@@ -1551,7 +1551,6 @@ struct AsmProcessor
 	AsmArgMode mode;
 
 	insnTemplate = new OutBuffer;
-	// %% todo: special case for something..
 	if (opInfo->linkType == Out_Mnemonic)
 	    mnemonic = alternateMnemonics[opInfo->link];
 	else
@@ -1831,9 +1830,27 @@ struct AsmProcessor
 		    Expression * e = (Expression *) operand->symbolDisplacement.data[0];
 		    Declaration * decl = 0;
 
+		    /* We are generating a memory reference, but the
+		       operand could be a floating point constant.  If this
+		       is the case, it will be turned into a const static
+		       VAR_DECL later in AsmStatement::toIR. */
+		    
 		    if (e->op == TOKvar)
 			decl = ((VarExp *) e)->var;
 
+		    /* The GCC will error (or just warn) on references to
+		       nonlocal vars, but it cannot catch closure vars. */
+		    
+		    /* TODO: Could allow this for the frame-relative
+		       case if it's a closure var.  We do not know if
+		       it is a closure var at this point, however.
+		       Just accept it for now and return false from
+		       getFrameRelativeOffset if it turns out not to be.
+		    */
+		    if (decl && !(decl->isDataseg() || decl->isCodeseg()) &&
+			decl->toParent2() != sc->func)
+			stmt->error("cannot access nonlocal variable %s", decl->toChars());
+		    
 		    if (operand->baseReg != Reg_Invalid &&
 			decl && ! decl->isDataseg()) {
 
@@ -2052,6 +2069,16 @@ struct AsmProcessor
 	    } else {
 		stmt->error("too many symbols in operand");
 	    }
+	} else if (exp->op == TOKfloat64) {
+	    if (exp->type && exp->type->isscalar())
+	    {
+		// DMD doesn't check Tcomplex*, and counts Tcomplex32 as Tfloat64
+		TY ty = exp->type->toBasetype()->ty;
+		operand->dataSizeHint = ty == Tfloat80 || ty == Timaginary80 ?
+		    Extended_Ptr : (PtrType) exp->type->size(0);
+	    }
+	    // Will be converted to a VAR_DECL in AsmStatement::toIR
+	    operand->symbolDisplacement.push( exp );
 	} else if (exp == Handled) {
 	    // nothing
 	} else {
@@ -2466,12 +2493,16 @@ struct AsmProcessor
 
 		e = e->semantic(sc);
 
-		// Special case for floating point constant declarations.
+		/* Special case for floating point literals: Try to
+		   resolve to a memory reference.  If not possible,
+		   leave it as a literal and generate an anonymous
+		   var in AsmStatement::toIR.
+		*/
 		if (e->op == TOKfloat64) {
 		    Dsymbol * sym = sc->search(stmt->loc, ident, NULL);
 		    if ( sym ) {
 			VarDeclaration *v = sym->isVarDeclaration();
-			if ( v ) {
+			if ( v && v->isDataseg() ) {
 			    Expression *ve = new VarExp(stmt->loc, v);
 			    ve->type = e->type;
 			    e = ve;
