@@ -658,7 +658,7 @@ handle_section_attribute (tree *node, tree ARG_UNUSED (name), tree args,
 	{
 	  if (TREE_CODE (decl) == VAR_DECL
 	      && current_function_decl != NULL_TREE
-	      && !TREE_STATIC (decl))
+	      && ! (TREE_STATIC (decl) || DECL_EXTERNAL (decl)))
 	    {
 	      error ("%Jsection attribute cannot be specified for "
 		     "local variables", decl);
