@@ -43,86 +43,75 @@ d_gcc_force_templates()
 	IRState::emitTemplates == TEall;
 }
 
-
 void
 IRState::emitLocalVar(VarDeclaration * v, bool no_init)
 {
     if (v->isDataseg() || (v->storage_class & STCfield))
 	return;
 
-    tree var_decl = v->toSymbol()->Stree;
-    tree init_exp = NULL_TREE;
+    Symbol * sym = v->toSymbol();
+    tree var_decl = sym->Stree;
 
-    assert( ! TREE_STATIC(var_decl ));
-    
+    gcc_assert(! TREE_STATIC( var_decl ));
     if (TREE_CODE(var_decl) == CONST_DECL)
 	return;
 
     DECL_CONTEXT( var_decl ) = getLocalContext();
-    pushdecl(var_decl);
 
-    if ( ! no_init && ! DECL_INITIAL( var_decl ) && v->init ) {
+    tree var_exp;
+#if V2
+    if (sym->SclosureField)
+	var_exp = var(v);
+    else
+#endif
+    {
+	var_exp = var_decl;
+	pushdecl(var_decl);
+#if D_GCC_VER < 40
+	expand_decl(var_decl);
+#endif
+    }
+
+    tree init_exp = NULL_TREE; // complete initializer expression (include MODIFY_EXPR, e.g.)
+    tree init_val = NULL_TREE;
+
+    if (! no_init && ! DECL_INITIAL( var_decl ) && v->init ) {
 	if (! v->init->isVoidInitializer()) {
 	    ExpInitializer * exp_init = v->init->isExpInitializer();
-	    assert(exp_init);
-
 	    Expression * ie = exp_init->toExpression();
-	    do {
-		if (ie->op == TOKassign) {
-		    AssignExp * a_exp = (AssignExp *) ie;
-		    if (a_exp->e1->op == TOKvar &&
-			((VarExp *) a_exp->e1)->var == v) {
-
-			DECL_INITIAL( var_decl ) = convertForAssignment(a_exp->e2, v->type);
-			break;
-		    }
-		}
+	    if ( ! (init_val = assignValue(ie, v)) )
 		init_exp = ie->toElem(this);
-	    } while (0);
-	} else {
-	    no_init = true;
 	}
+	else
+	    no_init = true;
     }
 
-    // %% Do we want to do this for special variables?
-    if (inVolatile() && DECL_INITIAL( var_decl ) != NULL_TREE) {
-	TREE_THIS_VOLATILE( DECL_INITIAL( var_decl )) = 1;
-    }
+    if (! no_init) {
+	g.ofile->doLineNote(v->loc);
+
+	if (! init_val)
+	    init_val = DECL_INITIAL(var_decl);
 #if D_GCC_VER < 40
-    expand_decl(var_decl);
-
-    if ( no_init ) {
-	// nothing
-    } else if ( init_exp ) {
-	g.ofile->doLineNote(v->loc);
-	expand_expr_stmt_value(init_exp, 0, 1);
-    } else if ( DECL_INITIAL( var_decl )) {
-	g.ofile->doLineNote(v->loc);
-	expand_decl_init(var_decl); // %%new: merge with above, irs->expandDecl
-    } else {
-	d_warning(OPT_Wuninitialized, "uninitialized variable '%s'", v->ident ? v->ident->string : "(no name)");
-    }
-#else
-    // pushdecl will allow the VAR_DECL to be placed in a BIND_EXPR when
-    // endBindings is called, but we want the initialization *here*
-    if (no_init) {
-	// nothing
-    } else if ( init_exp ) {
-	//g.ofile->doLineNote(v->loc);
-	//expand_expr_stmt_value(init_exp, 0, 1);
-    } else if ( DECL_INITIAL( var_decl )) {
-	init_exp = DECL_INITIAL( var_decl );
-	DECL_INITIAL( var_decl ) = NULL_TREE;
-	init_exp = build2(MODIFY_EXPR, void_type_node, var_decl, init_exp);
-	//g.ofile->doLineNote(v->loc);
-	//expand_decl_init(var_decl); // %%new: merge with above, irs->expandDecl
-    } else {
-	d_warning(OPT_Wuninitialized, "uninitialized variable '%s'", v->ident ? v->ident->string : "(no name)");
-    }
-	
-    if (init_exp)
-	addExp(init_exp);
+	if (init_val
+#if V2
+	    && ! sym->SclosureField
 #endif
+	    )
+	{
+	    // not sure if there is any advantage to doing this...
+	    DECL_INITIAL(var_decl) = init_val;
+	    expand_decl_init(var_decl);
+	}
+	else
+#endif
+	if (! init_exp && init_val)
+	    init_exp = vmodify(var_exp, init_val);
+
+	if (init_exp)
+	    addExp(init_exp);
+	else if (! init_val && v->size(v->loc)) // Zero-length arrays do not have an initializer
+	    d_warning(OPT_Wuninitialized, "uninitialized variable '%s'", v->ident ? v->ident->string : "(no name)");
+    }
 }
 
 tree
@@ -220,6 +209,48 @@ IRState::expandDecl(tree t_decl)
 #endif
 }
 
+#if V2    
+tree
+IRState::var(VarDeclaration * v)
+{
+    bool is_closure_var = v->toSymbol()->SclosureField != NULL;
+
+    /* // testing...
+    bool test2 = false;
+    Dsymbol * vp = v->toParent2();
+    if (vp)
+    {
+	FuncDeclaration * f = vp->isFuncDeclaration();
+	/* //testing...
+	if (f && getFrameInfo(f)->creates_closure)
+	{
+	    for (unsigned i = 0; i < f->closureVars.dim; ++i)
+		if (v == f->closureVars.data[i])
+		{
+		    test2 = true;
+		    break;
+		}
+	}
+	gcc_assert(is_closure_var == test2);
+    }
+    //...testing */
+    
+    if (is_closure_var)
+    {
+	FuncDeclaration * f = v->toParent2()->isFuncDeclaration();
+	
+	tree cf = getClosureRef(f);
+	tree field = v->toSymbol()->SclosureField;
+	gcc_assert(field != NULL_TREE);
+	return component(indirect(cf), field);
+    }
+    else
+	// Static var or auto var that the back end will handle for us
+	return v->toSymbol()->Stree;
+}
+#endif
+
+
 tree
 IRState::convertTo(Expression * exp, Type * target_type)
 {
@@ -235,7 +266,7 @@ IRState::convertTo(tree exp, Type * exp_type, Type * target_type)
     assert(exp_type);
     exp_type = exp_type->toBasetype();
     
-    if (typesCompatible(exp_type, target_type))
+    if (typesSame(exp_type, target_type))
 	return exp;
 
     switch (exp_type->ty) {
@@ -375,6 +406,12 @@ IRState::convertTo(tree exp, Type * exp_type, Type * target_type)
 	    return build1(NOP_EXPR, target_type->toCtype(), exp);
 	// else, default conversion, which should product an error
 	break;
+    case Tpointer:
+	/* For some reason, convert_to_integer converts pointers
+	   to a signed type. */
+	if (target_type->isintegral())
+	    exp = d_convert_basic(d_type_for_size(POINTER_SIZE, 1), exp);
+	break;
     default:
 	if (exp_type->isreal() && target_type->isimaginary() ||
 	    exp_type->isimaginary() && target_type->isreal()) {
@@ -691,7 +728,6 @@ IRState::assertCall(Loc loc, LibCall libcall)
     return libCall(libcall, 2, args);
 }
 
-volatile int baka;
 tree
 IRState::assertCall(Loc loc, Expression * msg)
 {
@@ -821,13 +857,18 @@ IRState::call(Expression * expr, /*TypeFunction * func_type, */ Array * argument
     } else if (expr->op == TOKvar) {
 	FuncDeclaration * fd = ((VarExp *) expr)->var->isFuncDeclaration();
 	tf = (TypeFunction *) fd->type;
-	if (fd && fd->isNested()) {
+	if (fd) {
+	    if (fd->isNested()) {
 #if D_NO_TRAMPOLINES
-	    object = getStaticChain(fd);
+		object = getFrameForFunction(fd);
 #else
-	    // Pass fake argument for nested functions
-	    object = d_null_pointer;
+		// Pass fake argument for nested functions
+		object = d_null_pointer;
 #endif
+	    } else if (fd->needThis()) {
+		expr->error("need 'this' to access member %s", fd->toChars());
+		object = d_null_pointer; // continue processing...
+	    }
 	}
     } else {
 	tf = get_function_type(t);
@@ -956,8 +997,9 @@ static const char * libcall_ids[LIBCALL_count] =
       "_D9invariant12_d_invariantFC6ObjectZv",
       "_d_newclass", "_d_newarrayT",
       "_d_newarrayiT",
-      "_d_newarraymTp", "_d_newarraymiTp",
-      "_d_delclass", "_d_delarray", "_d_delmemory", "_d_callfinalizer",
+      "_d_newarraymTp", "_d_newarraymiTp", "_d_allocmemory",
+      "_d_delclass", "_d_delinterface", "_d_delarray",
+      "_d_delmemory", "_d_callfinalizer", "_d_callinterfacefinalizer",
       "_d_arraysetlengthT", "_d_arraysetlengthiT",
       "_d_dynamic_cast", "_d_interface_cast",
       "_adEq", "_adCmp", "_adCmpChar",
@@ -969,6 +1011,10 @@ static const char * libcall_ids[LIBCALL_count] =
       "_d_arraycatT", "_d_arraycatnT",
       "_d_arrayappendT",
       /*"_d_arrayappendc", */"_d_arrayappendcTp",
+#if V2
+      "_d_arrayassign", "_d_arrayctor", "_d_arraysetassign",
+      "_d_arraysetctor",
+#endif
       "_d_monitorenter", "_d_monitorexit",
       "_d_criticalenter", "_d_criticalexit",
       "_d_throw",
@@ -1053,8 +1099,13 @@ IRState::getLibCallDecl(LibCall lib_call)
 	    if (lib_call == LIBCALL_NEWARRAYMITP)
 		arg_types.push( Type::tsize_t );
 	    break;
+	case LIBCALL_ALLOCMEMORY:
+	    arg_types.push( Type::tsize_t );
+	    return_type = Type::tvoidptr;
+	    break;
 	case LIBCALL_DELCLASS:
-	    arg_types.push(getObjectType()->pointerTo());
+	case LIBCALL_DELINTERFACE:
+	    arg_types.push(Type::tvoid->pointerTo());
 	    break;
 	case LIBCALL_DELARRAY:
 	    arg_types.push(Type::tvoid->arrayOf()->pointerTo());
@@ -1063,7 +1114,8 @@ IRState::getLibCallDecl(LibCall lib_call)
 	    arg_types.push(Type::tvoid->pointerTo()->pointerTo());
 	    break;
 	case LIBCALL_CALLFINALIZER:
-	    arg_types.push(Type::tvoid->pointerTo()); // declared void*; could be Object....
+	case LIBCALL_CALLINTERFACEFINALIZER:
+	    arg_types.push(Type::tvoid->pointerTo());
 	    break;
 	case LIBCALL_ARRAYSETLENGTHT:
 	case LIBCALL_ARRAYSETLENGTHIT:
@@ -1181,6 +1233,22 @@ IRState::getLibCallDecl(LibCall lib_call)
 	    arg_types.push(Type::tvoid->pointerTo()); // varargs = true;
 	    return_type = Type::tvoid->arrayOf();
 	    break;
+#if V2
+	case LIBCALL_ARRAYASSIGN:
+	case LIBCALL_ARRAYCTOR:
+	    arg_types.push(Type::typeinfo->type);
+	    arg_types.push(Type::tvoid->arrayOf());
+	    return_type = Type::tvoid->arrayOf();
+	    break;
+	case LIBCALL_ARRAYSETASSIGN:
+	case LIBCALL_ARRAYSETCTOR:
+	    arg_types.push(Type::tvoid->pointerTo());
+	    arg_types.push(Type::tvoid->pointerTo());
+	    arg_types.push(Type::tsize_t);
+	    arg_types.push(Type::typeinfo->type);
+	    return_type = Type::tvoid->pointerTo();
+	    break;
+#endif
 	case LIBCALL_MONITORENTER:
 	case LIBCALL_MONITOREXIT:
 	case LIBCALL_THROW:
@@ -1870,6 +1938,21 @@ IRState::imagPart(tree c) {
     return build1(IMAGPART_EXPR, TREE_TYPE(TREE_TYPE(c)), c);
 }
 
+tree
+IRState::assignValue(Expression * e, VarDeclaration * v)
+{
+    if (e->op == TOKassign || e->op == TOKconstruct || e->op == TOKblit)
+    {
+	AssignExp * a_exp = (AssignExp *) e;
+	if (a_exp->e1->op == TOKvar && ((VarExp *) a_exp->e1)->var == v)
+	    return convertForAssignment(a_exp->e2, v->type);
+	//else
+	    //return e->toElem(this);
+    }
+    return NULL_TREE;
+}
+
+
 tree
 IRState::twoFieldType(tree rec_type, tree ft1, tree ft2, Type * d_type, const char * n1, const char * n2)
 {
@@ -2213,8 +2296,16 @@ IRState::attributes(Expressions * in_attrs)
 
 	    if (c->arguments) {
 		for (unsigned ai = 0; ai < c->arguments->dim; ai++) {
-		    // g.irs?
-		    args.cons( ((Expression *) c->arguments->data[ai])->toElem(&gen) );
+		    Expression * ae = (Expression *) c->arguments->data[ai];
+		    tree aet;
+		    if (ae->op == TOKstring && ((StringExp *) ae)->sz == 1)
+		    {
+			StringExp * s = (StringExp *) ae;
+			aet = build_string(s->len, (const char*) s->string);
+		    }
+		    else
+			aet = ae->toElem(&gen);
+		    args.cons(aet);
 		}
 	    }
 	}
@@ -2289,120 +2380,333 @@ IRState::label(Loc loc, Identifier * ident) {
 }
 
 tree
-IRState::getStaticChain(FuncDeclaration * f)
+IRState::getFrameForFunction(FuncDeclaration * f)
 {
     if (f->fbody)
-	return getStaticChain(f->toSymbol()->Stree);
-    else {
+	return getFrameForSymbol(f);
+    else
+    {
 	// Should error on line that references f
 	f->error("nested function missing body");
 	return d_null_pointer;
     }
 }
+tree
+IRState::getFrameForNestedClass(ClassDeclaration *c)
+{
+    return getFrameForSymbol(c);
+}
+
+/* If nested_sym is a nested function, return the static chain to be
+   used when invoking that function.
+
+   If nested_sym is a nested class, return the static chain to be used
+   when creating an instance of the class.
+
+   This method is protected to enforce the type checking of
+   getFrameForFunction and getFrameForNestedClass.
+   getFrameForFunction also checks that the nestd function is properly
+   defined.
+*/
 
 tree
-IRState::getStaticChainOfNestedClass(tree t)
+IRState::getFrameForSymbol(Dsymbol * nested_sym)
 {
-    gcc_assert(TYPE_P( t ));
-    return getStaticChain(t);
-}
+    FuncDeclaration * nested_func = 0;
+    FuncDeclaration * outer_func = 0;
 
-/* Return the static chain to be used when invoking the nested function F.
-   F may also be a type with a function context and indicates the
-   static chain is for some function nested directly in that context.  This
-   allowed so that it is not necessary to find a method of a nested
-   class.
+    if ( (nested_func = nested_sym->isFuncDeclaration()) )
+    {
+	// gcc_assert(nested_func->isNested())
+	outer_func = nested_func->toParent2()->isFuncDeclaration();
+	gcc_assert(outer_func != NULL);
+    }
+    else
+    {
+	/* It's a class.  NewExp::toElem has already determined its
+	   outer scope is not another class, so it must be a
+	   function. */
 
-   This is protected to enforce the check in
-   getStaticChain(FuncDeclaration *).
-*/
-tree IRState::getStaticChain(tree t)
-{
-#if D_GCC_VER < 40
-    rtx chain_rtx = NULL_RTX;
-#else
-    tree context_f = NULL_TREE;
-#endif
-    
-    if (TYPE_P( t )) {
-	/* Consider:
-
-	f() {
-	  class X {
-	    m() { }
-	  }
-	  g() {
-	    h() {
-	       new X // <-- you are here
-	    }
-	  }
-	}
-
-	Search upwards through the nested function chain until a function
-	at the same level as the methods X of is found.  The result may,
-	in fact, be a method of X.  If the current function is F, just
-	use virtual_stack_vars_rtx.
-
-	*/
+	Dsymbol * sym = nested_sym;
 	
-	do  {
-	    t = get_containing_scope(t);
-	} while (t && TREE_CODE( t ) != FUNCTION_DECL );
+	while (sym && ! (outer_func = sym->isFuncDeclaration()))
+	    sym = sym->toParent2();
 
-	/* Make sure we can access the frame. */
+	/* Make sure we can access the frame of outer_func.
 
-	if (t != current_function_decl) {
-	    tree f = current_function_decl;
-	    tree c = f;
+	   For GCC < 4.0:
 
+	    f() {
+	      class X {
+		m() { }
+	      }
+	      g() {
+		h() {
+		   new X // <-- you are here
+		}
+	      }
+	    }
+
+	   In order to get the static chain we must know a
+	   function nested in F.  If we are currently in such a
+	   nested function, use that.
+
+	   f() {
+	       class X { m() { } }
+	       new X // <-- you are here
+	   }
+
+	   If we are at the level of the function containing the
+	   class, the answer is just 'virtual_stack_vars_rtx'.
+	*/
+
+	if (outer_func != func) {
+	       
+	    Dsymbol * o = nested_func = func;
 	    do {
-		f = c;
-		if (DECL_NO_STATIC_CHAIN(f))
+		if (! nested_func->isNested())
 		    goto cannot_access_frame;
-		c = decl_function_context(f);
-	    } while (c && c != t);
-	    
-	    if (! c) {
+		while ( (o = o->toParent2()) )
+		    if ( (nested_func = o->isFuncDeclaration()) )
+			break;
+	    } while (o && o != outer_func);
+
+	    if (! o) {
 	    cannot_access_frame:
 		error("cannot access frame of function '%s' from '%s'",
-		    IDENTIFIER_POINTER (DECL_NAME (t)),
-		    func->ident->string);
+		    outer_func->ident->string, func->ident->string);
 		return d_null_pointer;
 	    }
-
-	    t = f;
-	} else {
-#if D_GCC_VER < 40
-	    chain_rtx = virtual_stack_vars_rtx;
-#else
-	    context_f = t;
-#endif
 	}
+	// else, the answer is 'virtual_stack_vars_rtx'
     }
-	
-#if D_GCC_VER < 40
-    if (! chain_rtx)
-	chain_rtx = lookup_static_chain(t);
 
+#if V2
+    if (getFrameInfo(outer_func)->creates_closure)
+	return getClosureRef(outer_func);
+#endif
+
+#if D_GCC_VER < 40
     tree result = make_node (RTL_EXPR);
     TREE_TYPE (result) = ptr_type_node;
-    RTL_EXPR_RTL (result) = chain_rtx;
+    RTL_EXPR_RTL (result) = nested_func ?
+	lookup_static_chain(nested_func->toSymbol()->Stree) :
+	virtual_stack_vars_rtx;
     return result;
 #else
-    if (! context_f)
-	context_f = decl_function_context(t); // duplicates last decl_function_context call
-    gcc_assert(context_f != NULL_TREE && TREE_CODE(context_f) == FUNCTION_DECL);
-    return build1(STATIC_CHAIN_EXPR, ptr_type_node, context_f);
+    if (! outer_func)
+	outer_func = nested_func->toParent2()->isFuncDeclaration();
+    gcc_assert(outer_func != NULL);
+    return build1(STATIC_CHAIN_EXPR, ptr_type_node, outer_func->toSymbol()->Stree);
 #endif
 }
 
+#if V2
+
+/* For the purposes this is used, fd is assumed to be a nested
+   function or a method of a class that is (eventually) nested in a
+   function.
+*/
+static bool
+isFuncNestedInFunc(FuncDeclaration * fd, FuncDeclaration *fo)
+{
+    if (fd == fo)
+	return false;
+    
+    while (fd)
+    {
+	AggregateDeclaration * ad;
+	ClassDeclaration * cd;
+
+	if (fd == fo)
+	{
+	    //fprintf(stderr, "%s is nested in %s\n", fd->toChars(), fo->toChars());
+	    return true;
+	}
+	else if (fd->isNested())
+	    fd = fd->toParent2()->isFuncDeclaration();
+	else if ( (ad = fd->isThis()) && (cd = ad->isClassDeclaration()) )
+	{
+	    fd = NULL;
+	    while (cd && cd->isNested())
+	    {
+		Dsymbol * dsym = cd->toParent2();
+		if ( (fd = dsym->isFuncDeclaration()) )
+		    break;
+		else
+		    cd = dsym->isClassDeclaration();
+	    }
+	}
+	else
+	    break;
+    }
+    
+    //fprintf(stderr, "%s is NOT nested in %s\n", fd->toChars(), fo->toChars());
+    return false;
+}
+
+FuncFrameInfo *
+IRState::getFrameInfo(FuncDeclaration *fd)
+{
+    Symbol * fds = fd->toSymbol();
+    if (fds->frameInfo)
+	return fds->frameInfo;
+
+    FuncFrameInfo * ffi = new FuncFrameInfo;
+    ffi->creates_closure = false;
+    ffi->closure_rec = NULL_TREE;
+
+    fds->frameInfo = ffi;
+    
+    Dsymbol * s = fd->toParent2();
+
+    if (fd->needsClosure())
+	ffi->creates_closure = true;
+    else
+    {
+	/* If fd is nested (deeply) in a function 'g' that creates a
+	   closure and there exists a function 'h' nested (deeply) in
+	   fd and 'h' accesses the frame of 'g', then fd must also
+	   create a closure.
+
+	   This is for the sake of a simple implementation.  An alternative
+	   is, when determining the frame to pass to 'h', pass the pointer
+	   to 'g' (the deepest 'g' whose frame is accessed by 'h') instead
+	   of the usual frame that 'h' would take.
+	*/
+	FuncDeclaration * ff = fd;
+
+	while (ff)
+	{
+	    AggregateDeclaration * ad;
+	    ClassDeclaration * cd;
+
+	    if (ff != fd)
+	    {
+		if (getFrameInfo(ff)->creates_closure)
+		{
+		    for (int i = 0; i < ff->closureVars.dim; i++)
+		    {   VarDeclaration *v = (VarDeclaration *)ff->closureVars.data[i];
+			for (int j = 0; j < v->nestedrefs.dim; j++)
+			{   FuncDeclaration *fi = (FuncDeclaration *)v->nestedrefs.data[j];
+			    if (isFuncNestedInFunc(fi, fd))
+			    {
+				ffi->creates_closure = true;
+				goto L_done;
+			    }
+			}
+		    }
+		}
+	    }
+	    
+	    if (ff->isNested())
+		ff = ff->toParent2()->isFuncDeclaration();
+	    else if ( (ad = ff->isThis()) && (cd = ad->isClassDeclaration()) )
+	    {
+		ff = NULL;
+		while (cd && cd->isNested())
+		{
+		    Dsymbol * dsym = cd->toParent2();
+		    if ( (ff = dsym->isFuncDeclaration()) )
+			break;
+		    else
+			cd = dsym->isClassDeclaration();
+		}
+	    }
+	    else
+		break;
+	}
+
+	L_done:
+	    ;
+    }
+
+    /*fprintf(stderr, "%s  %s\n", ffi->creates_closure ? "YES" : "NO ",
+      fd->toChars());*/
+
+    return ffi;
+}
+
+// Return a pointer to the closure block of outer_func
+tree
+IRState::getClosureRef(FuncDeclaration * outer_func)
+{
+    tree result = closureLink();
+    FuncDeclaration * fd = closureFunc;
+
+    while (fd && fd != outer_func)
+    {
+	AggregateDeclaration * ad;
+	ClassDeclaration * cd;
+
+	gcc_assert(getFrameInfo(fd)->creates_closure); // remove this if we loosen creates_closure
+
+	// like compon(indirect, field0) parent closure link is the first field;
+	result = indirect(result, ptr_type_node);
+
+	if (fd->isNested())
+	{
+	    fd = fd->toParent2()->isFuncDeclaration();
+	}
+	else if ( (ad = fd->isThis()) && (cd = ad->isClassDeclaration()) )
+	{
+	    fd = NULL;
+	    while (cd && cd->isNested())
+	    {
+		/* Shouldn't need to do this.  getClosureRef is only
+		   used to get the pointer to a function's frame (not
+		   a class instances.)  With the current implementation,
+		   the link the closure record always points to the
+		   outer function's frame even if there are intervening
+		   nested classes.  So, we can just skip over those...
+		/*
+		tree vthis_field = cd->vthis->toSymbol()->Stree;
+		result = nop(result, cd->type->toCtype());
+		result = component(indirect(result), vthis_field);
+		*/
+
+		Dsymbol * dsym = cd->toParent2();
+		if ( (fd = dsym->isFuncDeclaration()) )
+		    break;
+		else
+		    cd = dsym->isClassDeclaration();
+	    }
+	}
+	else
+	    break;
+    }
+    
+    if (fd == outer_func)
+    {
+	tree closure_rec = getFrameInfo(outer_func)->closure_rec;
+	result = nop(result, build_pointer_type(closure_rec));
+	return result;
+    }
+    else
+    {
+	func->error("cannot access frame of %s", outer_func->toChars());
+	return d_null_pointer;
+    }
+}
+
+#endif
+
+/* Return true if function F needs to have the static chain passed to
+   it.  This only applies to nested function handling provided by the
+   GCC back end (not D closures.)
+*/
 bool
 IRState::functionNeedsChain(FuncDeclaration *f)
 {
     Dsymbol * s;
     ClassDeclaration * a;
-    
-    if (f->isNested())
+    FuncDeclaration *pf = 0;
+
+    if (f->isNested()
+#if V2
+	&& ! getFrameInfo(f->toParent2()->isFuncDeclaration())->creates_closure
+#endif
+	)
 	return true;
     if (f->isStatic())
 	return false;
@@ -2411,7 +2715,11 @@ IRState::functionNeedsChain(FuncDeclaration *f)
 
     while ( s && (a = s->isClassDeclaration()) && a->isNested() ) {
 	s = s->toParent2();
-	if (s->isFuncDeclaration())
+	if ( (pf = s->isFuncDeclaration())
+#if V2
+	    && ! getFrameInfo(pf)->creates_closure
+#endif
+	     )
 	    return true;
     }
     return false;
@@ -3014,9 +3322,18 @@ ArrayScope::finish(tree e)
 {
     if (v)
     {
-	tree t = v->toSymbol()->Stree;
+	Symbol * s = v->toSymbol();
+	tree t = s->Stree;
 	if (TREE_CODE(t) == VAR_DECL)
+	{
+#if V2
+	    if (s->SclosureField)
+		return irs->compound( irs->vmodify(irs->var(v),
+			DECL_INITIAL(t) ), e);
+	    else
+#endif
 	    return gen.binding(v->toSymbol()->Stree, e);
+	}
     }
 
     return e;
