@@ -38,10 +38,13 @@ typedef enum {
     LIBCALL_NEWARRAYMTP,
     //LIBCALL_NEWARRAYMIT,
     LIBCALL_NEWARRAYMITP,
+    LIBCALL_ALLOCMEMORY,
     LIBCALL_DELCLASS,
+    LIBCALL_DELINTERFACE,
     LIBCALL_DELARRAY,
     LIBCALL_DELMEMORY,
     LIBCALL_CALLFINALIZER,
+    LIBCALL_CALLINTERFACEFINALIZER,
     LIBCALL_ARRAYSETLENGTHT,
     LIBCALL_ARRAYSETLENGTHIT,
     LIBCALL_DYNAMIC_CAST,
@@ -65,6 +68,12 @@ typedef enum {
     LIBCALL_ARRAYAPPENDT,
     //LIBCALL_ARRAYAPPENDCT,
     LIBCALL_ARRAYAPPENDCTP,
+#if V2
+    LIBCALL_ARRAYASSIGN,
+    LIBCALL_ARRAYCTOR,
+    LIBCALL_ARRAYSETASSIGN,
+    LIBCALL_ARRAYSETCTOR,
+#endif
     LIBCALL_MONITORENTER,
     LIBCALL_MONITOREXIT,
     LIBCALL_CRITICALENTER,
@@ -80,6 +89,12 @@ typedef enum {
     LIBCALL_count
 } LibCall;
 
+struct FuncFrameInfo
+{
+    bool creates_closure;
+    tree closure_rec;
+};
+
 class ArrayScope;
 
 // Code generation routines should be in a separate namespace, but so many
@@ -117,6 +132,12 @@ struct IRState : IRBase
     tree exprVar(tree t_type);
     tree maybeExprVar(tree exp, tree * out_var);
     void expandDecl(tree t_decl);
+
+#if V2    
+    tree var(VarDeclaration * v);
+#else
+    tree var(VarDeclaration * v) { return v->toSymbol()->Stree; }
+#endif
     
     // ** Type conversion
     
@@ -132,9 +153,16 @@ struct IRState : IRBase
 	return convertForCondition(exp->toElem(this), exp->type); }
     tree convertForCondition(tree exp_tree, Type * exp_type);
     tree toDArray(Expression * exp);
+    static bool typesSame(Type * t1, Type * t2) {
+#if V2
+	return t1->mutableOf()->equals(t2->mutableOf());
+#else
+	return t1->equals(t2);
+#endif
+    }
     static bool typesCompatible(Type * t1, Type * t2) {
 #if V2
-	return t1->constConv(t2) >= MATCHconst;
+	return t1->implicitConvTo(t2) >= MATCHconst;
 #else
 	return t1->equals(t2);
 #endif
@@ -317,6 +345,14 @@ struct IRState : IRBase
 	else
 	    return build(COMPOUND_EXPR, TREE_TYPE(b), a, b);
     }
+    tree maybeVoidCompound(tree a, tree b) {
+	if (a == NULL_TREE)
+	    return b;
+	else if (b == NULL_TREE)
+	    return a;
+	else
+	    return build(COMPOUND_EXPR, void_type_node, a, b);
+    }
 
     static tree component(tree v, tree f)
     {
@@ -366,6 +402,8 @@ struct IRState : IRBase
 	return t;
     }
 
+    tree assignValue(Expression * e, VarDeclaration * v);
+
     static TemplateEmission emitTemplates;
     static bool splitDynArrayVarArgs;
     static bool useBuiltins;
@@ -405,10 +443,31 @@ public:
     static tree label(Loc loc, Identifier * ident = 0);
 
     // Static chain for nested functions
-    tree getStaticChain(FuncDeclaration * f);
-    tree getStaticChainOfNestedClass(tree t);
+    tree getFrameForFunction(FuncDeclaration * f);
+    tree getFrameForNestedClass(ClassDeclaration * c);
+#if V2
+    void useClosure(FuncDeclaration * f, tree l) {
+	_closureLink = l;
+	closureFunc = f;
+    }
+    void useParentClosure() {
+	if (parent)
+	{
+	    _closureLink = ((IRState*)parent)->_closureLink;
+	    closureFunc = ((IRState*)parent)->closureFunc;
+	}
+    }
+    tree closureLink() { return _closureLink; }
+#endif
 protected:
-    tree getStaticChain(tree t);
+    tree getFrameForSymbol(Dsymbol * nested_sym);
+#if V2
+    tree getClosureRef(FuncDeclaration *outer_func);
+    FuncDeclaration * closureFunc;
+    tree _closureLink;
+public:
+    static FuncFrameInfo * getFrameInfo(FuncDeclaration *fd);
+#endif
 public:
     static bool functionNeedsChain(FuncDeclaration *f);
 
