@@ -194,13 +194,68 @@ Symbol *VarDeclaration::toSymbol()
 	    ident_to_use = ident->string;
 
 	enum tree_code decl_kind;
+
+#if ! V2
+	bool is_template_const = false;
+	// Logic copied from toobj.c VarDeclaration::toObjFile
+
+	Dsymbol * parent = this->toParent();
+#if 1	/* private statics should still get a global symbol, in case
+	 * another module inlines a function that references it.
+	 */
+	if (/*protection == PROTprivate ||*/
+	    !parent || parent->ident == NULL || parent->isFuncDeclaration())
+	{
+	    // nothing
+	}
+	else
+#endif
+	{
+	    do
+	    {
+		/* Global template data members need to be in comdat's
+		 * in case multiple .obj files instantiate the same
+		 * template with the same types.
+		 */
+		if (parent->isTemplateInstance() && !parent->isTemplateMixin())
+		{
+		    /* These symbol constants have already been copied,
+		     * so no reason to output them.
+		     * Note that currently there is no way to take
+		     * the address of such a const.
+		     */
+		    if (isConst() && type->toBasetype()->ty != Tsarray &&
+			init && init->isExpInitializer())
+		    {
+			is_template_const = true;
+			break;
+		    }
+
+		    break;
+		}
+		parent = parent->parent;
+	    } while (parent);
+	}
+#endif
+	
 	if (storage_class & STCparameter)
 	    decl_kind = PARM_DECL;
-	else if (isConst() && ! gen.isDeclarationReferenceType( this ) &&
-	    type->isscalar() &&
-	    /* i386 asm needs VAR_DECL for loading floating point constants.
-	       That should be fixed instead of this exception. */
-	    ! type->isfloating() 
+	else if (
+#if V2
+# if 0
+		 // from VarDeclaration::getConstInitializer
+		 (isConst() || isInvariant()) && (storage_class & STCinit)
+# else
+		 (storage_class & STCmanifest)
+# endif
+#else
+		 is_template_const ||
+		 (
+		   isConst()
+		   && ! gen.isDeclarationReferenceType( this ) &&
+		   type->isscalar() && ! isDataseg()
+		 )
+#endif
 		 )
 	    decl_kind = CONST_DECL;
 	else
@@ -248,13 +303,23 @@ Symbol *VarDeclaration::toSymbol()
 		e = type->defaultInit();
 
 	    if (e)
-		DECL_INITIAL( var_decl ) = e->toElem(g.irs);
+	    {
+		DECL_INITIAL( var_decl ) = g.irs->assignValue(e, this);
+		if (! DECL_INITIAL(var_decl))
+		    DECL_INITIAL( var_decl ) = e->toElem(g.irs);
+	    }
 	}
 
 	// Can't set TREE_STATIC, etc. until we get to toObjFile as this could be
 	// called from a varaible in an imported module
 	// %% (out const X x) doesn't mean the reference is const...
-	if ( isConst() && ! gen.isDeclarationReferenceType( this )) {
+	if (
+#if V2
+	    (isConst() || isInvariant()) && (storage_class & STCinit)
+#else
+	    isConst()
+#endif
+	    && ! gen.isDeclarationReferenceType( this )) {
 	    // %% CONST_DECLS don't have storage, so we can't use those,
 	    // but it would be nice to get the benefit of them (could handle in
 	    // VarExp -- makeAddressOf could switch back to the VAR_DECL
@@ -271,7 +336,13 @@ Symbol *VarDeclaration::toSymbol()
 
 #if D_GCC_VER < 40
 	// With cgraph, backend can figure it out (improves inline nested funcs?)
-	if (nestedref && decl_kind != CONST_DECL) {
+	if (
+#if V2
+	    nestedrefs.dim
+#else
+	    nestedref
+#endif
+	    && decl_kind != CONST_DECL) {
 	    DECL_NONLOCAL( var_decl ) = 1;
 	    TREE_ADDRESSABLE( var_decl ) = 1;
 	}
@@ -669,10 +740,12 @@ Symbol *static_sym()
  */
 
 // Not used in GCC
-Classsym *fake_classsym(char * /*name*/)
+/*
+Classsym *fake_classsym(char * name)
 {
     return 0;
 }
+*/
 
 /*************************************
  * Create the "ClassInfo" symbol
@@ -746,7 +819,6 @@ Symbol *Module::toSymbol()
 	tree decl = build_decl(VAR_DECL, get_identifier(csym->Sident),
 	    TREE_TYPE(some_type->toCtype())); // want the RECORD_TYPE, not the REFERENCE_TYPE
 	csym->Stree = decl;
-	csym->ScontextDecl = g.ofile->fileContext;
 	
 	dkeep(decl);
 	
@@ -843,7 +915,6 @@ Symbol *AggregateDeclaration::toInitializer()
 Symbol *TypedefDeclaration::toInitializer()
 {
     Symbol *s;
-    Classsym *stag;
 
     if (!sinit)
     {
@@ -860,7 +931,6 @@ Symbol *TypedefDeclaration::toInitializer()
 Symbol *EnumDeclaration::toInitializer()
 {
     Symbol *s;
-    Classsym *stag;
 
     if (!sinit)
     {
@@ -878,6 +948,17 @@ Symbol *EnumDeclaration::toInitializer()
 	slist_add(s);
 	sinit = s;
     }
+    if (! sinit->Stree && g.ofile != NULL)
+    {
+	tree t = build_decl(VAR_DECL, get_identifier(sinit->Sident), type->toCtype());
+	sinit->Stree = t;
+	dkeep(t);
+
+	g.ofile->setupStaticStorage(this, t);
+	TREE_CONSTANT( t ) = 1;
+	TREE_READONLY( t ) = 1;
+	DECL_CONTEXT( t ) = 0;
+    }
     return sinit;
 }
 
