@@ -110,8 +110,13 @@ static char lang_name[6] = "GNU D";
 
 #if D_GCC_VER >= 40
 #define LANG_HOOKS_TYPES_COMPATIBLE_P d_types_compatible_p
+
+#if V2
+#undef LANG_HOOKS_TREE_INLINING_CONVERT_PARM_FOR_INLINING
+#define LANG_HOOKS_TREE_INLINING_CONVERT_PARM_FOR_INLINING d_convert_parm_for_inlining
 #endif
 
+#endif
 
 
 ////static tree d_type_for_size PARAMS ((unsigned, int));
@@ -149,6 +154,7 @@ d_init_options (unsigned int, const char **)
     global.params.warnings = 0;
     global.params.obj = 1;
     global.params.Dversion = 2;
+    global.params.quiet = 1;
 
     global.params.linkswitches = new Array();
     global.params.libfiles = new Array();
@@ -279,6 +285,35 @@ d_init ()
        Intel-specific. */
 #ifdef TARGET_64BIT
     global.params.isX86_64 = TARGET_64BIT ? 1 : 0;
+#else
+    /* TARGET_64BIT is only defined on biarched archs defaulted to 64-bit
+     * (as amd64 or s390x) so for full 64-bit archs (as ia64 or alpha) we
+     * need to test it more. */
+#  ifdef D_CPU_VERSYM64
+	/* We are "defaulting" to 32-bit, which mean that if both D_CPU_VERSYM
+	 * and D_CPU_VERSYM64 are defined, and not TARGET_64BIT, we will use
+	 * 32 bits. This will be overidden for full 64-bit archs */
+	global.params.isX86_64 = 0;
+#    ifndef D_CPU_VERSYM
+	/* So this is typically for alpha and ia64 */
+	global.params.isX86_64 = 1;
+#    endif
+#  else
+#    ifdef D_CPU_VERSYM /* D_CPU_VERSYM is defined and D_CPU_VERSYM64 is not. */
+	global.params.isX86_64 = 0;
+#    else
+	/* If none of D_CPU_VERSYM and D_CPU_VERSYM64 defined check size_t
+	 * length instead. */
+	switch (sizeof(size_t)) {
+	    case 4:
+	    	global.params.isX86_64 = 0;
+		break;
+	    case 8:
+	        global.params.isX86_64 = 1;
+		break;
+	}
+#    endif
+#  endif
 #endif
     
     Type::init();
@@ -291,13 +326,16 @@ d_init ()
     maybe_fixup_cygwin();
 
     VersionCondition::addPredefinedGlobalIdent("GNU");
-#ifdef D_CPU_VERSYM
+#ifdef D_CPU_VERSYM64
+    if (global.params.isX86_64 == 1)
+    	cpu_versym = D_CPU_VERSYM64;
+#  ifdef D_CPU_VERSYM
+    else
+	cpu_versym = D_CPU_VERSYM;
+#  endif
+#else
+#  ifdef D_CPU_VERSYM
     cpu_versym = D_CPU_VERSYM;
-#endif
-#ifdef TARGET_64BIT
-#  ifdef D_CPU_VERSYM64
-    if (TARGET_64BIT)
-	cpu_versym = D_CPU_VERSYM64;
 #  endif
 #endif
     if (cpu_versym)
@@ -313,6 +351,13 @@ d_init ()
 	is_target_win32 = true;
 #endif
 
+#ifdef TARGET_THUMB
+    if (TARGET_THUMB)
+        VersionCondition::addPredefinedGlobalIdent("Thumb");
+    else
+        VersionCondition::addPredefinedGlobalIdent("Arm");
+#endif
+
     if (BYTES_BIG_ENDIAN)
 	VersionCondition::addPredefinedGlobalIdent("BigEndian");
     else
@@ -338,6 +383,8 @@ d_init ()
     //if (global.params.cov)
     if (flag_test_coverage)
 	VersionCondition::addPredefinedGlobalIdent("D_Coverage");
+    if (global.params.useUnitTests)
+	VersionCondition::addPredefinedGlobalIdent("unittest");
 
     {
 	line_maps lm;
@@ -530,6 +577,9 @@ d_handle_option (size_t scode, const char *arg, int value)
       case OPT_fdebug_c:
 	  strcpy(lang_name, value ? "GNU C" : "GNU D");
 	  break;
+      case OPT_fignore_unknown_pragmas:
+	  global.params.ignoreUnsupportedPragmas = value;
+	  break;
 #ifdef _DH
       case OPT_fintfc:
 	  global.params.doHdrGeneration = value;
@@ -658,10 +708,30 @@ d_write_global_declarations()
     tree * vec = (tree *) globalFunctions.data;
     wrapup_global_declarations(vec, globalFunctions.dim);
     check_global_declarations(vec, globalFunctions.dim);
-#if D_GCC_VER >= 40    
-    cgraph_optimize ();
+#if D_GCC_VER >= 40
+
+#if D_GCC_VER >= 41
+    cgraph_optimize();
+#endif
+    
+
     for (int i = 0; i < globalFunctions.dim; i++)
 	debug_hooks->global_decl(vec[i]);
+
+#if D_GCC_VER == 40
+    /* For 4.0.x, if cgraph_optimize is called before the loop over
+       globalFunctions, dwarf2out can generate a DIE tree that has a
+       nested class as the parent of an outer class (this causes an
+       ICE.) Not sure if this is due to a bug in the D front end, but calling
+       debug_hooks->global_decl ensures the outer elements are
+       processed first.
+
+       For later versions, calling cgraph_optimize later causes more
+       problems.
+    */
+    cgraph_optimize();
+#endif
+    
 #endif
 }
 
@@ -966,7 +1036,7 @@ d_parse_file (int /*set_yydebug*/)
 	if (global.params.verbose)
 	    printf("code      %s\n", m->toChars());
 	if (! flag_syntax_only)
-	    m->genobjfile();
+	    m->genobjfile(false);
 	if (! global.errors && ! errorcount)
 	{
 	    if (global.params.doDocComments)
@@ -1646,6 +1716,32 @@ d_types_compatible_p (tree x, tree y)
 	
     return TYPE_MAIN_VARIANT (x) == TYPE_MAIN_VARIANT (y);
 }
+
+#if V2
+
+/* DMD 2 makes a parameter delclaration's type 'const(T)' if the
+   parameter is a simple STCin or STCconst.  The TypeFunction's
+   Argument's type stays unqualified, however.
+
+   This mismatch causes a problem with optimization and inlining.  For
+   RECORD_TYPE arguments, failure will occur in (setup_one_parameter
+   -> fold_convert).  d_types_compatible_p hacks lead to failures in
+   the sra pass.
+
+   Fortunately, the middle end provides a simple workaround by using
+   this hook.
+*/
+
+tree
+d_convert_parm_for_inlining  (tree parm, tree value, tree fndecl, int argnum)
+{
+    if (TREE_TYPE(parm) != TREE_TYPE(value))
+	return build1(NOP_EXPR, TREE_TYPE(parm), value);
+    else
+	return value;
+}
+#endif
+    
 #endif
 
 
