@@ -24,7 +24,6 @@ unsigned  ObjectFile::moduleSearchIndex;
 Array ObjectFile::deferredThunks;
 Array ObjectFile::staticCtorList;
 Array ObjectFile::staticDtorList;
-tree  ObjectFile::fileContext;
 
 typedef struct {
     tree decl;
@@ -34,8 +33,6 @@ typedef struct {
 
 ObjectFile::ObjectFile()
 {
-    fileContext = build_decl(TRANSLATION_UNIT_DECL, NULL, NULL);
-    dkeep(fileContext);
 }
 
 void
@@ -152,23 +149,76 @@ ObjectFile::setDeclLoc(tree t, Dsymbol * decl)
 }
 
 void
-ObjectFile::makeDeclOneOnly(tree decl_tree)
+ObjectFile::giveDeclUniqueName(tree decl, const char * prefix)
+{
+    /* It would be nice to be able to use TRANSLATION_UNIT_DECL
+       so lhd_set_decl_assembler_name would do this automatically.
+       Unforntuately, the non-NULL decl context confuses dwarf2out.
+
+       Maybe this is fixed in later versions of GCC.
+    */
+
+    const char *name;
+    if (prefix)
+	name = prefix;
+    else if (DECL_NAME(decl))
+	name = IDENTIFIER_POINTER(DECL_NAME(decl));
+    else
+	name = "___s";
+    char *label = d_asm_format_private_name(name, DECL_UID(decl));
+    SET_DECL_ASSEMBLER_NAME(decl, get_identifier(label));
+    free(label);
+}
+
+#if V2
+static bool
+is_function_nested_in_function(Dsymbol * dsym)
+{
+    FuncDeclaration * fd = dsym->isFuncDeclaration();
+    AggregateDeclaration * ad;
+    ClassDeclaration * cd;
+
+    if (! fd)
+	return false;
+    else if (fd->isNested())
+	return true;
+    else if ( (ad = fd->isThis()) && (cd = ad->isClassDeclaration()) )
+    {
+	while (cd && cd->isNested())
+	{
+	    dsym = cd->toParent2();
+	    if (dsym->isFuncDeclaration())
+		return true;
+	    else
+		cd = dsym->isClassDeclaration();
+	}
+    }
+    
+    return false;
+}   
+#endif
+
+void
+ObjectFile::makeDeclOneOnly(tree decl_tree, Dsymbol * dsym)
 {
     /* First method: Use one-only/coalesced attribute.
        
        If user has specified -femit-templates=private, honor that
        even if the target supports one-only. */
     if ( ! D_DECL_IS_TEMPLATE( decl_tree ) || gen.emitTemplates != TEprivate ) {
-	/* Weak definitions have to be public.  Nested functions may not
-	   be emitted as public even if TREE_PUBLIC is set.
+	/* Weak definitions have to be public.  Nested functions may or
+	   may not be emitted as public even if TREE_PUBLIC is set.
 	   There is no way to tell if the back end implements
 	   make_decl_one_only with DECL_WEAK, so this check is
 	   done first.	*/
-	
-	if (TREE_CODE(decl_tree) == FUNCTION_DECL &&
-	    decl_function_context(decl_tree) != NULL_TREE &&
-	    ! DECL_NO_STATIC_CHAIN(decl_tree)) {
-
+	if ( (TREE_CODE(decl_tree) == FUNCTION_DECL &&
+		decl_function_context(decl_tree) != NULL_TREE &&
+		! DECL_NO_STATIC_CHAIN(decl_tree))
+#if V2
+	    || ( dsym && is_function_nested_in_function(dsym) )
+#endif
+	     ) {
+	    
 	    return;
 	}
 	    
@@ -291,7 +341,7 @@ ObjectFile::setupSymbolStorage(Dsymbol * dsym, tree decl_tree, bool force_static
 	}
 
 	if (D_DECL_ONE_ONLY( decl_tree ))
-	    makeDeclOneOnly( decl_tree );
+	    makeDeclOneOnly(decl_tree, dsym);
 
 	if ( func_decl && func_decl->isNested() )
 	    TREE_PUBLIC( decl_tree ) = 0;
@@ -487,10 +537,11 @@ ObjectFile::initTypeDecl(tree t, tree decl)
 void
 ObjectFile::declareType(tree t, tree decl)
 {
-    bool top_level = DECL_CONTEXT(decl) == fileContext || ! DECL_CONTEXT(decl);
+    bool top_level = /*DECL_CONTEXT(decl) == fileContext || */
+	! DECL_CONTEXT(decl);
 	
-    //EXPER//always nyan//if (top_level)
-	rodc(decl, top_level);
+    // okay to do this?
+    rodc(decl, top_level);
 }
 
 tree
@@ -684,7 +735,7 @@ ObjectFile::doSimpleFunction(const char * name, tree expr, bool static_ctor, boo
     func->fbody = new ExpStatement(mod->loc,
 	new WrappedExp(mod->loc, TOKcomma, expr, Type::tvoid));
 
-    func->toObjFile();
+    func->toObjFile(false);
 
     return func;
 }
@@ -809,9 +860,7 @@ check_static_sym(Symbol * sym)
 
 	tree t_ini = dt2tree( sym->Sdt ); // %% recursion problems?
 	tree t_var = build_decl( VAR_DECL, NULL_TREE, TREE_TYPE( t_ini ));
-	char * name = d_asm_format_private_name("___s", DECL_UID( t_var ));
-	SET_DECL_ASSEMBLER_NAME( t_var, get_identifier( name ));
-	free(name);
+	g.ofile->giveDeclUniqueName(t_var);
 	DECL_INITIAL( t_var ) = t_ini;
 	TREE_STATIC( t_var ) = 1;
 	if (sym->Sseg == CDATA) {
@@ -828,17 +877,22 @@ check_static_sym(Symbol * sym)
     return sym->Stree;
 }
 
+static tree
+dt2tree_list_of_elems(dt_t * dt);
 
 static tree
 dt2node(dt_t * dt)
 {
-    tree a;
     switch (dt->dt) {
     case DT_azeros:
     case DT_common:
-	a = make_node(CONSTRUCTOR);
-	TREE_TYPE(a) = gen.arrayType(Type::tuns8, dt->DTint);
-	return a;
+	{
+	    tree a = make_node(CONSTRUCTOR);
+	    TREE_TYPE(a) = gen.arrayType(Type::tuns8, dt->DTint);
+	    TREE_READONLY(a) = 1;
+	    TREE_CONSTANT(a) = 1;
+	    return a;
+	}
     case DT_nbytes:
 	{
 	    tree s = build_string(dt->DTint, (char *) dt->DTpointer);
@@ -863,12 +917,14 @@ dt2node(dt_t * dt)
 	return dt->DTtree;
     case DT_container:
 	{
-	    /* It is necessary to give static array data its
-	       original type.  Otherwise, the SRA pass will
-	       not find the array elements.
+	    /* It is necessary to give static array data its original
+	       type.  Otherwise, the SRA pass will not find the array
+	       elements.
 
-	       SRA accesses struct elements by field offset,
-	       so the ad hoc type from dt2tree is fine. */
+	       SRA accesses struct elements by field offset, so the ad
+	       hoc type from dt2tree is fine.  It must still be a
+	       CONSTRUCTOR, or the CCP pass may use it incorrectly.
+	    */
 	    Type *tb = NULL;
 	    if (dt->DTtype)
 		tb = dt->DTtype->toBasetype();
@@ -887,11 +943,17 @@ dt2node(dt_t * dt)
 		}
 		tree ctor = make_node(CONSTRUCTOR);
 		TREE_TYPE(ctor) = dt->DTtype->toCtype();
+
+		// DT data should always be constant.  If the decl is not TREE_CONSTANT, fine.
+		TREE_CONSTANT(ctor) = 1;
+
 		TREE_READONLY(ctor) = 1;
 		TREE_STATIC(ctor) = 1;
 		CONSTRUCTOR_ELTS(ctor) = ctor_elts.head;
 		return ctor;
 	    }
+	    else if (tb && tb->ty == Tstruct)
+		return dt2tree_list_of_elems(dt->DTvalues);
 	    else
 		return dt2tree(dt->DTvalues);
 	}
@@ -907,47 +969,55 @@ dt2tree(dt_t * dt)
     if (dt && /*dt->dt == DT_container || */dt->DTnext == NULL)
 	return dt2node(dt);
     else
-    {
-	// Generate type on the fly
-	CtorEltMaker elts;
-	ListMaker fields;
-	tree offset = size_zero_node;
+	return dt2tree_list_of_elems(dt);
+}
 
-	tree aggtype = make_node(RECORD_TYPE);
+static tree
+dt2tree_list_of_elems(dt_t * dt)
+{
+    // Generate type on the fly
+    CtorEltMaker elts;
+    ListMaker fields;
+    tree offset = size_zero_node;
 
-	while (dt) {
-	    tree value = dt2node(dt);
-	    tree field = build_decl(FIELD_DECL, NULL_TREE, TREE_TYPE(value));
-	    DECL_CONTEXT(field) = aggtype;
-	    DECL_FIELD_OFFSET(field) = offset;
-	    DECL_FIELD_BIT_OFFSET(field) = bitsize_zero_node;
-	    DECL_ARTIFICIAL(field) = 1;
-	    DECL_IGNORED_P(field) = 1;
-	    layout_decl(field, 0);
+    tree aggtype = make_node(RECORD_TYPE);
 
-	    fields.chain(field);
-	    elts.cons(field, value);
-	    offset = size_binop(PLUS_EXPR, offset, TYPE_SIZE_UNIT( TREE_TYPE( value )));
+    while (dt) {
+	tree value = dt2node(dt);
+	tree field = build_decl(FIELD_DECL, NULL_TREE, TREE_TYPE(value));
+	DECL_CONTEXT(field) = aggtype;
+	DECL_FIELD_OFFSET(field) = offset;
+	DECL_FIELD_BIT_OFFSET(field) = bitsize_zero_node;
+	DECL_ARTIFICIAL(field) = 1;
+	DECL_IGNORED_P(field) = 1;
+	layout_decl(field, 0);
 
-	    dt = dt->DTnext;
-	}
-    
-	TYPE_FIELDS(aggtype) = fields.head; // or finish_laout
-	TYPE_SIZE(aggtype) = convert( bitsizetype,
-	    size_binop( MULT_EXPR, offset, size_int( BITS_PER_UNIT )));
-	TYPE_SIZE_UNIT(aggtype) = offset;
-	// okay no alignment -- decl (which has the correct type) should take care of it..
-	// align=bits per word?
-	compute_record_mode(aggtype);
+	fields.chain(field);
+	elts.cons(field, value);
+	offset = size_binop(PLUS_EXPR, offset, TYPE_SIZE_UNIT( TREE_TYPE( value )));
 
-	tree ctor = make_node(CONSTRUCTOR);
-	TREE_TYPE(ctor) = aggtype;
-	TREE_READONLY(ctor) = 1;
-	// dt created data is always static
-	TREE_STATIC(ctor) = 1;
-	CONSTRUCTOR_ELTS(ctor) = elts.head;
-	return ctor;
+	dt = dt->DTnext;
     }
+
+    TYPE_FIELDS(aggtype) = fields.head; // or finish_laout
+    TYPE_SIZE(aggtype) = convert( bitsizetype,
+	size_binop( MULT_EXPR, offset, size_int( BITS_PER_UNIT )));
+    TYPE_SIZE_UNIT(aggtype) = offset;
+    // okay no alignment -- decl (which has the correct type) should take care of it..
+    // align=bits per word?
+    compute_record_mode(aggtype);
+
+    tree ctor = make_node(CONSTRUCTOR);
+    TREE_TYPE(ctor) = aggtype;
+    TREE_READONLY(ctor) = 1;
+    // dt created data is always static
+    TREE_STATIC(ctor) = 1;
+    // should always be constant
+    TREE_CONSTANT(ctor) = 1;
+
+    // DT data should always be constant.  If the decl is not TREE_CONSTANT, fine.
+    CONSTRUCTOR_ELTS(ctor) = elts.head;
+    return ctor;
 }
 
 void
@@ -990,6 +1060,19 @@ obj_includelib(char *name)
     d_warning(OPT_Wunknown_pragmas, "pragma(lib) not implemented");
 }
 
+void
+obj_startaddress(Symbol *s)
+{
+    d_warning(OPT_Wunknown_pragmas, "pragma(startaddress) not implemented");
+}
+
+void
+obj_append(Dsymbol *s)
+{
+    // GDC does not do multi-obj, so just write it out now.
+    s->toObjFile(false);
+}
+
 void
 obj_moduleinfo(Symbol *sym)
 {
@@ -1013,8 +1096,9 @@ obj_moduleinfo(Symbol *sym)
     tree f0 = TYPE_FIELDS( mod_ref_type );
     tree f1 = TREE_CHAIN( f0 );
 
-    tree our_mod_ref = build_decl(VAR_DECL, get_identifier("__mod_ref"), mod_ref_type);
+    tree our_mod_ref = build_decl(VAR_DECL, NULL_TREE, mod_ref_type);
     dkeep(our_mod_ref);
+    g.ofile->giveDeclUniqueName(our_mod_ref, "__mod_ref");
 
     tree init = make_node(CONSTRUCTOR);
     CtorEltMaker ce;
@@ -1028,7 +1112,6 @@ obj_moduleinfo(Symbol *sym)
     TREE_PRIVATE( our_mod_ref ) = 1;
     TREE_STATIC( our_mod_ref ) = 1;
     DECL_INITIAL( our_mod_ref ) = init;
-    DECL_CONTEXT( our_mod_ref ) = g.ofile->fileContext;
     g.ofile->rodc(our_mod_ref, 1);
 
     tree the_mod_ref = build_decl(VAR_DECL, get_identifier("_Dmodule_ref"),
