@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2006 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -152,6 +152,20 @@ ClassDeclaration::ClassDeclaration(Loc loc, Identifier *id, BaseClasses *basecla
 		    Type::typeinfotypelist->error("%s", msg);
 		Type::typeinfotypelist = this;
 	    }
+
+#if V2
+	    if (id == Id::TypeInfo_Const)
+	    {	if (Type::typeinfoconst)
+		    Type::typeinfoconst->error("%s", msg);
+		Type::typeinfoconst = this;
+	    }
+
+	    if (id == Id::TypeInfo_Invariant)
+	    {	if (Type::typeinfoinvariant)
+		    Type::typeinfoinvariant->error("%s", msg);
+		Type::typeinfoinvariant = this;
+	    }
+#endif
 	}
 
 	if (id == Id::Object)
@@ -174,14 +188,11 @@ ClassDeclaration::ClassDeclaration(Loc loc, Identifier *id, BaseClasses *basecla
     }
 
     com = 0;
-#if 0
-    if (id == Id::IUnknown)		// IUnknown is the root of all COM objects
-	com = 1;
-#endif
     isauto = 0;
     isabstract = 0;
     isnested = 0;
     vthis = NULL;
+    inuse = 0;
 }
 
 Dsymbol *ClassDeclaration::syntaxCopy(Dsymbol *s)
@@ -258,6 +269,11 @@ void ClassDeclaration::semantic(Scope *sc)
     methods.setDim(0);
 #endif
 
+    if (sc->stc & STCdeprecated)
+    {
+	isdeprecated = 1;
+    }
+
     // Expand any tuples in baseclasses[]
     for (i = 0; i < baseclasses.dim; )
     {	BaseClass *b = (BaseClass *)baseclasses.data[i];
@@ -295,6 +311,18 @@ void ClassDeclaration::semantic(Scope *sc)
 	else
 	{
 	    tc = (TypeClass *)(tb);
+
+	    if (tc->sym->isDeprecated())
+	    {
+		if (!isDeprecated())
+		{
+		    // Deriving from deprecated class makes this one deprecated too
+		    isdeprecated = 1;
+
+		    tc->checkDeprecated(loc, sc);
+		}
+	    }
+
 	    if (tc->sym->isInterfaceDeclaration())
 		;
 	    else
@@ -349,6 +377,17 @@ void ClassDeclaration::semantic(Scope *sc)
 	}
 	else
 	{
+	    if (tc->sym->isDeprecated())
+	    {
+		if (!isDeprecated())
+		{
+		    // Deriving from deprecated class makes this one deprecated too
+		    isdeprecated = 1;
+
+		    tc->checkDeprecated(loc, sc);
+		}
+	    }
+
 	    // Check for duplicate interfaces
 	    for (size_t j = (baseClass ? 1 : 0); j < i; j++)
 	    {
@@ -489,8 +528,6 @@ void ClassDeclaration::semantic(Scope *sc)
 	isauto = 1;
     if (storage_class & STCabstract)
 	isabstract = 1;
-    if (storage_class & STCdeprecated)
-	isdeprecated = 1;
 
     sc = sc->push(this);
     sc->stc &= ~(STCfinal | STCauto | STCscope | STCstatic |
@@ -539,7 +576,7 @@ void ClassDeclaration::semantic(Scope *sc)
 	scope->setNoFree();
 	scope->module->addDeferredSemantic(this);
 
-	//printf("\tsemantic('%s') failed\n", toChars());
+	//printf("\tsemantic('%s') failed due to forward references\n", toChars());
 	return;
     }
 
@@ -577,7 +614,7 @@ void ClassDeclaration::semantic(Scope *sc)
 	ctor->fbody = new CompoundStatement(0, new Statements());
 	members->push(ctor);
 	ctor->addMember(sc, this, 1);
-	*sc = scsave;
+	*sc = scsave;	// why? What about sc->nofree?
 	sc->offset = structsize;
 	ctor->semantic(sc);
 	defaultCtor = ctor;
@@ -617,6 +654,7 @@ void ClassDeclaration::semantic(Scope *sc)
     sizeok = 1;
     Module::dprogress++;
 
+    dtor = buildDtor(sc);
 
     sc->pop();
 
@@ -764,6 +802,35 @@ Dsymbol *ClassDeclaration::search(Loc loc, Identifier *ident, int flags)
     return s;
 }
 
+/**********************************************************
+ * fd is in the vtbl[] for this class.
+ * Return 1 if function is hidden (not findable through search).
+ */
+
+#if V2
+int isf(void *param, FuncDeclaration *fd)
+{
+    //printf("param = %p, fd = %p %s\n", param, fd, fd->toChars());
+    return param == fd;
+}
+
+int ClassDeclaration::isFuncHidden(FuncDeclaration *fd)
+{
+    //printf("ClassDeclaration::isFuncHidden(%s)\n", fd->toChars());
+    Dsymbol *s = search(0, fd->ident, 4|2);
+    if (!s)
+    {	//printf("not found\n");
+	/* Because, due to a hack, if there are multiple definitions
+	 * of fd->ident, NULL is returned.
+	 */
+	return 0;
+    }
+    FuncDeclaration *fdstart = s->toAlias()->isFuncDeclaration();
+    //printf("%s fdstart = %p\n", s->kind(), fdstart);
+    return !overloadApply(fdstart, &isf, fd);
+}
+#endif
+
 /****************
  * Find virtual function matching identifier and type.
  * Used to build virtual function tables for interface implementations.
@@ -801,18 +868,17 @@ FuncDeclaration *ClassDeclaration::findFunc(Identifier *ident, TypeFunction *tf)
 }
 
 void ClassDeclaration::interfaceSemantic(Scope *sc)
-{   int i;
-
+{
     vtblInterfaces = new BaseClasses();
     vtblInterfaces->reserve(interfaces_dim);
 
-    for (i = 0; i < interfaces_dim; i++)
+    for (size_t i = 0; i < interfaces_dim; i++)
     {
 	BaseClass *b = interfaces[i];
 
 	// If this is an interface, and it derives from a COM interface,
 	// then this is a COM interface too.
-	if (b->base->isCOMclass())
+	if (b->base->isCOMinterface())
 	    com = 1;
 
 	vtblInterfaces->push(b);
@@ -828,6 +894,11 @@ int ClassDeclaration::isCOMclass()
     return com;
 }
 
+int ClassDeclaration::isCOMinterface()
+{
+    return 0;
+}
+
 
 /****************************************
  */
@@ -850,6 +921,7 @@ int ClassDeclaration::isAbstract()
     return FALSE;
 }
 
+
 /****************************************
  * Returns !=0 if there's an extra member which is the 'this'
  * pointer to the enclosing context (enclosing class or function)
@@ -915,6 +987,8 @@ void InterfaceDeclaration::semantic(Scope *sc)
 {   int i;
 
     //printf("InterfaceDeclaration::semantic(%s), type = %p\n", toChars(), type);
+    if (inuse)
+	return;
     if (!scope)
     {	type = type->semantic(loc, sc);
 	handle = handle->semantic(loc, sc);
@@ -942,6 +1016,11 @@ void InterfaceDeclaration::semantic(Scope *sc)
     else
 	attributes = sc->attributes;
 
+    if (sc->stc & STCdeprecated)
+    {
+	isdeprecated = 1;
+    }
+
     // Expand any tuples in baseclasses[]
     for (i = 0; i < baseclasses.dim; )
     {	BaseClass *b = (BaseClass *)baseclasses.data[0];
@@ -999,7 +1078,7 @@ void InterfaceDeclaration::semantic(Scope *sc)
 		baseclasses.remove(i);
 		continue;
 	    }
-	    if (!b->base->symtab || b->base->scope)
+	    if (!b->base->symtab || b->base->scope || b->base->inuse)
 	    {
 		//error("forward reference of base class %s", baseClass->toChars());
 		// Forward reference of base, try again later
@@ -1060,16 +1139,18 @@ void InterfaceDeclaration::semantic(Scope *sc)
     sc = sc->push(this);
     sc->attributes = NULL;
     sc->parent = this;
-    if (isCOMclass())
+    if (isCOMinterface())
 	sc->linkage = LINKwindows;
     sc->structalign = 8;
     structalign = sc->structalign;
     sc->offset = PTRSIZE * 2;
+    inuse++;
     for (i = 0; i < members->dim; i++)
     {
 	Dsymbol *s = (Dsymbol *)members->data[i];
 	s->semantic(sc);
     }
+    inuse--;
     //members->print();
     sc->pop();
     //printf("-InterfaceDeclaration::semantic(%s), type = %p\n", toChars(), type);
@@ -1150,18 +1231,23 @@ int InterfaceDeclaration::isBaseOf(BaseClass *bc, target_ptrdiff_t *poffset)
 
 /****************************************
  * Determine if slot 0 of the vtbl[] is reserved for something else.
- * For class objects, yes, this is where the classinfo ptr goes.
+ * For class objects, yes, this is where the ClassInfo ptr goes.
  * For COM interfaces, no.
  * For non-COM interfaces, yes, this is where the Interface ptr goes.
  */
 
 int InterfaceDeclaration::vtblOffset()
 {
-    if (isCOMclass())
+    if (isCOMinterface())
 	return 0;
     return 1;
 }
 
+int InterfaceDeclaration::isCOMinterface()
+{
+    return com;
+}
+
 /*******************************************
  */
 
