@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -39,9 +39,14 @@ long __cdecl __ehfilter(LPEXCEPTION_POINTERS ep);
 #include "cond.h"
 #include "expression.h"
 #include "lexer.h"
+#include "lib.h"
 
+void browse(const char *url);
 void getenv_setargv(const char *envvar, int *pargc, char** *pargv);
 
+void obj_start(char *srcfile);
+void obj_end(Library *library, File *objfile);
+
 Global global;
 
 Global::Global()
@@ -64,9 +69,21 @@ Global::Global()
     obj_ext = "o";
 #endif
 
-    copyright = "Copyright (c) 1999-2007 by Digital Mars";
+#ifndef IN_GCC
+#if _WIN32
+    lib_ext  = "lib";
+#elif linux
+    lib_ext  = "a";
+#else
+#error "fix this"
+#endif
+#else
+    lib_ext = "a";
+#endif
+
+    copyright = "Copyright (c) 1999-2008 by Digital Mars";
     written = "written by Walter Bright";
-    version = "v1.021";
+    version = "v1.030";
     global.structalign = 8;
 
     memset(&params, 0, sizeof(Param));
@@ -176,6 +193,8 @@ Usage:\n\
   -debug         compile in debug code\n\
   -debug=level   compile in debug code <= level\n\
   -debug=ident   compile in debug code identified by ident\n\
+  -debuglib=name    set symbolic debug library to name\n\
+  -defaultlib=name  set default library to name\n\
   -g             add symbolic debug info\n\
   -gc            add symbolic debug info, pretend to be C\n\
   -H             generate 'header' file\n\
@@ -183,13 +202,16 @@ Usage:\n\
   -Hffilename    write 'header' file to filename\n\
   --help         print help\n\
   -Ipath         where to look for imports\n\
-  -Jpath         where to look for string imports\n\
+  -ignore        ignore unsupported pragmas\n\
   -inline        do function inlining\n\
+  -Jpath         where to look for string imports\n\
   -Llinkerflag   pass linkerflag to link\n\
+  -lib           generate library rather than object files\n\
+  -man           open web browser on manual page\n\
   -nofloat       do not emit reference to floating point\n\
   -O             optimize\n\
   -o-            do not write object file\n\
-  -odobjdir      write object files to directory objdir\n\
+  -odobjdir      write object & library files to directory objdir\n\
   -offilename	 name output file to filename\n\
   -op            do not strip paths from source file\n\
   -profile	 profile runtime performance of generated code\n\
@@ -215,6 +237,7 @@ int main(int argc, char *argv[])
 {
     int i;
     Array files;
+    Array libmodules;
     char *p;
     Module *m;
     int status = EXIT_SUCCESS;
@@ -252,13 +275,18 @@ int main(int argc, char *argv[])
     global.params.useInline = 0;
     global.params.obj = 1;
     global.params.Dversion = 2;
+    global.params.quiet = 1;
 
     global.params.linkswitches = new Array();
     global.params.libfiles = new Array();
     global.params.objfiles = new Array();
     global.params.ddocfiles = new Array();
 
+#if _WIN32
     global.params.defaultlibname = "phobos";
+#elif linux
+    global.params.defaultlibname = "phobos";
+#endif
     global.params.debuglibname = global.params.defaultlibname;
 
     // Predefine version identifiers
@@ -277,6 +305,9 @@ int main(int argc, char *argv[])
     //VersionCondition::addPredefinedGlobalIdent("D_Bits");
     VersionCondition::addPredefinedGlobalIdent("D_InlineAsm");
     VersionCondition::addPredefinedGlobalIdent("D_InlineAsm_X86");
+#if V2
+    VersionCondition::addPredefinedGlobalIdent("D_Version2");
+#endif
     VersionCondition::addPredefinedGlobalIdent("all");
 
 #if _WIN32
@@ -307,6 +338,8 @@ int main(int argc, char *argv[])
 		global.params.cov = 1;
 	    else if (strcmp(p + 1, "fPIC") == 0)
 		global.params.pic = 1;
+	    else if (strcmp(p + 1, "multiobj") == 0)
+		global.params.multiobj = 1;
 	    else if (strcmp(p + 1, "g") == 0)
 		global.params.symdebug = 1;
 	    else if (strcmp(p + 1, "gc") == 0)
@@ -320,7 +353,14 @@ int main(int argc, char *argv[])
 	    else if (strcmp(p + 1, "v") == 0)
 		global.params.verbose = 1;
 	    else if (strcmp(p + 1, "v1") == 0)
+	    {
+#if V1
 		global.params.Dversion = 1;
+#else
+		error("use DMD 1.0 series compilers for -v1 switch");
+		break;
+#endif
+            }
 	    else if (strcmp(p + 1, "w") == 0)
 		global.params.warnings = 1;
 	    else if (strcmp(p + 1, "O") == 0)
@@ -406,8 +446,12 @@ int main(int argc, char *argv[])
 		}
 	    }
 #endif
+	    else if (strcmp(p + 1, "ignore") == 0)
+		global.params.ignoreUnsupportedPragmas = 1;
 	    else if (strcmp(p + 1, "inline") == 0)
 		global.params.useInline = 1;
+	    else if (strcmp(p + 1, "lib") == 0)
+		global.params.lib = 1;
 	    else if (strcmp(p + 1, "nofloat") == 0)
 		global.params.nofloat = 1;
 	    else if (strcmp(p + 1, "quiet") == 0)
@@ -507,6 +551,24 @@ int main(int argc, char *argv[])
 	    {
 		global.params.debuglibname = p + 1 + 9;
 	    }
+	    else if (memcmp(p + 1, "man", 3) == 0)
+	    {
+#if _WIN32
+#if V1
+		browse("http://www.digitalmars.com/d/1.0/dmd-windows.html");
+#else
+		browse("http://www.digitalmars.com/d/2.0/dmd-windows.html");
+#endif
+#endif
+#if linux
+#if V1
+		browse("http://www.digitalmars.com/d/1.0/dmd-linux.html");
+#else
+		browse("http://www.digitalmars.com/d/2.0/dmd-linux.html");
+#endif
+#endif
+		exit(EXIT_SUCCESS);
+	    }
 	    else if (strcmp(p + 1, "run") == 0)
 	    {	global.params.run = 1;
 		global.params.runargs_length = ((i >= argcstart) ? argc : argcstart) - i - 1;
@@ -534,7 +596,17 @@ int main(int argc, char *argv[])
 	    }
 	}
 	else
+	{
+#if !TARGET_LINUX
+	    char *ext = FileName::ext(p);
+	    if (ext && stricmp(ext, "exe") == 0)
+	    {
+		global.params.objname = p;
+		continue;
+	    }
+#endif
 	    files.push(p);
+	}
     }
     if (global.errors)
     {
@@ -560,13 +632,37 @@ int main(int argc, char *argv[])
     if (global.params.useUnitTests)
 	global.params.useAssert = 1;
 
-    if (!global.params.obj)
+    if (!global.params.obj || global.params.lib)
 	global.params.link = 0;
 
     if (global.params.link)
     {
 	global.params.exefile = global.params.objname;
+	global.params.oneobj = 1;
+	if (global.params.objname)
+	{
+	    /* Use this to name the one object file with the same
+	     * name as the exe file.
+	     */
+	    global.params.objname = FileName::forceExt(global.params.objname, global.obj_ext)->toChars();
+
+	    /* If output directory is given, use that path rather than
+	     * the exe file path.
+	     */
+	    if (global.params.objdir)
+	    {	char *name = FileName::name(global.params.objname);
+		global.params.objname = FileName::combine(global.params.objdir, name);
+	    }
+	}
+    }
+    else if (global.params.lib)
+    {
+	global.params.libname = global.params.objname;
 	global.params.objname = NULL;
+
+	// Haven't investigated handling these options with multiobj
+	if (!global.params.cov && !global.params.trace)
+	    global.params.multiobj = 1;
     }
     else if (global.params.run)
     {
@@ -577,12 +673,17 @@ int main(int argc, char *argv[])
     {
 	if (global.params.objname && files.dim > 1)
 	{
-	    error("multiple source files, but only one .obj name");
-	    fatal();
+	    global.params.oneobj = 1;
+	    //error("multiple source files, but only one .obj name");
+	    //fatal();
 	}
     }
     if (global.params.cov)
 	VersionCondition::addPredefinedGlobalIdent("D_Coverage");
+#if V2
+    if (global.params.useUnitTests)
+	VersionCondition::addPredefinedGlobalIdent("unittest");
+#endif
 
     // Initialization
     Type::init();
@@ -631,8 +732,9 @@ int main(int argc, char *argv[])
     // Create Modules
     Array modules;
     modules.reserve(files.dim);
+    int firstmodule = 1;
     for (i = 0; i < files.dim; i++)
-    {	Identifier *id;
+    {
 	char *ext;
 	char *name;
 
@@ -650,7 +752,8 @@ int main(int argc, char *argv[])
 	p = FileName::name(p);		// strip path
 	ext = FileName::ext(p);
 	if (ext)
-	{
+	{   /* Deduce what to do with a file based on its extension
+	     */
 #if TARGET_LINUX
 	    if (strcmp(ext, global.obj_ext) == 0)
 #else
@@ -658,16 +761,18 @@ int main(int argc, char *argv[])
 #endif
 	    {
 		global.params.objfiles->push(files.data[i]);
+		libmodules.push(files.data[i]);
 		continue;
 	    }
 
 #if TARGET_LINUX
-	    if (strcmp(ext, "a") == 0)
+	    if (strcmp(ext, global.lib_ext) == 0)
 #else
-	    if (stricmp(ext, "lib") == 0)
+	    if (stricmp(ext, global.lib_ext) == 0)
 #endif
 	    {
 		global.params.libfiles->push(files.data[i]);
+		libmodules.push(files.data[i]);
 		continue;
 	    }
 
@@ -692,12 +797,15 @@ int main(int argc, char *argv[])
 
 	    if (stricmp(ext, "exe") == 0)
 	    {
-		global.params.exefile = (char *)files.data[i];
-		continue;
+		assert(0);	// should have already been handled
 	    }
 #endif
 
+	    /* Examine extension to see if it is a valid
+	     * D source file extension
+	     */
 	    if (stricmp(ext, global.mars_ext) == 0 ||
+		stricmp(ext, "dd") == 0 ||
 		stricmp(ext, "htm") == 0 ||
 		stricmp(ext, "html") == 0 ||
 		stricmp(ext, "xhtml") == 0)
@@ -728,11 +836,18 @@ int main(int argc, char *argv[])
 		goto Linvalid;
 	}
 
-	id = new Identifier(name, 0);
+	/* At this point, name is the D source file name stripped of
+	 * its path and extension.
+	 */
+
+	Identifier *id = new Identifier(name, 0);
 	m = new Module((char *) files.data[i], id, global.params.doDocComments, global.params.doHdrGeneration);
 	modules.push(m);
 
-	global.params.objfiles->push(m->objfile->name->str);
+	if (firstmodule)
+	{   global.params.objfiles->push(m->objfile->name->str);
+	    firstmodule = 0;
+	}
     }
 
 #if _WIN32
@@ -740,6 +855,7 @@ int main(int argc, char *argv[])
   {
 #endif
     // Read files, parse them
+    int anydocfiles = 0;
     for (i = 0; i < modules.dim; i++)
     {
 	m = (Module *)modules.data[i];
@@ -748,11 +864,13 @@ int main(int argc, char *argv[])
 	if (!Module::rootModule)
 	    Module::rootModule = m;
 	m->importedFrom = m;
-	m->deleteObjFile();
+	if (!global.params.oneobj || i == 0 || m->isDocFile)
+	    m->deleteObjFile();
 	m->read(0);
 	m->parse();
 	if (m->isDocFile)
 	{
+	    anydocfiles = 1;
 	    m->gendocfile();
 
 	    // Remove m from list of modules
@@ -773,6 +891,12 @@ int main(int argc, char *argv[])
 		global.params.link = 0;
 	}
     }
+    if (anydocfiles && modules.dim &&
+	(global.params.oneobj || global.params.objname))
+    {
+	error("conflicting Ddoc and obj generation options");
+	fatal();
+    }
     if (global.errors)
 	fatal();
 #ifdef _DH
@@ -862,22 +986,68 @@ int main(int argc, char *argv[])
     if (global.errors)
 	fatal();
 
-    // Generate output files
-    for (i = 0; i < modules.dim; i++)
+    Library *library = NULL;
+    if (global.params.lib)
     {
-	m = (Module *)modules.data[i];
-	if (global.params.verbose)
-	    printf("code      %s\n", m->toChars());
-	if (global.params.obj)
-	    m->genobjfile();
-	if (global.errors)
-	    m->deleteObjFile();
-	else
+	library = new Library();
+	library->setFilename(global.params.objdir, global.params.libname);
+
+	// Add input object and input library files to output library
+	for (int i = 0; i < libmodules.dim; i++)
 	{
-	    if (global.params.doDocComments)
-		m->gendocfile();
+	    char *p = (char *)libmodules.data[i];
+	    library->addObject(p, NULL, 0);
 	}
     }
+
+    // Generate output files
+    if (global.params.oneobj)
+    {
+	for (i = 0; i < modules.dim; i++)
+	{
+	    m = (Module *)modules.data[i];
+	    if (global.params.verbose)
+		printf("code      %s\n", m->toChars());
+	    if (i == 0)
+		obj_start(m->srcfile->toChars());
+	    m->genobjfile(0);
+	    if (!global.errors && global.params.doDocComments)
+		m->gendocfile();
+	}
+	if (!global.errors && modules.dim)
+	{
+	    obj_end(library, ((Module *)modules.data[0])->objfile);
+	}
+    }
+    else
+    {
+	for (i = 0; i < modules.dim; i++)
+	{
+	    m = (Module *)modules.data[i];
+	    if (global.params.verbose)
+		printf("code      %s\n", m->toChars());
+	    if (global.params.obj)
+	    {   obj_start(m->srcfile->toChars());
+		m->genobjfile(global.params.multiobj);
+		obj_end(library, m->objfile);
+		obj_write_deferred(library);
+	    }
+	    if (global.errors)
+	    {
+		if (!global.params.lib)
+		    m->deleteObjFile();
+	    }
+	    else
+	    {
+		if (global.params.doDocComments)
+		    m->gendocfile();
+	    }
+	}
+    }
+
+    if (global.params.lib && !global.errors)
+	library->write();
+
 #if _WIN32
   }
   __except (__ehfilter(GetExceptionInformation()))
@@ -910,8 +1080,10 @@ int main(int argc, char *argv[])
 		 */
 		for (i = 0; i < modules.dim; i++)
 		{
-		    m = (Module *)modules.data[i];
+		    Module *m = (Module *)modules.data[i];
 		    m->deleteObjFile();
+		    if (global.params.oneobj)
+			break;
 		}
 		deleteExeFile();
 	    }
