@@ -85,6 +85,7 @@ Module::Module(char *filename, Identifier *ident, int doDocComment, int doHdrGen
     vmoduleinfo = NULL;
     massert = NULL;
     marray = NULL;
+    sictor = NULL;
     sctor = NULL;
     sdtor = NULL;
     stest = NULL;
@@ -103,11 +104,13 @@ Module::Module(char *filename, Identifier *ident, int doDocComment, int doHdrGen
 
     macrotable = NULL;
     escapetable = NULL;
+    doppelganger = 0;
     cov = NULL;
     covb = NULL;
 
     srcfilename = FileName::defaultExt(filename, global.mars_ext);
-    if (!srcfilename->equalsExt(global.mars_ext))
+    if (!srcfilename->equalsExt(global.mars_ext) &&
+        !srcfilename->equalsExt("dd"))
     {
 	if (srcfilename->equalsExt("html") ||
 	    srcfilename->equalsExt("htm")  ||
@@ -780,18 +783,15 @@ void Module::gensymfile()
 {
     OutBuffer buf;
     HdrGenState hgs;
-    int i;
 
     //printf("Module::gensymfile()\n");
 
     buf.printf("// Sym file generated from '%s'", srcfile->toChars());
     buf.writenl();
 
-    for (i = 0; i < members->dim; i++)
-    {
-	Dsymbol *s;
+    for (int i = 0; i < members->dim; i++)
+    {	Dsymbol *s = (Dsymbol *)members->data[i];
 
-	s = (Dsymbol *)members->data[i];
 	s->toCBuffer(&buf, &hgs);
     }
 
