@@ -739,7 +739,7 @@ static void templateResolve(Match *m, TemplateDeclaration *td, Scope *sc, Loc lo
     FuncDeclaration *fd;
 
     assert(td);
-    fd = td->deduce(sc, loc, targsi, arguments);
+    fd = td->deduceFunctionTemplate(sc, loc, targsi, arguments);
     if (!fd)
 	return;
     m->anyf = fd;
