@@ -528,7 +528,7 @@ int CompoundStatement::usesEH()
 int CompoundStatement::fallOffEnd()
 {   int falloff = TRUE;
 
-    //printf("CompoundStatement::fallOffEnd()\n");
+    //printf("CompoundStatement::fallOffEnd() %s\n", toChars());
     for (int i = 0; i < statements->dim; i++)
     {	Statement *s = (Statement *)statements->data[i];
 
@@ -648,25 +648,18 @@ int UnrolledLoopStatement::usesEH()
 }
 
 int UnrolledLoopStatement::fallOffEnd()
-{   int falloff = TRUE;
-
+{
     //printf("UnrolledLoopStatement::fallOffEnd()\n");
     for (size_t i = 0; i < statements->dim; i++)
     {	Statement *s = (Statement *)statements->data[i];
 
-	if (!s)
-	    continue;
-
-	if (!falloff && global.params.warnings && !s->comeFrom())
-	{
-	    fprintf(stdmsg, "warning - ");
-	    s->error("statement is not reachable");
-	}
-	falloff = s->fallOffEnd();
+	if (s)
+	    s->fallOffEnd();
     }
-    return falloff;
+    return TRUE;
 }
 
+
 int UnrolledLoopStatement::comeFrom()
 {   int comefrom = FALSE;
 
@@ -1998,6 +1991,7 @@ Statement *SwitchStatement::semantic(Scope *sc)
     {	condition = condition->integralPromotions(sc);
 	condition->checkIntegral();
     }
+    condition = condition->optimize(WANTvalue);
 
     sc = sc->push();
     sc->sbreak = this;
@@ -3222,7 +3216,7 @@ int TryFinallyStatement::usesEH()
 int TryFinallyStatement::fallOffEnd()
 {   int result;
 
-    result = body->fallOffEnd();
+    result = body ? body->fallOffEnd() : TRUE;
 //    if (finalbody)
 //	result = finalbody->fallOffEnd();
     return result;
