@@ -1,5 +1,5 @@
 
-// Copyright (c) 1999-2004 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -98,9 +98,11 @@ static void find_module_deps(Module * the_module, Array * out_deps)
 
 void Module::genmoduleinfo()
 {
+    //printf("Module::genmoduleinfo() %s\n", toChars());
+
     Symbol *msym = toSymbol();
     unsigned offset;
-    unsigned sizeof_ModuleInfo = 12 * PTRSIZE;
+    unsigned sizeof_ModuleInfo = 14 * PTRSIZE;
 
     //////////////////////////////////////////////
 
@@ -118,6 +120,8 @@ void Module::genmoduleinfo()
 	    void *ctor;
 	    void *dtor;
 	    void *unitTest;
+	    const(MemberInfo[]) function(string) xgetMembers;	// module getMembers() function
+	    void *ictor;
        }
      */
     dt_t *dt = NULL;
@@ -125,7 +129,9 @@ void Module::genmoduleinfo()
     if (moduleinfo)
 	dtxoff(&dt, moduleinfo->toVtblSymbol(), 0, TYnptr); // vtbl for ModuleInfo
     else
+    {	//printf("moduleinfo is null\n");
 	dtdword(&dt, 0);		// BUG: should be an assert()
+    }
     dtdword(&dt, 0);			// monitor
 
     // name[]
@@ -135,14 +141,11 @@ void Module::genmoduleinfo()
     dtabytes(&dt, TYnptr, 0, namelen + 1, name);
 
     ClassDeclarations aclasses;
-    int i;
 
     //printf("members->dim = %d\n", members->dim);
-    for (i = 0; i < members->dim; i++)
-    {
-	Dsymbol *member;
+    for (int i = 0; i < members->dim; i++)
+    {	Dsymbol *member = (Dsymbol *)members->data[i];
 
-	member = (Dsymbol *)members->data[i];
 	//printf("\tmember '%s'\n", member->toChars());
 	member->addLocalClass(&aclasses);
     }
@@ -161,7 +164,7 @@ void Module::genmoduleinfo()
     else
 #endif
 #endif
-    for (i = 0; i < aimports.dim; i++)
+    for (int i = 0; i < aimports.dim; i++)
     {	Module *m = (Module *)aimports.data[i];
 	if (!m->needModuleInfo())
 	    aimports_dim--;
@@ -180,9 +183,9 @@ void Module::genmoduleinfo()
 	dtdword(&dt, 0);
 
     if (needmoduleinfo)
-	dti32(&dt, 0, true);		// flags (4 means MIstandalone)
+	dti32(&dt, 8|0, true);		// flags (4 means MIstandalone)
     else
-	dti32(&dt, 4, true);		// flags (4 means MIstandalone)
+	dti32(&dt, 8|4, true);		// flags (4 means MIstandalone)
 
     if (sctor)
 	dtxoff(&dt, sctor, 0, TYnptr);
@@ -199,6 +202,13 @@ void Module::genmoduleinfo()
     else
 	dtdword(&dt, 0);
 
+    dtdword(&dt, 0);			// xgetMembers
+
+    if (sictor)
+	dtxoff(&dt, sictor, 0, TYnptr);
+    else
+	dtdword(&dt, 0);
+
     //////////////////////////////////////////////
 
 #ifdef IN_GCC
@@ -217,11 +227,9 @@ void Module::genmoduleinfo()
     else
 #endif
 #endif
-    for (i = 0; i < aimports.dim; i++)
-    {
-	Module *m;
+   for (int i = 0; i < aimports.dim; i++)
+    {	Module *m = (Module *)aimports.data[i];
 
-	m = (Module *)aimports.data[i];
 	if (m->needModuleInfo())
 	{   Symbol *s = m->toSymbol();
 	    s->Sflags |= SFLweak;
@@ -229,11 +237,9 @@ void Module::genmoduleinfo()
 	}
     }
 
-    for (i = 0; i < aclasses.dim; i++)
+    for (int i = 0; i < aclasses.dim; i++)
     {
-	ClassDeclaration *cd;
-
-	cd = (ClassDeclaration *)aclasses.data[i];
+	ClassDeclaration *cd = (ClassDeclaration *)aclasses.data[i];
 	dtxoff(&dt, cd->toSymbol(), 0, TYnptr);
     }
 
@@ -250,7 +256,8 @@ void Module::genmoduleinfo()
 }
 
 /* ================================================================== */
-void Dsymbol::toObjFile()
+
+void Dsymbol::toObjFile(int multiobj)
 {
     //printf("Dsymbol::toObjFile('%s')\n", toChars());
     // ignore
@@ -270,7 +277,7 @@ void Dsymbol::toObjFile()
 		ds = ((DsymbolExp *) o)->s;
 		d = ds->isDeclaration();
 		if (d)
-		    d->toObjFile();
+		    d->toObjFile(multiobj);
 	    }
 		
 	}		
@@ -279,11 +286,10 @@ void Dsymbol::toObjFile()
 }
 /* ================================================================== */
 
-void ClassDeclaration::toObjFile()
+void ClassDeclaration::toObjFile(int multiobj)
 {   unsigned i;
     unsigned offset;
     Symbol *sinit;
-    Symbol *sdtor;
     enum_SC scclass;
 
     //printf("ClassDeclaration::toObjFile('%s')\n", toChars());
@@ -307,11 +313,12 @@ void ClassDeclaration::toObjFile()
 	Dsymbol *member;
 
 	member = (Dsymbol *)members->data[i];
-	member->toObjFile();
+	member->toObjFile(0);
     }
 
+#if 0
     // Build destructor by aggregating dtors[]
-#ifndef IN_GCC
+    Symbol *sdtor;
     switch (dtors.dim)
     {	case 0:
 	    // No destructors for this class
@@ -368,8 +375,6 @@ void ClassDeclaration::toObjFile()
 	    writefunc(sdtor);
 	}
     }
-#else
-    sdtor = d_gcc_aggregate_dtors(this);
 #endif
 
     // Generate C symbols
@@ -458,8 +463,8 @@ void ClassDeclaration::toObjFile()
 	dtdword(&dt, 0);
 
     // destructor
-    if (sdtor)
-	dtxoff(&dt, sdtor, 0, TYnptr);
+    if (dtor)
+	dtxoff(&dt, dtor->toSymbol(), 0, TYnptr);
     else
 	dtdword(&dt, 0);
 
@@ -470,7 +475,7 @@ void ClassDeclaration::toObjFile()
 	dtdword(&dt, 0);
 
     // flags
-    int flags = 4 | com;
+    int flags = 4 | isCOMclass();
     if (ctor)
 	flags |= 8;
     for (ClassDeclaration *cd = this; cd; cd = cd->baseClass)
@@ -790,7 +795,7 @@ unsigned ClassDeclaration::baseVtblOffset(BaseClass *bc)
 
 /* ================================================================== */
 
-void InterfaceDeclaration::toObjFile()
+void InterfaceDeclaration::toObjFile(int multiobj)
 {   unsigned i;
     unsigned offset;
     Symbol *sinit;
@@ -816,7 +821,7 @@ void InterfaceDeclaration::toObjFile()
 
 	member = (Dsymbol *)members->data[i];
 	if (!member->isFuncDeclaration())
-	    member->toObjFile();
+	    member->toObjFile(0);
     }
 
     // Generate C symbols
@@ -890,7 +895,7 @@ void InterfaceDeclaration::toObjFile()
     dtdword(&dt, 0);
 
     // flags
-    dti32(&dt, 4 | com, true);
+    dti32(&dt, 4 | isCOMinterface(), true);
 
     // deallocator
     dtdword(&dt, 0);
@@ -934,7 +939,7 @@ void InterfaceDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void StructDeclaration::toObjFile()
+void StructDeclaration::toObjFile(int multiobj)
 {   unsigned i;
 
     //printf("StructDeclaration::toObjFile('%s')\n", toChars());
@@ -994,23 +999,24 @@ void StructDeclaration::toObjFile()
 	    Dsymbol *member;
 
 	    member = (Dsymbol *)members->data[i];
-	    member->toObjFile();
+	    member->toObjFile(0);
 	}
     }
 }
 
 /* ================================================================== */
 
-void VarDeclaration::toObjFile()
+void VarDeclaration::toObjFile(int multiobj)
 {
     Symbol *s;
     unsigned sz;
     Dsymbol *parent;
 
     //printf("VarDeclaration::toObjFile(%p '%s' type=%s) protection %d\n", this, toChars(), type->toChars(), protection);
+    //printf("\talign = %d\n", type->alignsize());
 
     if (aliassym)
-    {	toAlias()->toObjFile();
+    {	toAlias()->toObjFile(0);
 	return;
     }
 
@@ -1025,7 +1031,9 @@ void VarDeclaration::toObjFile()
 	 */
 	if (/*protection == PROTprivate ||*/
 	    !parent || parent->ident == NULL || parent->isFuncDeclaration())
+	{
 	    s->Sclass = SCstatic;
+	}
 	else
 #endif
 	{
@@ -1068,57 +1076,18 @@ void VarDeclaration::toObjFile()
 
 	    tb = type->toBasetype();
 	    if (tb->ty == Tsarray && ie &&
-		!tb->next->equals(ie->exp->type->toBasetype()->next) &&
-		ie->exp->implicitConvTo(tb->next)
+		!tb->nextOf()->equals(ie->exp->type->toBasetype()->nextOf()) &&
+		ie->exp->implicitConvTo(tb->nextOf())
 		)
 	    {
 		int dim;
 
 		dim = ((TypeSArray *)tb)->dim->toInteger();
 
-		if (tb->next->toBasetype()->ty == Tbit)
-		{   integer_t value;
-
-#ifndef IN_GCC
-		    value = ie->exp->toInteger();
-		    value = (value & 1) ? ~(integer_t)0 : (integer_t)0;
-		    if (value == 0)
-		    {
-			dtnzeros(&s->Sdt, ((unsigned)dim + 31) / 32 * PTRSIZE);
-		    }
-		    else
-		    {
-			while (dim >= 32)
-			{
-			    dtnbytes(&s->Sdt, 4, (char *)&value);
-			    dim -= 32;
-			}
-			if (dim)
-			{
-			    value = (1 << dim) - 1;
-			    dtnbytes(&s->Sdt, 4, (char *)&value);
-			}
-		    }
-#else
-
-		    s->Sdt = NULL;
-		    if (ie->exp->toInteger()) {
-			Bits databits;
-			databits.resize(dim);
-			databits.set();
-			dtnbits(& s->Sdt, databits.allocdim * sizeof(databits.data[0]), (char *)databits.data, sizeof(databits.data[0]));
-		    } else {
-			dtnzeros(&s->Sdt, ((unsigned)dim + 31) / 32 * 4);
-		    }
-#endif
-		}
-		else
+		// Duplicate Sdt 'dim-1' times, as we already have the first one
+		while (--dim > 0)
 		{
-		    // Duplicate Sdt 'dim-1' times, as we already have the first one
-		    while (--dim > 0)
-		    {
-			ie->exp->toDt(&s->Sdt);
-		    }
+		    ie->exp->toDt(&s->Sdt);
 		}
 	    }
 	}
@@ -1172,7 +1141,7 @@ void VarDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void TypedefDeclaration::toObjFile()
+void TypedefDeclaration::toObjFile(int multiobj)
 {
     //printf("TypedefDeclaration::toObjFile('%s')\n", toChars());
 
@@ -1211,7 +1180,7 @@ void TypedefDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void EnumDeclaration::toObjFile()
+void EnumDeclaration::toObjFile(int multiobj)
 {
     //printf("EnumDeclaration::toObjFile('%s')\n", toChars());
 
@@ -1242,7 +1211,12 @@ void EnumDeclaration::toObjFile()
 #if ELFOBJ // Burton
 	sinit->Sseg = CDATA;
 #endif /* ELFOBJ */
+#ifndef IN_GCC
 	dtnbytes(&sinit->Sdt, tc->size(0), (char *)&tc->sym->defaultval);
+#else
+	Expression * e = new IntegerExp(0, tc->sym->defaultval, tc);
+	e->toDt(& sinit->Sdt);
+#endif
 	//sinit->Sdt = tc->sym->init->toDt();
 	outdata(sinit);
     }
