@@ -71,7 +71,9 @@ Expression *Type::getInternalTypeInfo(Scope *sc)
     switch (t->ty)
     {
 	case Tsarray:
+#if 0
 	    t = t->next->arrayOf();	// convert to corresponding dynamic array type
+#endif
 	    break;
 
 	case Tclass:
@@ -133,7 +135,7 @@ Expression *Type::getTypeInfo(Scope *sc)
 	    }
 	    else			// if in obj generation pass
 	    {
-		t->vtinfo->toObjFile();
+		t->vtinfo->toObjFile(global.params.multiobj);
 	    }
 	}
     }
@@ -588,7 +590,7 @@ void TypeInfoTupleDeclaration::toDt(dt_t **pdt)
     dtxoff(pdt, s, 0, TYnptr);		    // elements.ptr
 }
 
-void TypeInfoDeclaration::toObjFile()
+void TypeInfoDeclaration::toObjFile(int multiobj)
 {
     Symbol *s;
     unsigned sz;
@@ -596,6 +598,12 @@ void TypeInfoDeclaration::toObjFile()
 
     //printf("TypeInfoDeclaration::toObjFile(%p '%s') protection %d\n", this, toChars(), protection);
 
+    if (multiobj)
+    {
+	obj_append(this);
+	return;
+    }
+
     s = toSymbol();
     sz = type->size();
 
