@@ -1,5 +1,5 @@
 
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -31,10 +31,10 @@
 
 Expression *Expression::implicitCastTo(Scope *sc, Type *t)
 {
-    //printf("Expression::implicitCastTo(%s) => %s\n", type->toChars(), t->toChars());
-    //printf("%s\n", toChars());
+    //printf("Expression::implicitCastTo(%s of type %s) => %s\n", toChars(), type->toChars(), t->toChars());
 
-    if (implicitConvTo(t))
+    MATCH match = implicitConvTo(t);
+    if (match)
     {
 	if (global.params.warnings &&
 	    Type::impcnvWarn[type->toBasetype()->ty][t->toBasetype()->ty] &&
@@ -49,6 +49,12 @@ Expression *Expression::implicitCastTo(Scope *sc, Type *t)
 	    error("implicit conversion of expression (%s) of type %s to %s can cause loss of data",
 		toChars(), type->toChars(), t->toChars());
 	}
+	if (match == MATCHconst && t == type->constOf())
+	{
+	    Expression *e = copy();
+	    e->type = t;
+	    return e;
+	}
 	return castTo(sc, t);
     }
 
@@ -57,6 +63,7 @@ Expression *Expression::implicitCastTo(Scope *sc, Type *t)
 	return e->implicitCastTo(sc, t);
 
 #if 0
+printf("ty = %d\n", type->ty);
 print();
 type->print();
 printf("to:\n");
@@ -104,7 +111,7 @@ MATCH Expression::implicitConvTo(Type *t)
     if (e->type == t)
 	return MATCHexact;
     if (e != this)
-    {	//printf("optimized to %s\n", e->toChars());
+    {	//printf("\toptimized to %s of type %s\n", e->toChars(), e->type->toChars());
 	return e->implicitConvTo(t);
     }
     MATCH match = type->implicitConvTo(t);
@@ -138,13 +145,14 @@ MATCH IntegerExp::implicitConvTo(Type *t)
     printf("IntegerExp::implicitConvTo(this=%s, type=%s, t=%s)\n",
 	toChars(), type->toChars(), t->toChars());
 #endif
-    if (type->equals(t))
-	return MATCHexact;
+    MATCH m = type->implicitConvTo(t);
+    if (m >= MATCHconst)
+	return m;
 
     TY ty = type->toBasetype()->ty;
     TY toty = t->toBasetype()->ty;
 
-    if (type->implicitConvTo(t) == MATCHnomatch && t->ty == Tenum)
+    if (m == MATCHnomatch && t->ty == Tenum)
 	goto Lno;
 
     switch (ty)
@@ -363,6 +371,14 @@ MATCH NullExp::implicitConvTo(Type *t)
 #endif
     if (this->type->equals(t))
 	return MATCHexact;
+
+    /* Allow implicit conversions from invariant to mutable|const,
+     * and mutable to invariant. It works because, after all, a null
+     * doesn't actually point to anything.
+     */
+    if (t->invariantOf()->equals(type->invariantOf()))
+	return MATCHconst;
+
     // NULL implicitly converts to any pointer type or dynamic array
     if (type->ty == Tpointer && type->nextOf()->ty == Tvoid)
     {
@@ -376,6 +392,37 @@ MATCH NullExp::implicitConvTo(Type *t)
     return Expression::implicitConvTo(t);
 }
 
+MATCH StructLiteralExp::implicitConvTo(Type *t)
+{
+#if 0
+    printf("StructLiteralExp::implicitConvTo(this=%s, type=%s, t=%s)\n",
+	toChars(), type->toChars(), t->toChars());
+#endif
+    MATCH m = Expression::implicitConvTo(t);
+    if (m != MATCHnomatch)
+	return m;
+    if (type->ty == t->ty && type->ty == Tstruct &&
+	((TypeStruct *)type)->sym == ((TypeStruct *)t)->sym)
+    {
+	m = MATCHconst;
+	for (int i = 0; i < elements->dim; i++)
+	{   Expression *e = (Expression *)elements->data[i];
+	    Type *te = e->type;
+	    if (t->mod == 0)
+		te = te->mutableOf();
+	    else
+	    {	assert(t->mod == MODinvariant);
+		te = te->invariantOf();
+	    }
+	    MATCH m2 = e->implicitConvTo(te);
+	    //printf("\t%s => %s, match = %d\n", e->toChars(), te->toChars(), m2);
+	    if (m2 < m)
+		m = m2;
+	}
+    }
+    return m;
+}
+
 MATCH StringExp::implicitConvTo(Type *t)
 {   MATCH m;
 
@@ -514,10 +561,30 @@ MATCH AddrExp::implicitConvTo(Type *t)
     if (result == MATCHnomatch)
     {
 	// Look for pointers to functions where the functions are overloaded.
-	VarExp *ve;
-	FuncDeclaration *f;
 
 	t = t->toBasetype();
+
+	if (e1->op == TOKoverloadset &&
+	    (t->ty == Tpointer || t->ty == Tdelegate) && t->nextOf()->ty == Tfunction)
+	{   OverExp *eo = (OverExp *)e1;
+	    FuncDeclaration *f = NULL;
+	    for (int i = 0; i < eo->vars->a.dim; i++)
+	    {   Dsymbol *s = (Dsymbol *)eo->vars->a.data[i];
+		FuncDeclaration *f2 = s->isFuncDeclaration();
+		assert(f2);
+		if (f2->overloadExactMatch(t->nextOf()))
+		{   if (f)
+			/* Error if match in more than one overload set,
+			 * even if one is a 'better' match than the other.
+			 */
+			ScopeDsymbol::multiplyDefined(loc, f, f2);
+		    else
+			f = f2;
+		    result = MATCHexact;
+		}
+	    }
+	}
+
 	if (type->ty == Tpointer && type->nextOf()->ty == Tfunction &&
 	    t->ty == Tpointer && t->nextOf()->ty == Tfunction &&
 	    e1->op == TOKvar)
@@ -527,8 +594,8 @@ MATCH AddrExp::implicitConvTo(Type *t)
 	     * converted to a SymOffExp.
 	     */
 	    assert(0);
-	    ve = (VarExp *)e1;
-	    f = ve->var->isFuncDeclaration();
+	    VarExp *ve = (VarExp *)e1;
+	    FuncDeclaration *f = ve->var->isFuncDeclaration();
 	    if (f && f->overloadExactMatch(t->nextOf()))
 		result = MATCHexact;
 	}
@@ -563,7 +630,9 @@ MATCH SymOffExp::implicitConvTo(Type *t)
 		if (f)
 		{   if ((t->ty == Tdelegate && (f->needThis() || f->isNested())) ||
 			(t->ty == Tpointer && !(f->needThis() || f->isNested())))
+		    {
 			result = MATCHexact;
+		    }
 		}
 	    }
 	}
@@ -618,9 +687,7 @@ MATCH CondExp::implicitConvTo(Type *t)
  */
 
 Expression *Expression::castTo(Scope *sc, Type *t)
-{   Expression *e;
-    Type *tb;
-
+{
     //printf("Expression::castTo(this=%s, t=%s)\n", toChars(), t->toChars());
 #if 0
     printf("Expression::castTo(this=%s, type=%s, t=%s)\n",
@@ -628,19 +695,19 @@ Expression *Expression::castTo(Scope *sc, Type *t)
 #endif
     if (type == t)
 	return this;
-    e = this;
-    tb = t->toBasetype();
-    type = type->toBasetype();
-    if (tb != type)
+    Expression *e = this;
+    Type *tb = t->toBasetype();
+    Type *typeb = type->toBasetype();
+    if (tb != typeb)
     {
 	// Do (type *) cast of (type [dim])
 	if (tb->ty == Tpointer &&
-	    type->ty == Tsarray
+	    typeb->ty == Tsarray
 	   )
 	{
 	    //printf("Converting [dim] to *\n");
 
-	    if (type->size(loc) == 0)
+	    if (typeb->size(loc) == 0)
 		e = new NullExp(loc);
 	    else
 		e = new AddrExp(loc, e);
@@ -658,6 +725,11 @@ Expression *Expression::castTo(Scope *sc, Type *t)
 	    e = new CastExp(loc, e, tb);
 	}
     }
+    else
+    {
+	e = e->copy();	// because of COW for assignment to e->type
+    }
+    assert(e != this);
     e->type = t;
     //printf("Returning: %s\n", e->toChars());
     return e;
@@ -665,40 +737,55 @@ Expression *Expression::castTo(Scope *sc, Type *t)
 
 
 Expression *RealExp::castTo(Scope *sc, Type *t)
-{
-    if (type->isreal() && t->isreal())
-	type = t;
-    else if (type->isimaginary() && t->isimaginary())
-	type = t;
-    else
-	return Expression::castTo(sc, t);
-    return this;
+{   Expression *e = this;
+    if (type != t)
+    {
+	if ((type->isreal() && t->isreal()) ||
+	    (type->isimaginary() && t->isimaginary())
+	   )
+	{   e = copy();
+	    e->type = t;
+	}
+	else
+	    e = Expression::castTo(sc, t);
+    }
+    return e;
 }
 
 
 Expression *ComplexExp::castTo(Scope *sc, Type *t)
-{
-    if (type->iscomplex() && t->iscomplex())
-	type = t;
-    else
-	return Expression::castTo(sc, t);
-    return this;
+{   Expression *e = this;
+    if (type != t)
+    {
+	if (type->iscomplex() && t->iscomplex())
+	{   e = copy();
+	    e->type = t;
+	}
+	else
+	    e = Expression::castTo(sc, t);
+    }
+    return e;
 }
 
 
 Expression *NullExp::castTo(Scope *sc, Type *t)
-{   Expression *e;
+{   NullExp *e;
     Type *tb;
 
     //printf("NullExp::castTo(t = %p)\n", t);
-    committed = 1;
-    e = this;
+    if (type == t)
+    {
+	committed = 1;
+	return this;
+    }
+    e = (NullExp *)copy();
+    e->committed = 1;
     tb = t->toBasetype();
-    type = type->toBasetype();
-    if (tb != type)
+    e->type = type->toBasetype();
+    if (tb != e->type)
     {
 	// NULL implicitly converts to any pointer type or dynamic array
-	if (type->ty == Tpointer && type->nextOf()->ty == Tvoid &&
+	if (e->type->ty == Tpointer && e->type->nextOf()->ty == Tvoid &&
 	    (tb->ty == Tpointer || tb->ty == Tarray || tb->ty == Taarray ||
 	     tb->ty == Tdelegate))
 	{
@@ -718,8 +805,7 @@ Expression *NullExp::castTo(Scope *sc, Type *t)
 	}
 	else
 	{
-	    return Expression::castTo(sc, t);
-	    //e = new CastExp(loc, e, tb);
+	    return e->Expression::castTo(sc, t);
 	}
     }
     e->type = t;
@@ -959,17 +1045,44 @@ Expression *AddrExp::castTo(Scope *sc, Type *t)
     if (tb != type)
     {
 	// Look for pointers to functions where the functions are overloaded.
-	VarExp *ve;
-	FuncDeclaration *f;
+
+	if (e1->op == TOKoverloadset &&
+	    (t->ty == Tpointer || t->ty == Tdelegate) && t->nextOf()->ty == Tfunction)
+	{   OverExp *eo = (OverExp *)e1;
+	    FuncDeclaration *f = NULL;
+	    for (int i = 0; i < eo->vars->a.dim; i++)
+	    {   Dsymbol *s = (Dsymbol *)eo->vars->a.data[i];
+		FuncDeclaration *f2 = s->isFuncDeclaration();
+		assert(f2);
+		if (f2->overloadExactMatch(t->nextOf()))
+		{   if (f)
+			/* Error if match in more than one overload set,
+			 * even if one is a 'better' match than the other.
+			 */
+			ScopeDsymbol::multiplyDefined(loc, f, f2);
+		    else
+			f = f2;
+		}
+	    }
+	    if (f)
+	    {	f->tookAddressOf = 1;
+		SymOffExp *se = new SymOffExp(loc, f, 0, 0);
+		se->semantic(sc);
+		// Let SymOffExp::castTo() do the heavy lifting
+		return se->castTo(sc, t);
+	    }
+	}
+
 
 	if (type->ty == Tpointer && type->nextOf()->ty == Tfunction &&
 	    tb->ty == Tpointer && tb->nextOf()->ty == Tfunction &&
 	    e1->op == TOKvar)
 	{
-	    ve = (VarExp *)e1;
-	    f = ve->var->isFuncDeclaration();
+	    VarExp *ve = (VarExp *)e1;
+	    FuncDeclaration *f = ve->var->isFuncDeclaration();
 	    if (f)
 	    {
+		assert(0);	// should be SymOffExp instead
 		f = f->overloadExactMatch(tb->nextOf());
 		if (f)
 		{
@@ -989,29 +1102,32 @@ Expression *AddrExp::castTo(Scope *sc, Type *t)
 
 
 Expression *TupleExp::castTo(Scope *sc, Type *t)
-{
-    for (size_t i = 0; i < exps->dim; i++)
-    {   Expression *e = (Expression *)exps->data[i];
-	e = e->castTo(sc, t);
-	exps->data[i] = (void *)e;
+{   TupleExp *e = (TupleExp *)copy();
+    e->exps = (Expressions *)exps->copy();
+    for (size_t i = 0; i < e->exps->dim; i++)
+    {   Expression *ex = (Expression *)e->exps->data[i];
+	ex = ex->castTo(sc, t);
+	e->exps->data[i] = (void *)ex;
     }
-    return this;
+    return e;
 }
 
 
 Expression *ArrayLiteralExp::castTo(Scope *sc, Type *t)
 {
 #if 0
-    printf("ArrayLiteralExp::castTo(this=%s, type=%s, t=%s)\n",
+    printf("ArrayLiteralExp::castTo(this=%s, type=%s, => %s)\n",
 	toChars(), type->toChars(), t->toChars());
 #endif
     if (type == t)
 	return this;
+    ArrayLiteralExp *e = this;
     Type *typeb = type->toBasetype();
     Type *tb = t->toBasetype();
     if ((tb->ty == Tarray || tb->ty == Tsarray) &&
 	(typeb->ty == Tarray || typeb->ty == Tsarray) &&
-	tb->nextOf()->toBasetype()->ty != Tvoid)
+	// Not trying to convert non-void[] to void[]
+	!(tb->nextOf()->toBasetype()->ty == Tvoid && typeb->nextOf()->toBasetype()->ty != Tvoid))
     {
 	if (tb->ty == Tsarray)
 	{   TypeSArray *tsa = (TypeSArray *)tb;
@@ -1019,65 +1135,73 @@ Expression *ArrayLiteralExp::castTo(Scope *sc, Type *t)
 		goto L1;
 	}
 
+	e = (ArrayLiteralExp *)copy();
+	e->elements = (Expressions *)elements->copy();
 	for (int i = 0; i < elements->dim; i++)
-	{   Expression *e = (Expression *)elements->data[i];
-	    e = e->castTo(sc, tb->nextOf());
-	    elements->data[i] = (void *)e;
+	{   Expression *ex = (Expression *)elements->data[i];
+	    ex = ex->castTo(sc, tb->nextOf());
+	    e->elements->data[i] = (void *)ex;
 	}
-	type = t;
-	return this;
+	e->type = t;
+	return e;
     }
     if (tb->ty == Tpointer && typeb->ty == Tsarray)
     {
-	type = typeb->nextOf()->pointerTo();
+	e = (ArrayLiteralExp *)copy();
+	e->type = typeb->nextOf()->pointerTo();
     }
 L1:
-    return Expression::castTo(sc, t);
+    return e->Expression::castTo(sc, t);
 }
 
 Expression *AssocArrayLiteralExp::castTo(Scope *sc, Type *t)
 {
+    if (type == t)
+	return this;
+    AssocArrayLiteralExp *e = this;
     Type *typeb = type->toBasetype();
     Type *tb = t->toBasetype();
     if (tb->ty == Taarray && typeb->ty == Taarray &&
 	tb->nextOf()->toBasetype()->ty != Tvoid)
     {
+	e = (AssocArrayLiteralExp *)copy();
+	e->keys = (Expressions *)keys->copy();
+	e->values = (Expressions *)values->copy();
 	assert(keys->dim == values->dim);
 	for (size_t i = 0; i < keys->dim; i++)
-	{   Expression *e = (Expression *)values->data[i];
-	    e = e->castTo(sc, tb->nextOf());
-	    values->data[i] = (void *)e;
+	{   Expression *ex = (Expression *)values->data[i];
+	    ex = ex->castTo(sc, tb->nextOf());
+	    e->values->data[i] = (void *)ex;
 
-	    e = (Expression *)keys->data[i];
-	    e = e->castTo(sc, ((TypeAArray *)tb)->index);
-	    keys->data[i] = (void *)e;
+	    ex = (Expression *)keys->data[i];
+	    ex = ex->castTo(sc, ((TypeAArray *)tb)->index);
+	    e->keys->data[i] = (void *)ex;
 	}
-	type = t;
-	return this;
+	e->type = t;
+	return e;
     }
 L1:
-    return Expression::castTo(sc, t);
+    return e->Expression::castTo(sc, t);
 }
 
 Expression *SymOffExp::castTo(Scope *sc, Type *t)
 {
-    Type *tb;
-
 #if 0
     printf("SymOffExp::castTo(this=%s, type=%s, t=%s)\n",
 	toChars(), type->toChars(), t->toChars());
 #endif
-    Expression *e = this;
-
-    tb = t->toBasetype();
-    type = type->toBasetype();
-    if (tb != type)
+    if (type == t && hasOverloads == 0)
+	return this;
+    Expression *e;
+    Type *tb = t->toBasetype();
+    Type *typeb = type->toBasetype();
+    if (tb != typeb)
     {
 	// Look for pointers to functions where the functions are overloaded.
 	FuncDeclaration *f;
 
 	if (hasOverloads &&
-	    type->ty == Tpointer && type->nextOf()->ty == Tfunction &&
+	    typeb->ty == Tpointer && typeb->nextOf()->ty == Tfunction &&
 	    (tb->ty == Tpointer || tb->ty == Tdelegate) && tb->nextOf()->ty == Tfunction)
 	{
 	    f = var->isFuncDeclaration();
@@ -1101,38 +1225,38 @@ Expression *SymOffExp::castTo(Scope *sc, Type *t)
 			e = new SymOffExp(loc, f, 0);
 			e->type = t;
 		    }
+		    f->tookAddressOf = 1;
 		    return e;
 		}
 	    }
 	}
 	e = Expression::castTo(sc, t);
     }
-    if (e == this && hasOverloads)
-    {	e = syntaxCopy();
+    else
+    {	e = copy();
+	e->type = t;
 	((SymOffExp *)e)->hasOverloads = 0;
     }
-    e->type = t;
     return e;
 }
 
 Expression *DelegateExp::castTo(Scope *sc, Type *t)
 {
-    Type *tb;
 #if 0
     printf("DelegateExp::castTo(this=%s, type=%s, t=%s)\n",
 	toChars(), type->toChars(), t->toChars());
 #endif
-    Expression *e = this;
     static char msg[] = "cannot form delegate due to covariant return type";
 
-    tb = t->toBasetype();
-    type = type->toBasetype();
-    if (tb != type)
+    Expression *e = this;
+    Type *tb = t->toBasetype();
+    Type *typeb = type->toBasetype();
+    if (tb != typeb)
     {
 	// Look for delegates to functions where the functions are overloaded.
 	FuncDeclaration *f;
 
-	if (type->ty == Tdelegate && type->nextOf()->ty == Tfunction &&
+	if (typeb->ty == Tdelegate && typeb->nextOf()->ty == Tfunction &&
 	    tb->ty == Tdelegate && tb->nextOf()->ty == Tfunction)
 	{
 	    if (func)
@@ -1142,6 +1266,7 @@ Expression *DelegateExp::castTo(Scope *sc, Type *t)
 		{   target_ptrdiff_t offset;
 		    if (f->tintro && f->tintro->nextOf()->isBaseOf(f->type->nextOf(), &offset) && offset)
 			error("%s", msg);
+		    f->tookAddressOf = 1;
 		    e = new DelegateExp(loc, e1, f);
 		    e->type = t;
 		    return e;
@@ -1155,10 +1280,12 @@ Expression *DelegateExp::castTo(Scope *sc, Type *t)
     else
     {	target_ptrdiff_t offset;
 
+	func->tookAddressOf = 1;
 	if (func->tintro && func->tintro->nextOf()->isBaseOf(func->type->nextOf(), &offset) && offset)
 	    error("%s", msg);
+	e = copy();
+	e->type = t;
     }
-    e->type = t;
     return e;
 }
 
@@ -1232,7 +1359,7 @@ Expression *BinExp::typeCombine(Scope *sc)
     Type *t;
     TY ty;
 
-    //printf("BinExp::typeCombine()\n");
+    //printf("BinExp::typeCombine() %s\n", toChars());
     //dump(0);
 
     e1 = e1->integralPromotions(sc);
@@ -1242,6 +1369,7 @@ Expression *BinExp::typeCombine(Scope *sc)
     t1 = e1->type;
     t2 = e2->type;
     assert(t1);
+    t = t1;
 
     //if (t1) printf("\tt1 = %s\n", t1->toChars());
     //if (t2) printf("\tt2 = %s\n", t2->toChars());
@@ -1298,19 +1426,35 @@ Expression *BinExp::typeCombine(Scope *sc)
 	return this;
     }
 
+    t1 = t1b;
+    t2 = t2b;
+
     if (op == TOKcat)
     {
 	if ((t1->ty == Tsarray || t1->ty == Tarray) &&
 	    (t2->ty == Tsarray || t2->ty == Tarray) &&
-	    (t1->nextOf()->mod || t2->nextOf()->mod))
+	    (t1->nextOf()->mod || t2->nextOf()->mod) &&
+	    (t1->nextOf()->mod != t2->nextOf()->mod)
+	   )
 	{
-	    t1 = t1->constOf();
-	    t2 = t2->constOf();
+	    t1 = t1->nextOf()->mutableOf()->constOf()->arrayOf();
+	    t2 = t2->nextOf()->mutableOf()->constOf()->arrayOf();
+//t1 = t1->constOf();
+//t2 = t2->constOf();
+	    if (e1->op == TOKstring && !((StringExp *)e1)->committed)
+		e1->type = t1;
+	    else
+		e1 = e1->castTo(sc, t1);
+	    if (e2->op == TOKstring && !((StringExp *)e2)->committed)
+		e2->type = t2;
+	    else
+		e2 = e2->castTo(sc, t2);
+	    t = t1;
+	    goto Lagain;
 	}
     }
 
 Lagain:
-    t = t1;
     if (t1 == t2)
     {
 	if ((t1->ty == Tstruct || t1->ty == Tclass) &&
@@ -1331,8 +1475,9 @@ Lagain:
 	    ;
 	else if (t1n->mod != t2n->mod)
 	{
-	    t1 = t1->constOf();
-	    t2 = t2->constOf();
+	    t1 = t1n->mutableOf()->constOf()->pointerTo();
+	    t2 = t2n->mutableOf()->constOf()->pointerTo();
+	    t = t1;
 	    goto Lagain;
 	}
 	else if (t1n->ty == Tclass && t2n->ty == Tclass)
@@ -1387,36 +1532,65 @@ Lagain:
 	     t1->nextOf()->mod != t2->nextOf()->mod
 	    )
     {
-	t1 = t1->constOf();
-	t2 = t2->constOf();
+	if (t1->ty == Tpointer)
+	    t1 = t1->nextOf()->mutableOf()->constOf()->pointerTo();
+	else
+	    t1 = t1->nextOf()->mutableOf()->constOf()->arrayOf();
+
+	if (t2->ty == Tpointer)
+	    t2 = t2->nextOf()->mutableOf()->constOf()->pointerTo();
+	else
+	    t2 = t2->nextOf()->mutableOf()->constOf()->arrayOf();
+	t = t1;
 	goto Lagain;
     }
     else if (t1->ty == Tclass || t2->ty == Tclass)
-    {	int i1;
-	int i2;
-
-	i1 = e2->implicitConvTo(t1);
-	i2 = e1->implicitConvTo(t2);
-
-	if (i1 && i2)
+    {
+	while (1)
 	{
-	    // We have the case of class vs. void*, so pick class
-	    if (t1->ty == Tpointer)
-		i1 = 0;
-	    else if (t2->ty == Tpointer)
-		i2 = 0;
-	}
+	    int i1 = e2->implicitConvTo(t1);
+	    int i2 = e1->implicitConvTo(t2);
 
-	if (i2)
-	{
-	    goto Lt2;
+	    if (i1 && i2)
+	    {
+		// We have the case of class vs. void*, so pick class
+		if (t1->ty == Tpointer)
+		    i1 = 0;
+		else if (t2->ty == Tpointer)
+		    i2 = 0;
+	    }
+
+	    if (i2)
+	    {
+		goto Lt2;
+	    }
+	    else if (i1)
+	    {
+		goto Lt1;
+	    }
+	    else if (t1->ty == Tclass && t2->ty == Tclass)
+	    {	TypeClass *tc1 = (TypeClass *)t1;
+		TypeClass *tc2 = (TypeClass *)t2;
+
+		/* Pick 'tightest' type
+		 */
+		ClassDeclaration *cd1 = tc1->sym->baseClass;
+		ClassDeclaration *cd2 = tc1->sym->baseClass;
+
+		if (cd1 && cd2)
+		{   t1 = cd1->type;
+		    t2 = cd2->type;
+		}
+		else if (cd1)
+		    t1 = cd1->type;
+		else if (cd2)
+		    t2 = cd2->type;
+		else
+		    goto Lincompatible;
+	    }
+	    else
+		goto Lincompatible;
 	}
-	else if (i1)
-	{
-	    goto Lt1;
-	}
-	else
-	    goto Lincompatible;
     }
     else if (t1->ty == Tstruct && t2->ty == Tstruct)
     {
@@ -1427,7 +1601,6 @@ Lagain:
     {
 	goto Lt2;
     }
-//else if (e2->op == TOKstring) { printf("test2\n"); }
     else if ((e2->op == TOKstring || e2->op == TOKnull) && e2->implicitConvTo(t1))
     {
 	goto Lt1;
@@ -1460,6 +1633,12 @@ Lagain:
 Lret:
     if (!type)
 	type = t;
+#if 0
+    printf("-BinExp::typeCombine() %s\n", toChars());
+    if (e1->type) printf("\tt1 = %s\n", e1->type->toChars());
+    if (e2->type) printf("\tt2 = %s\n", e2->type->toChars());
+    printf("\ttype = %s\n", type->toChars());
+#endif
     //dump(0);
     return this;
 
@@ -1481,9 +1660,10 @@ Lt2:
  */
 
 Expression *Expression::integralPromotions(Scope *sc)
-{   Expression *e;
+{
+    Expression *e = this;
 
-    e = this;
+    //printf("integralPromotions %s %s\n", e->toChars(), e->type->toChars());
     switch (type->toBasetype()->ty)
     {
 	case Tvoid:
