@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -153,6 +153,7 @@ ClassDeclaration::ClassDeclaration(Loc loc, Identifier *id, BaseClasses *basecla
 		Type::typeinfotypelist = this;
 	    }
 
+#if V2
 	    if (id == Id::TypeInfo_Const)
 	    {	if (Type::typeinfoconst)
 		    Type::typeinfoconst->error("%s", msg);
@@ -164,6 +165,7 @@ ClassDeclaration::ClassDeclaration(Loc loc, Identifier *id, BaseClasses *basecla
 		    Type::typeinfoinvariant->error("%s", msg);
 		Type::typeinfoinvariant = this;
 	    }
+#endif
 	}
 
 	if (id == Id::Object)
@@ -186,14 +188,11 @@ ClassDeclaration::ClassDeclaration(Loc loc, Identifier *id, BaseClasses *basecla
     }
 
     com = 0;
-#if 0
-    if (id == Id::IUnknown)		// IUnknown is the root of all COM objects
-	com = 1;
-#endif
     isauto = 0;
     isabstract = 0;
     isnested = 0;
     vthis = NULL;
+    inuse = 0;
 }
 
 Dsymbol *ClassDeclaration::syntaxCopy(Dsymbol *s)
@@ -270,6 +269,14 @@ void ClassDeclaration::semantic(Scope *sc)
     methods.setDim(0);
 #endif
 
+    if (sc->stc & STCdeprecated)
+    {
+	isdeprecated = 1;
+    }
+
+    if (sc->linkage == LINKcpp)
+	error("cannot create C++ classes");
+
     // Expand any tuples in baseclasses[]
     for (i = 0; i < baseclasses.dim; )
     {	BaseClass *b = (BaseClass *)baseclasses.data[i];
@@ -307,6 +314,18 @@ void ClassDeclaration::semantic(Scope *sc)
 	else
 	{
 	    tc = (TypeClass *)(tb);
+
+	    if (tc->sym->isDeprecated())
+	    {
+		if (!isDeprecated())
+		{
+		    // Deriving from deprecated class makes this one deprecated too
+		    isdeprecated = 1;
+
+		    tc->checkDeprecated(loc, sc);
+		}
+	    }
+
 	    if (tc->sym->isInterfaceDeclaration())
 		;
 	    else
@@ -361,6 +380,17 @@ void ClassDeclaration::semantic(Scope *sc)
 	}
 	else
 	{
+	    if (tc->sym->isDeprecated())
+	    {
+		if (!isDeprecated())
+		{
+		    // Deriving from deprecated class makes this one deprecated too
+		    isdeprecated = 1;
+
+		    tc->checkDeprecated(loc, sc);
+		}
+	    }
+
 	    // Check for duplicate interfaces
 	    for (size_t j = (baseClass ? 1 : 0); j < i; j++)
 	    {
@@ -429,6 +459,7 @@ void ClassDeclaration::semantic(Scope *sc)
 	com = baseClass->isCOMclass();
 	isauto = baseClass->isauto;
 	vthis = baseClass->vthis;
+	storage_class |= baseClass->storage_class & (STCconst | STCinvariant);
     }
     else
     {
@@ -501,12 +532,15 @@ void ClassDeclaration::semantic(Scope *sc)
 	isauto = 1;
     if (storage_class & STCabstract)
 	isabstract = 1;
-    if (storage_class & STCdeprecated)
-	isdeprecated = 1;
+    if (storage_class & STCinvariant)
+	type = type->invariantOf();
+    else if (storage_class & STCconst)
+	type = type->constOf();
 
     sc = sc->push(this);
     sc->stc &= ~(STCfinal | STCauto | STCscope | STCstatic |
-		 STCabstract | STCdeprecated);
+		 STCabstract | STCdeprecated | STCconst | STCinvariant | STCtls);
+    sc->stc |= storage_class & (STCconst | STCinvariant);
     sc->attributes = NULL;
     sc->parent = this;
     sc->inunion = 0;
@@ -551,7 +585,7 @@ void ClassDeclaration::semantic(Scope *sc)
 	scope->setNoFree();
 	scope->module->addDeferredSemantic(this);
 
-	//printf("\tsemantic('%s') failed\n", toChars());
+	//printf("\tsemantic('%s') failed due to forward references\n", toChars());
 	return;
     }
 
@@ -629,6 +663,7 @@ void ClassDeclaration::semantic(Scope *sc)
     sizeok = 1;
     Module::dprogress++;
 
+    dtor = buildDtor(sc);
 
     sc->pop();
 
@@ -781,6 +816,7 @@ Dsymbol *ClassDeclaration::search(Loc loc, Identifier *ident, int flags)
  * Return 1 if function is hidden (not findable through search).
  */
 
+#if V2
 int isf(void *param, FuncDeclaration *fd)
 {
     //printf("param = %p, fd = %p %s\n", param, fd, fd->toChars());
@@ -798,10 +834,26 @@ int ClassDeclaration::isFuncHidden(FuncDeclaration *fd)
 	 */
 	return 0;
     }
-    FuncDeclaration *fdstart = s->toAlias()->isFuncDeclaration();
-    //printf("%s fdstart = %p\n", s->kind(), fdstart);
-    return !overloadApply(fdstart, &isf, fd);
+    s = s->toAlias();
+    OverloadSet *os = s->isOverloadSet();
+    if (os)
+    {
+	for (int i = 0; i < os->a.dim; i++)
+	{   Dsymbol *s = (Dsymbol *)os->a.data[i];
+	    FuncDeclaration *f2 = s->isFuncDeclaration();
+	    if (f2 && overloadApply(f2, &isf, fd))
+		return 0;
+	}
+	return 1;
+    }
+    else
+    {
+	FuncDeclaration *fdstart = s->isFuncDeclaration();
+	//printf("%s fdstart = %p\n", s->kind(), fdstart);
+	return !overloadApply(fdstart, &isf, fd);
+    }
 }
+#endif
 
 /****************
  * Find virtual function matching identifier and type.
@@ -840,20 +892,24 @@ FuncDeclaration *ClassDeclaration::findFunc(Identifier *ident, TypeFunction *tf)
 }
 
 void ClassDeclaration::interfaceSemantic(Scope *sc)
-{   int i;
+{
+    InterfaceDeclaration *id = isInterfaceDeclaration();
 
     vtblInterfaces = new BaseClasses();
     vtblInterfaces->reserve(interfaces_dim);
 
-    for (i = 0; i < interfaces_dim; i++)
+    for (size_t i = 0; i < interfaces_dim; i++)
     {
 	BaseClass *b = interfaces[i];
 
 	// If this is an interface, and it derives from a COM interface,
 	// then this is a COM interface too.
-	if (b->base->isCOMclass())
+	if (b->base->isCOMinterface())
 	    com = 1;
 
+	if (b->base->isCPPinterface() && id)
+	    id->cpp = 1;
+
 	vtblInterfaces->push(b);
 	b->copyBaseInterfaces(vtblInterfaces);
     }
@@ -867,6 +923,16 @@ int ClassDeclaration::isCOMclass()
     return com;
 }
 
+int ClassDeclaration::isCOMinterface()
+{
+    return 0;
+}
+
+int ClassDeclaration::isCPPinterface()
+{
+    return 0;
+}
+
 
 /****************************************
  */
@@ -889,6 +955,7 @@ int ClassDeclaration::isAbstract()
     return FALSE;
 }
 
+
 /****************************************
  * Returns !=0 if there's an extra member which is the 'this'
  * pointer to the enclosing context (enclosing class or function)
@@ -933,8 +1000,11 @@ InterfaceDeclaration::InterfaceDeclaration(Loc loc, Identifier *id, BaseClasses
     : ClassDeclaration(loc, id, baseclasses)
 {
     com = 0;
-    if (id == Id::IUnknown)		// IUnknown is the root of all COM objects
-	com = 1;
+    cpp = 0;
+    if (id == Id::IUnknown)	// IUnknown is the root of all COM interfaces
+    {	com = 1;
+	cpp = 1;		// IUnknown is also a C++ interface
+    }
 }
 
 Dsymbol *InterfaceDeclaration::syntaxCopy(Dsymbol *s)
@@ -954,6 +1024,8 @@ void InterfaceDeclaration::semantic(Scope *sc)
 {   int i;
 
     //printf("InterfaceDeclaration::semantic(%s), type = %p\n", toChars(), type);
+    if (inuse)
+	return;
     if (!scope)
     {	type = type->semantic(loc, sc);
 	handle = handle->semantic(loc, sc);
@@ -981,6 +1053,11 @@ void InterfaceDeclaration::semantic(Scope *sc)
     else
 	attributes = sc->attributes;
 
+    if (sc->stc & STCdeprecated)
+    {
+	isdeprecated = 1;
+    }
+
     // Expand any tuples in baseclasses[]
     for (i = 0; i < baseclasses.dim; )
     {	BaseClass *b = (BaseClass *)baseclasses.data[0];
@@ -1002,6 +1079,9 @@ void InterfaceDeclaration::semantic(Scope *sc)
 	    i++;
     }
 
+    if (!baseclasses.dim && sc->linkage == LINKcpp)
+	cpp = 1;
+
     // Check for errors, handle forward references
     for (i = 0; i < baseclasses.dim; )
     {	TypeClass *tc;
@@ -1038,7 +1118,7 @@ void InterfaceDeclaration::semantic(Scope *sc)
 		baseclasses.remove(i);
 		continue;
 	    }
-	    if (!b->base->symtab || b->base->scope)
+	    if (!b->base->symtab || b->base->scope || b->base->inuse)
 	    {
 		//error("forward reference of base class %s", baseClass->toChars());
 		// Forward reference of base, try again later
@@ -1049,6 +1129,10 @@ void InterfaceDeclaration::semantic(Scope *sc)
 		return;
 	    }
 	}
+#if 0
+	// Inherit const/invariant from base class
+	storage_class |= b->base->storage_class & (STCconst | STCinvariant);
+#endif
 	i++;
     }
 
@@ -1090,6 +1174,9 @@ void InterfaceDeclaration::semantic(Scope *sc)
 	;
     }
 
+    protection = sc->protection;
+    storage_class |= sc->stc & (STCconst | STCinvariant);
+
     for (i = 0; i < members->dim; i++)
     {
 	Dsymbol *s = (Dsymbol *)members->data[i];
@@ -1097,18 +1184,25 @@ void InterfaceDeclaration::semantic(Scope *sc)
     }
 
     sc = sc->push(this);
+    sc->stc &= ~(STCfinal | STCauto | STCscope | STCstatic |
+                 STCabstract | STCdeprecated | STCconst | STCinvariant | STCtls);
+    sc->stc |= storage_class & (STCconst | STCinvariant);
     sc->attributes = NULL;
     sc->parent = this;
-    if (isCOMclass())
+    if (isCOMinterface())
 	sc->linkage = LINKwindows;
+    else if (isCPPinterface())
+	sc->linkage = LINKcpp;
     sc->structalign = 8;
     structalign = sc->structalign;
     sc->offset = PTRSIZE * 2;
+    inuse++;
     for (i = 0; i < members->dim; i++)
     {
 	Dsymbol *s = (Dsymbol *)members->data[i];
 	s->semantic(sc);
     }
+    inuse--;
     //members->print();
     sc->pop();
     //printf("-InterfaceDeclaration::semantic(%s), type = %p\n", toChars(), type);
@@ -1189,18 +1283,28 @@ int InterfaceDeclaration::isBaseOf(BaseClass *bc, target_ptrdiff_t *poffset)
 
 /****************************************
  * Determine if slot 0 of the vtbl[] is reserved for something else.
- * For class objects, yes, this is where the classinfo ptr goes.
+ * For class objects, yes, this is where the ClassInfo ptr goes.
  * For COM interfaces, no.
  * For non-COM interfaces, yes, this is where the Interface ptr goes.
  */
 
 int InterfaceDeclaration::vtblOffset()
 {
-    if (isCOMclass())
+    if (isCOMinterface() || isCPPinterface())
 	return 0;
     return 1;
 }
 
+int InterfaceDeclaration::isCOMinterface()
+{
+    return com;
+}
+
+int InterfaceDeclaration::isCPPinterface()
+{
+    return cpp;
+}
+
 /*******************************************
  */
 
