@@ -736,6 +736,36 @@ Expression *Equal(enum TOK op, Type *type, Expression *e1, Expression *e2)
 	    }
 	}
     }
+    else if (e1->op == TOKarrayliteral && e2->op == TOKstring)
+    {	// Swap operands and use common code
+	Expression *e = e1;
+	e1 = e2;
+	e2 = e;
+	goto Lsa;
+    }
+    else if (e1->op == TOKstring && e2->op == TOKarrayliteral)
+    {
+     Lsa:
+	StringExp *es1 = (StringExp *)e1;
+	ArrayLiteralExp *es2 = (ArrayLiteralExp *)e2;
+	size_t dim1 = es1->len;
+	size_t dim2 = es2->elements ? es2->elements->dim : 0;
+	if (dim1 != dim2)
+	    cmp = 0;
+	else
+	{
+	    for (size_t i = 0; i < dim1; i++)
+	    {
+		uinteger_t c = es1->charAt(i);
+		Expression *ee2 = (Expression *)es2->elements->data[i];
+		if (ee2->isConst() != 1)
+		    return EXP_CANT_INTERPRET;
+		cmp = (c == ee2->toInteger());
+		if (cmp == 0)
+		    break;
+	    }
+	}
+    }
     else if (e1->op == TOKstructliteral && e2->op == TOKstructliteral)
     {   StructLiteralExp *es1 = (StructLiteralExp *)e1;
 	StructLiteralExp *es2 = (StructLiteralExp *)e2;
@@ -856,7 +886,70 @@ Expression *Cmp(enum TOK op, Type *type, Expression *e1, Expression *e2)
     real_t r1;
     real_t r2;
 
-    if (e1->type->isreal())
+    //printf("Cmp(e1 = %s, e2 = %s)\n", e1->toChars(), e2->toChars());
+
+    if (e1->op == TOKstring && e2->op == TOKstring)
+    {	StringExp *es1 = (StringExp *)e1;
+	StringExp *es2 = (StringExp *)e2;
+	size_t sz = es1->sz;
+	assert(sz == es2->sz);
+
+	size_t len = es1->len;
+	if (es2->len < len)
+	    len = es2->len;
+
+	int cmp = 0;
+	if (sz == 1)
+	    cmp = memcmp(es1->string, es2->string, sz * len);
+	else if (sz == 2)
+	{
+	    const d_uns16 * p1 = (const d_uns16 *) es1->string;
+	    const d_uns16 * p2 = (const d_uns16 *) es2->string;
+	    for (size_t i = 0; i < len; ++i)
+		if (p1[i] > p2[i])
+		    { cmp = 1; break; }
+		else if (p1[i] < p2[i])
+		    { cmp = -1; break; }
+	}
+	else if (sz == 4)
+	{
+	    const d_uns32 * p1 = (const d_uns32 *) es1->string;
+	    const d_uns32 * p2 = (const d_uns32 *) es2->string;
+	    for (size_t i = 0; i < len; ++i)
+		if (p1[i] > p2[i])
+		    { cmp = 1; break; }
+		else if (p1[i] < p2[i])
+		    { cmp = -1; break; }
+	}
+	else
+	    assert(0);
+	
+	if (cmp == 0)
+	    cmp = es1->len - es2->len;
+
+	switch (op)
+	{
+	    case TOKlt:	n = cmp <  0;	break;
+	    case TOKle:	n = cmp <= 0;	break;
+	    case TOKgt:	n = cmp >  0;	break;
+	    case TOKge:	n = cmp >= 0;	break;
+
+	    case TOKleg:   n = 1;		break;
+	    case TOKlg:	   n = cmp != 0;	break;
+	    case TOKunord: n = 0;		break;
+	    case TOKue:	   n = cmp == 0;	break;
+	    case TOKug:	   n = cmp >  0;	break;
+	    case TOKuge:   n = cmp >= 0;	break;
+	    case TOKul:	   n = cmp <  0;	break;
+	    case TOKule:   n = cmp <= 0;	break;
+
+	    default:
+		assert(0);
+	}
+    }
+    else if (e1->isConst() != 1 || e2->isConst() != 1)
+	return EXP_CANT_INTERPRET;
+    else if (e1->type->isreal())
     {
 	r1 = e1->toReal();
 	r2 = e2->toReal();
@@ -1039,7 +1132,6 @@ Expression *Cast(Type *type, Type *to, Expression *e1)
 		default: assert(0);
 		}
 	    }
-
 	    d_int64 r = e1->toReal().toInt(rt, type);
 #else
 	    real_t r = e1->toReal();
@@ -1158,26 +1250,7 @@ Expression *Index(Type *type, Expression *e1, Expression *e2)
 	if (i >= es1->len)
 	    e1->error("string index %"PRIuMAX" is out of bounds [0 .. %"PRIuSIZE"]", i, es1->len);
 	else
-	{   integer_t value;
-
-	    switch (es1->sz)
-	    {
-		case 1:
-		    value = ((unsigned char *)es1->string)[i];
-		    break;
-
-		case 2:
-		    value = ((unsigned short *)es1->string)[i];
-		    break;
-
-		case 4:
-		    value = ((unsigned int *)es1->string)[i];
-		    break;
-
-		default:
-		    assert(0);
-		    break;
-	    }
+	{   unsigned value = es1->charAt(i);
 	    e = new IntegerExp(loc, value, type);
 	}
     }
