@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -35,6 +35,7 @@ Declaration::Declaration(Identifier *id)
     : Dsymbol(id)
 {
     type = NULL;
+    originalType = NULL;
     storage_class = STCundefined;
     protection = PROTundefined;
     linkage = LINKdefault;
@@ -91,16 +92,11 @@ enum PROT Declaration::prot()
  * Issue error if not.
  */
 
-void Declaration::checkModify(Loc loc, Scope *sc)
+void Declaration::checkModify(Loc loc, Scope *sc, Type *t)
 {
     if (sc->incontract && isParameter())
 	error(loc, "cannot modify parameter '%s' in contract", toChars());
 
-    VarDeclaration *v = isVarDeclaration();
-    if (v && v->canassign == 0 &&
-        (isConst() || isInvariant() || (isFinal() && !isCtorinit())))
-	error(loc, "cannot modify final/const/invariant variable '%s'", toChars());
-
     if (isCtorinit())
     {	// It's only modifiable if inside the right constructor
 	Dsymbol *s = sc->func;
@@ -115,6 +111,7 @@ void Declaration::checkModify(Loc loc, Scope *sc)
 		fd->toParent() == toParent()
 	       )
 	    {
+		VarDeclaration *v = isVarDeclaration();
 		assert(v);
 		v->ctorinit = 1;
 		//printf("setting ctorinit\n");
@@ -135,6 +132,26 @@ void Declaration::checkModify(Loc loc, Scope *sc)
 	    break;
 	}
     }
+    else
+    {
+	VarDeclaration *v = isVarDeclaration();
+	if (v && v->canassign == 0)
+	{
+	    char *p = NULL;
+	    if (isConst())
+		p = "const";
+	    else if (isInvariant())
+		p = "invariant";
+	    else if (storage_class & STCmanifest)
+		p = "manifest constant";
+	    else if (!t->isAssignable())
+		p = "struct with immutable members";
+	    if (p)
+	    {	error(loc, "cannot modify %s", p);
+		halt();
+	    }
+	}
+    }
 }
 
 
@@ -380,6 +397,7 @@ AliasDeclaration::AliasDeclaration(Loc loc, Identifier *id, Dsymbol *s)
 
 Dsymbol *AliasDeclaration::syntaxCopy(Dsymbol *s)
 {
+    //printf("AliasDeclaration::syntaxCopy()\n");
     assert(!s);
     AliasDeclaration *sa;
     if (type)
@@ -443,7 +461,7 @@ void AliasDeclaration::semantic(Scope *sc)
      * try to alias y to 3.
      */
     s = type->toDsymbol(sc);
-    if (s)
+    if (s && ((s->getType() && type->equals(s->getType())) || s->isEnumMember()))
 	goto L2;			// it's a symbolic alias
 
     //printf("alias type is %s\n", type->toChars());
@@ -455,18 +473,12 @@ void AliasDeclaration::semantic(Scope *sc)
     else if (e)
     {
 	// Try to convert Expression to Dsymbol
-        if (e->op == TOKvar)
-	{   s = ((VarExp *)e)->var;
+	s = getDsymbol(e);
+	if (s)
 	    goto L2;
-	}
-        else if (e->op == TOKfunction)
-	{   s = ((FuncExp *)e)->fd;
-	    goto L2;
-	}
-        else
-	{   error("cannot alias an expression %s", e->toChars());
-	    t = e->type;
-	}
+
+	error("cannot alias an expression %s", e->toChars());
+	t = e->type;
     }
     else if (t)
 	type = t;
@@ -604,13 +616,13 @@ VarDeclaration::VarDeclaration(Loc loc, Type *type, Identifier *id, Initializer
     this->loc = loc;
     offset = 0;
     noauto = 0;
-    nestedref = 0;
     inuse = 0;
     ctorinit = 0;
     aliassym = NULL;
     onstack = 0;
     canassign = 0;
     value = NULL;
+    scope = NULL;
 }
 
 Dsymbol *VarDeclaration::syntaxCopy(Dsymbol *s)
@@ -658,7 +670,9 @@ Dsymbol *VarDeclaration::syntaxCopy(Dsymbol *s)
 void VarDeclaration::semantic(Scope *sc)
 {
     //printf("VarDeclaration::semantic('%s', parent = '%s')\n", toChars(), sc->parent->toChars());
-    //printf(" type = %s\n", type->toChars());
+    //printf(" type = %s\n", type ? type->toChars() : "null");
+    //printf(" stc = x%x\n", sc->stc);
+    //printf(" storage_class = x%x\n", storage_class);
     //printf("linkage = %d\n", sc->linkage);
     //if (strcmp(toChars(), "mul") == 0) halt();
 
@@ -679,9 +693,14 @@ void VarDeclaration::semantic(Scope *sc)
 	 * declarations.
 	 */
 	storage_class &= ~STCauto;
+	originalType = type;
     }
     else
+    {	if (!originalType)
+	    originalType = type;
 	type = type->semantic(loc, sc);
+    }
+    //printf(" semantic type = %s\n", type ? type->toChars() : "null");
 
     type->checkDeprecated(loc, sc);
     linkage = sc->linkage;
@@ -726,6 +745,7 @@ void VarDeclaration::semantic(Scope *sc)
 	size_t nelems = Argument::dim(tt->arguments);
 	Objects *exps = new Objects();
 	exps->setDim(nelems);
+	Expression *ie = init ? init->toExpression() : NULL;
 
 	for (size_t i = 0; i < nelems; i++)
 	{   Argument *arg = Argument::getNth(tt->arguments, i);
@@ -736,7 +756,16 @@ void VarDeclaration::semantic(Scope *sc)
 	    char *name = (char *)buf.extractData();
 	    Identifier *id = new Identifier(name, TOKidentifier);
 
-	    VarDeclaration *v = new VarDeclaration(loc, arg->type, id, NULL);
+	    Expression *einit = ie;
+	    if (ie && ie->op == TOKtuple)
+	    {	einit = (Expression *)((TupleExp *)ie)->exps->data[i];
+	    }
+	    Initializer *ti = init;
+	    if (einit)
+	    {	ti = new ExpInitializer(einit->loc, einit);
+	    }
+
+	    VarDeclaration *v = new VarDeclaration(loc, arg->type, id, ti);
 	    //printf("declaring field %s of type %s\n", v->toChars(), v->type->toChars());
 	    v->semantic(sc);
 
@@ -755,33 +784,22 @@ void VarDeclaration::semantic(Scope *sc)
 	return;
     }
 
-    if (storage_class & STCfinal && !init && !fd)
-    {	// Initialize by constructor only
-	storage_class |= STCctorinit;
-    }
-
-    if (isConst())
-    {
-	type = type->constOf();
-	if (isParameter())
-	{   storage_class |= STCfinal;
-	    storage_class &= ~STCconst;
-	}
-    }
-    else if (storage_class & STCinvariant)
+Lagain:
+    if (storage_class & STCinvariant)
     {
 	type = type->invariantOf();
-	if (isParameter())
-	{   storage_class |= STCfinal;
-	    storage_class &= ~STCinvariant;
-	}
     }
-    type = type->toCanonConst();
-
-    if (storage_class & (STCconst | STCinvariant | STCstatic))
+    else if (storage_class & (STCconst | STCin))
     {
+	if (!type->isInvariant())
+	    type = type->constOf();
     }
-    else if (isSynchronized())
+    else if (type->isConst())
+	storage_class |= STCconst;
+    else if (type->isInvariant())
+	storage_class |= STCinvariant;
+
+    if (isSynchronized())
     {
 	error("variable %s cannot be synchronized", toChars());
     }
@@ -793,7 +811,12 @@ void VarDeclaration::semantic(Scope *sc)
     {
 	error("abstract cannot be applied to variable");
     }
-    else if (storage_class & STCtemplateparameter)
+    else if (storage_class & STCfinal)
+    {
+	error("final cannot be applied to variable");
+    }
+
+    if (storage_class & (STCstatic | STCextern | STCmanifest | STCtemplateparameter | STCtls))
     {
     }
     else
@@ -802,8 +825,15 @@ void VarDeclaration::semantic(Scope *sc)
 	if (!aad)
 	    aad = parent->isAggregateDeclaration();
 	if (aad)
-	{   assert(!(storage_class & (STCconst | STCinvariant | STCstatic)));
-	    aad->addField(sc, this);
+	{   assert(!(storage_class & (STCextern | STCstatic | STCtls)));
+
+	    if (storage_class & (STCconst | STCinvariant) && init)
+	    {
+		if (!type->toBasetype()->isTypeBasic())
+		    storage_class |= STCstatic;
+	    }
+	    else
+		aad->addField(sc, this);
 	}
 
 	InterfaceDeclaration *id = parent->isInterfaceDeclaration();
@@ -812,6 +842,8 @@ void VarDeclaration::semantic(Scope *sc)
 	    error("field not allowed in interface");
 	}
 
+	/* Templates cannot add fields to aggregates
+	 */
 	TemplateInstance *ti = parent->isTemplateInstance();
 	if (ti)
 	{
@@ -835,9 +867,9 @@ void VarDeclaration::semantic(Scope *sc)
 
     if (type->isauto() && !noauto)
     {
-	if (storage_class & (STCfield | STCout | STCref | STCstatic) || !fd)
+	if (storage_class & (STCfield | STCout | STCref | STCstatic | STCmanifest | STCtls) || !fd)
 	{
-	    error("globals, statics, fields, ref and out parameters cannot be auto");
+	    error("globals, statics, fields, manifest constants, ref and out parameters cannot be auto");
 	}
 
 	if (!(storage_class & (STCauto | STCscope)))
@@ -847,8 +879,20 @@ void VarDeclaration::semantic(Scope *sc)
 	}
     }
 
-    if (!init && !sc->inunion && !isStatic() && !isConst() && fd &&
-	!(storage_class & (STCfield | STCin | STCforeach)))
+    if ((isConst() || isInvariant()) && !init && !fd)
+    {	// Initialize by constructor only
+	storage_class |= STCctorinit;
+    }
+
+    if (init)
+	storage_class |= STCinit;     // remember we had an explicit initializer
+    else if (storage_class & STCmanifest)
+	error("manifest constants must have initializers");
+
+    enum TOK op = TOKconstruct;
+    if (!init && !sc->inunion && !isStatic() && fd &&
+	(!(storage_class & (STCfield | STCin | STCforeach | STCparameter)) || (storage_class & STCout)) &&
+	type->size() != 0)
     {
 	// Provide a default initializer
 	//printf("Providing default initializer for '%s'\n", toChars());
@@ -884,6 +928,8 @@ void VarDeclaration::semantic(Scope *sc)
 	{
 	    init = getExpInitializer();
 	}
+	// Default initializer is always a blit
+	op = TOKblit;
     }
 
     if (init)
@@ -913,13 +959,9 @@ void VarDeclaration::semantic(Scope *sc)
 	{
 	    // If local variable, use AssignExp to handle all the various
 	    // possibilities.
-	    if (fd && !isStatic() && !isConst() && !isInvariant() &&
+	    if (fd && !isStatic() && !(storage_class & STCmanifest) &&
 		!init->isVoidInitializer())
 	    {
-		Expression *e1;
-		Type *t;
-		sinteger_t dim;
-
 		//printf("fd = '%s', var = '%s'\n", fd->toChars(), toChars());
 		if (!ei)
 		{
@@ -937,20 +979,24 @@ void VarDeclaration::semantic(Scope *sc)
 		    init = ei;
 		}
 
-		e1 = new VarExp(loc, this);
+		Expression *e1 = new VarExp(loc, this);
 
-		t = type->toBasetype();
+		Type *t = type->toBasetype();
 		if (t->ty == Tsarray)
 		{
-		    dim = ((TypeSArray *)t)->dim->toInteger();
-		    // If multidimensional static array, treat as one large array
-		    while (1)
+		    ei->exp = ei->exp->semantic(sc);
+		    if (!ei->exp->implicitConvTo(type))
 		    {
-			t = t->nextOf()->toBasetype();
-			if (t->ty != Tsarray)
-			    break;
-			dim *= ((TypeSArray *)t)->dim->toInteger();
-			e1->type = new TypeSArray(t->nextOf(), new IntegerExp(0, dim, Type::tindex));
+			sinteger_t dim = ((TypeSArray *)t)->dim->toInteger();
+			// If multidimensional static array, treat as one large array
+			while (1)
+			{
+			    t = t->nextOf()->toBasetype();
+			    if (t->ty != Tsarray)
+				break;
+			    dim *= ((TypeSArray *)t)->dim->toInteger();
+			    e1->type = new TypeSArray(t->nextOf(), new IntegerExp(0, dim, Type::tindex));
+			}
 		    }
 		    e1 = new SliceExp(loc, e1, NULL, NULL);
 		}
@@ -958,10 +1004,14 @@ void VarDeclaration::semantic(Scope *sc)
 		{
 		    ei->exp = ei->exp->semantic(sc);
 		    if (!ei->exp->implicitConvTo(type))
-			ei->exp = new CastExp(loc, ei->exp, type);
+		    {	Type *ti = ei->exp->type->toBasetype();
+			// Don't cast away invariant or mutability in initializer
+			if (!(ti->ty == Tstruct && t->toDsymbol(sc) == ti->toDsymbol(sc)))
+			    ei->exp = new CastExp(loc, ei->exp, type);
+		    }
 		}
 		ei->exp = new AssignExp(loc, e1, ei->exp);
-		ei->exp->op = TOKconstruct;
+		ei->exp->op = op;
 		canassign++;
 		ei->exp = ei->exp->semantic(sc);
 		canassign--;
@@ -970,13 +1020,10 @@ void VarDeclaration::semantic(Scope *sc)
 	    else
 	    {
 		init = init->semantic(sc, type);
-		if (fd && (isConst() || isInvariant()) && !isStatic())
-		{   // Make it static
-		    storage_class |= STCstatic;
-		}
 	    }
 	}
-	else if (isConst() || isInvariant() || isFinal())
+	else if (storage_class & (STCconst | STCinvariant | STCmanifest) ||
+		 type->isConst() || type->isInvariant())
 	{
 	    /* Because we may need the results of a const declaration in a
 	     * subsequent type, such as an array dimension, before semantic2()
@@ -1009,14 +1056,31 @@ void VarDeclaration::semantic(Scope *sc)
 		{
 		    if (global.gag == 0)
 			global.errors = errors;	// act as if nothing happened
+
+		    /* Save scope for later use, to try again
+		     */
+		    scope = new Scope(*sc);
+		    scope->setNoFree();
 		}
 		else if (ei)
 		{
-		    e = e->optimize(WANTvalue | WANTinterpret);
+		    if (isDataseg())
+			/* static const/invariant does CTFE
+			 */
+			e = e->optimize(WANTvalue | WANTinterpret);
+		    else
+			e = e->optimize(WANTvalue);
 		    if (e->op == TOKint64 || e->op == TOKstring)
 		    {
 			ei->exp = e;		// no errors, keep result
 		    }
+		    else
+		    {
+			/* Save scope for later use, to try again
+			 */
+			scope = new Scope(*sc);
+			scope->setNoFree();
+		    }
 		}
 		else
 		    init = i2;		// no errors, keep result
@@ -1025,23 +1089,6 @@ void VarDeclaration::semantic(Scope *sc)
     }
 }
 
-ExpInitializer *VarDeclaration::getExpInitializer()
-{
-    ExpInitializer *ei;
-
-    if (init)
-	ei = init->isExpInitializer();
-    else
-    {
-	Expression *e = type->defaultInit();
-	if (e)
-	    ei = new ExpInitializer(loc, e);
-	else
-	    ei = NULL;
-    }
-    return ei;
-}
-
 void VarDeclaration::semantic2(Scope *sc)
 {
     //printf("VarDeclaration::semantic2('%s')\n", toChars());
@@ -1075,17 +1122,28 @@ Dsymbol *VarDeclaration::toAlias()
 
 void VarDeclaration::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 {
-    if (storage_class & STCconst)
-	buf->writestring("const ");
+    if (storage_class & STCmanifest)
+	buf->writestring("manifest ");
     if (storage_class & STCstatic)
 	buf->writestring("static ");
+    if (storage_class & STCtls)
+	buf->writestring("__tls ");
+    if (storage_class & STCconst)
+	buf->writestring("const ");
+    if (storage_class & STCinvariant)
+	buf->writestring("invariant ");
+
     if (type)
 	type->toCBuffer(buf, ident, hgs);
     else
 	buf->writestring(ident->toChars());
     if (init)
     {	buf->writestring(" = ");
-	init->toCBuffer(buf, hgs);
+	ExpInitializer *ie = init->isExpInitializer();
+	if (ie && (ie->exp->op == TOKconstruct || ie->exp->op == TOKblit))
+	    ((AssignExp *)ie->exp)->e2->toCBuffer(buf, hgs);
+	else
+	    init->toCBuffer(buf, hgs);
     }
     buf->writeByte(';');
     buf->writenl();
@@ -1099,7 +1157,8 @@ int VarDeclaration::needThis()
 
 int VarDeclaration::isImportedSymbol()
 {
-    if (protection == PROTexport && !init && (isStatic() || isConst() || parent->isModule()))
+    if (protection == PROTexport && !init &&
+	(storage_class & STCstatic || parent->isModule()))
 	return TRUE;
     return FALSE;
 }
@@ -1111,30 +1170,117 @@ void VarDeclaration::checkCtorConstInit()
 }
 
 /************************************
- * Check to see if variable is a reference to an enclosing function
- * or not.
+ * Check to see if this variable is actually in an enclosing function
+ * rather than the current one.
  */
 
 void VarDeclaration::checkNestedReference(Scope *sc, Loc loc)
 {
-    if (!isDataseg() && parent != sc->parent && parent)
+    if (parent && !isDataseg() && parent != sc->parent &&
+	!(storage_class & STCmanifest))
     {
+	// The function that this variable is in
 	FuncDeclaration *fdv = toParent()->isFuncDeclaration();
+	// The current function
 	FuncDeclaration *fdthis = sc->parent->isFuncDeclaration();
 
-	if (fdv && fdthis)
+	if (fdv && fdthis && fdv != fdthis)
 	{
 	    if (loc.filename)
 		fdthis->getLevel(loc, fdv);
-	    nestedref = 1;
-	    fdv->nestedFrameRef = 1;
+
+	    for (int i = 0; i < nestedrefs.dim; i++)
+	    {	FuncDeclaration *f = (FuncDeclaration *)nestedrefs.data[i];
+		if (f == fdthis)
+		    goto L1;
+	    }
+	    nestedrefs.push(fdthis);
+	  L1: ;
+
+
+	    for (int i = 0; i < fdv->closureVars.dim; i++)
+	    {	Dsymbol *s = (Dsymbol *)fdv->closureVars.data[i];
+		if (s == this)
+		    goto L2;
+	    }
+
+	    fdv->closureVars.push(this);
+	  L2: ;
+
 	    //printf("var %s in function %s is nested ref\n", toChars(), fdv->toChars());
 	}
     }
 }
 
+/****************************
+ * Get ExpInitializer for a variable, if there is one.
+ */
+
+ExpInitializer *VarDeclaration::getExpInitializer()
+{
+    ExpInitializer *ei;
+
+    if (init)
+	ei = init->isExpInitializer();
+    else
+    {
+	Expression *e = type->defaultInit();
+	if (e)
+	    ei = new ExpInitializer(loc, e);
+	else
+	    ei = NULL;
+    }
+    return ei;
+}
+
+/*******************************************
+ * If variable has a constant expression initializer, get it.
+ * Otherwise, return NULL.
+ */
+
+Expression *VarDeclaration::getConstInitializer()
+{
+    if ((isConst() || isInvariant() || storage_class & STCmanifest) &&
+	storage_class & STCinit)
+    {
+	ExpInitializer *ei = getExpInitializer();
+	if (ei)
+	    return ei->exp;
+    }
+
+    return NULL;
+}
+
+/*************************************
+ * Return !=0 if we can take the address of this variable.
+ */
+
+int VarDeclaration::canTakeAddressOf()
+{
+#if 0
+    /* Global variables and struct/class fields of the form:
+     *	const int x = 3;
+     * are not stored and hence cannot have their address taken.
+     */
+    if ((isConst() || isInvariant()) &&
+	storage_class & STCinit &&
+	(!(storage_class & (STCstatic | STCextern)) || (storage_class & STCfield)) &&
+	(!parent || toParent()->isModule() || toParent()->isTemplateInstance()) &&
+	type->toBasetype()->isTypeBasic()
+       )
+    {
+	return 0;
+    }
+#else
+    if (storage_class & STCmanifest)
+	return 0;
+#endif
+    return 1;
+}
+
 /*******************************
  * Does symbol go into data segment?
+ * Includes extern variables.
  */
 
 int VarDeclaration::isDataseg()
@@ -1144,31 +1290,111 @@ int VarDeclaration::isDataseg()
     printf("%x, %p, %p\n", storage_class & (STCstatic | STCconst), parent->isModule(), parent->isTemplateInstance());
     printf("parent = '%s'\n", parent->toChars());
 #endif
+    if (storage_class & STCmanifest)
+	return 0;
     Dsymbol *parent = this->toParent();
-    if (!parent && !(storage_class & (STCstatic | STCconst)))
+    if (!parent && !(storage_class & STCstatic))
     {	error("forward referenced");
 	type = Type::terror;
 	return 0;
     }
-    return (storage_class & (STCstatic | STCconst) ||
-	   parent->isModule() ||
-	   parent->isTemplateInstance());
+    return canTakeAddressOf() &&
+	(storage_class & (STCstatic | STCextern | STCtls) ||
+	 toParent()->isModule() ||
+	 toParent()->isTemplateInstance());
 }
 
 int VarDeclaration::hasPointers()
 {
+    //printf("VarDeclaration::hasPointers() %s, ty = %d\n", toChars(), type->ty);
     return (!isDataseg() && type->hasPointers());
 }
 
+/******************************************
+ * Return TRUE if variable needs to call the destructor.
+ */
+
+int VarDeclaration::needsAutoDtor()
+{
+    //printf("VarDeclaration::needsAutoDtor() %s\n", toChars());
+
+    if (noauto || storage_class & STCnodtor)
+	return FALSE;
+
+    // Destructors for structs and arrays of structs
+    Type *tv = type->toBasetype();
+    while (tv->ty == Tsarray)
+    {   TypeSArray *ta = (TypeSArray *)tv;
+	tv = tv->nextOf()->toBasetype();
+    }
+    if (tv->ty == Tstruct)
+    {   TypeStruct *ts = (TypeStruct *)tv;
+	StructDeclaration *sd = ts->sym;
+	if (sd->dtor)
+	    return TRUE;
+    }
+
+    // Destructors for classes
+    if (storage_class & (STCauto | STCscope))
+    {
+	if (type->isClassHandle())
+	    return TRUE;
+    }
+    return FALSE;
+}
+
+
 /******************************************
  * If a variable has an auto destructor call, return call for it.
  * Otherwise, return NULL.
  */
 
-Expression *VarDeclaration::callAutoDtor()
+Expression *VarDeclaration::callAutoDtor(Scope *sc)
 {   Expression *e = NULL;
 
-    if (storage_class & (STCauto | STCscope) && !noauto)
+    //printf("VarDeclaration::callAutoDtor() %s\n", toChars());
+
+    if (noauto || storage_class & STCnodtor)
+	return NULL;
+
+    // Destructors for structs and arrays of structs
+    bool array = false;
+    Type *tv = type->toBasetype();
+    while (tv->ty == Tsarray)
+    {   TypeSArray *ta = (TypeSArray *)tv;
+	array = true;
+	tv = tv->nextOf()->toBasetype();
+    }
+    if (tv->ty == Tstruct)
+    {   TypeStruct *ts = (TypeStruct *)tv;
+	StructDeclaration *sd = ts->sym;
+	if (sd->dtor)
+	{
+	    if (array)
+	    {
+		// Typeinfo.destroy(cast(void*)&v);
+		Expression *ea = new SymOffExp(loc, this, 0, 0);
+		ea = new CastExp(loc, ea, Type::tvoid->pointerTo());
+		Expressions *args = new Expressions();
+		args->push(ea);
+
+		Expression *et = type->getTypeInfo(sc);
+		et = new DotIdExp(loc, et, Id::destroy);
+
+		e = new CallExp(loc, et, args);
+	    }
+	    else
+	    {
+		e = new VarExp(loc, this);
+		e = new DotVarExp(loc, e, sd->dtor, 0);
+		e = new CallExp(loc, e);
+	    }
+	    return e;
+	}
+    }
+
+    // Destructors for classes
+    if (storage_class & (STCauto | STCscope))
     {
 	for (ClassDeclaration *cd = type->isClassHandle();
 	     cd;
@@ -1178,6 +1404,8 @@ Expression *VarDeclaration::callAutoDtor()
 	     * classes to determine if there's no way the monitor
 	     * could be set.
 	     */
+	    //if (cd->isInterfaceDeclaration())
+		//error("interface %s cannot be scope", cd->toChars());
 	    if (1 || onstack || cd->dtors.dim)	// if any destructors
 	    {
 		// delete this;
@@ -1256,17 +1484,21 @@ void TypeInfoDeclaration::semantic(Scope *sc)
 
 /***************************** TypeInfoConstDeclaration **********************/
 
+#if V2
 TypeInfoConstDeclaration::TypeInfoConstDeclaration(Type *tinfo)
     : TypeInfoDeclaration(tinfo, 0)
 {
 }
+#endif
 
 /***************************** TypeInfoInvariantDeclaration **********************/
 
+#if V2
 TypeInfoInvariantDeclaration::TypeInfoInvariantDeclaration(Type *tinfo)
     : TypeInfoDeclaration(tinfo, 0)
 {
 }
+#endif
 
 /***************************** TypeInfoStructDeclaration **********************/
 
