@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -37,6 +37,7 @@ struct ExpInitializer;
 struct StructDeclaration;
 struct TupleType;
 struct InterState;
+struct IRState;
 
 enum PROT;
 enum LINK;
@@ -68,6 +69,12 @@ enum STC
     STCscope	    = 0x80000,		// template parameter
     STCinvariant    = 0x100000,
     STCref	    = 0x200000,
+    STCinit	    = 0x400000,		// has explicit initializer
+    STCmanifest	    = 0x800000,		// manifest constant
+    STCnodtor	    = 0x1000000,	// don't run destructor
+    STCnothrow	    = 0x2000000,	// never throws exceptions
+    STCpure	    = 0x4000000,	// pure function
+    STCtls	    = 0x8000000,	// thread local
 };
 
 struct Match
@@ -79,7 +86,8 @@ struct Match
     FuncDeclaration *anyf;	// pick a func, any func, to use for error recovery
 };
 
-void overloadResolveX(Match *m, FuncDeclaration *f, Expressions *arguments);
+void overloadResolveX(Match *m, FuncDeclaration *f,
+	Expression *ethis, Expressions *arguments);
 int overloadApply(FuncDeclaration *fstart,
 	int (*fp)(void *, FuncDeclaration *),
 	void *param);
@@ -89,6 +97,7 @@ int overloadApply(FuncDeclaration *fstart,
 struct Declaration : Dsymbol
 {
     Type *type;
+    Type *originalType;		// before semantic analysis
     unsigned storage_class;
     enum PROT protection;
     enum LINK linkage;
@@ -98,7 +107,7 @@ struct Declaration : Dsymbol
     void semantic(Scope *sc);
     char *kind();
     target_size_t size(Loc loc);
-    void checkModify(Loc loc, Scope *sc);
+    void checkModify(Loc loc, Scope *sc, Type *t);
 
     void emitComment(Scope *sc);
     void toDocBuffer(OutBuffer *buf);
@@ -176,7 +185,7 @@ struct TypedefDeclaration : Declaration
 
     void toDocBuffer(OutBuffer *buf);
 
-    void toObjFile();			// compile to .obj file
+    void toObjFile(int multiobj);			// compile to .obj file
     void toDebug();
     int cvMember(unsigned char *p);
 
@@ -220,7 +229,7 @@ struct VarDeclaration : Declaration
     Initializer *init;
     target_size_t offset;
     int noauto;			// no auto semantics
-    int nestedref;		// referenced by a lexically nested function
+    FuncDeclarations nestedrefs; // referenced by these lexically nested functions
     int inuse;
     int ctorinit;		// it has been initialized in a ctor
     int onstack;		// 1: it has been allocated on the stack
@@ -229,6 +238,7 @@ struct VarDeclaration : Declaration
     Dsymbol *aliassym;		// if redone as alias to another symbol
     Expression *value;		// when interpreting, this is the value
 				// (NULL if value not determinable)
+    Scope *scope;		// !=NULL means context to use
 
     VarDeclaration(Loc loc, Type *t, Identifier *id, Initializer *init);
     Dsymbol *syntaxCopy(Dsymbol *);
@@ -244,20 +254,24 @@ struct VarDeclaration : Declaration
     int isImportedSymbol();
     int isDataseg();
     int hasPointers();
-    Expression *callAutoDtor();
+    int canTakeAddressOf();
+    int needsAutoDtor();
+    Expression *callAutoDtor(Scope *sc);
     ExpInitializer *getExpInitializer();
+    Expression *getConstInitializer();
     void checkCtorConstInit();
     void checkNestedReference(Scope *sc, Loc loc);
     Dsymbol *toAlias();
 
     Symbol *toSymbol();
-    void toObjFile();			// compile to .obj file
+    void toObjFile(int multiobj);			// compile to .obj file
     int cvMember(unsigned char *p);
 
     // Eliminate need for dynamic_cast
     VarDeclaration *isVarDeclaration() { return (VarDeclaration *)this; }
 };
 
+
 /**************************************************************/
 
 // This is a shell around a back end symbol
@@ -312,7 +326,7 @@ struct TypeInfoDeclaration : VarDeclaration
     void emitComment(Scope *sc);
 
     Symbol *toSymbol();
-    void toObjFile();			// compile to .obj file
+    void toObjFile(int multiobj);			// compile to .obj file
     virtual void toDt(dt_t **pdt);
 };
 
@@ -432,6 +446,22 @@ enum ILS
 };
 
 /**************************************************************/
+#if V2
+
+enum BUILTIN
+{
+    BUILTINunknown = -1,	// not known if this is a builtin
+    BUILTINnot,			// this is not a builtin
+    BUILTINsin,			// std.math.sin
+    BUILTINcos,			// std.math.cos
+    BUILTINtan,			// std.math.tan
+    BUILTINsqrt,		// std.math.sqrt
+    BUILTINfabs,		// std.math.fabs
+};
+
+Expression *eval_builtin(enum BUILTIN builtin, Expressions *arguments);
+
+#endif
 
 struct FuncDeclaration : Declaration
 {
@@ -463,7 +493,7 @@ struct FuncDeclaration : Declaration
     int inlineNest;			// !=0 if nested inline
     int cantInterpret;			// !=0 if cannot interpret function
     int semanticRun;			// !=0 if semantic3() had been run
-    int nestedFrameRef;			// !=0 if nested variables referenced frame ptr
+					// this function's frame ptr
     ForeachStatement *fes;		// if foreach body, this is the foreach
     int introducing;			// !=0 if 'introducing' function
     Type *tintro;			// if !=NULL, then this is the type
@@ -483,6 +513,20 @@ struct FuncDeclaration : Declaration
     VarDeclaration *nrvo_var;		// variable to replace with shidden
     Symbol *shidden;			// hidden pointer passed to function
 
+#if V2
+    enum BUILTIN builtin;		// set if this is a known, builtin
+					// function we can evaluate at compile
+					// time
+
+    int tookAddressOf;			// set if someone took the address of
+					// this function
+    Dsymbols closureVars;		// local variables in this function
+					// which are referenced by nested
+					// functions
+#else
+    int nestedFrameRef;			// !=0 if nested variables referenced
+#endif
+
     FuncDeclaration(Loc loc, Loc endloc, Identifier *id, enum STC storage_class, Type *type);
     Dsymbol *syntaxCopy(Dsymbol *);
     void semantic(Scope *sc);
@@ -491,9 +535,11 @@ struct FuncDeclaration : Declaration
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     void bodyToCBuffer(OutBuffer *buf, HdrGenState *hgs);
     int overrides(FuncDeclaration *fd);
+    int findVtblIndex(Array *vtbl, int dim);
     int overloadInsert(Dsymbol *s);
     FuncDeclaration *overloadExactMatch(Type *t);
-    FuncDeclaration *overloadResolve(Loc loc, Expressions *arguments);
+    FuncDeclaration *overloadResolve(Loc loc, Expression *ethis, Expressions *arguments, int flags = 0);
+    MATCH leastAsSpecialized(FuncDeclaration *g);
     LabelDsymbol *searchLabel(Identifier *ident);
     AggregateDeclaration *isThis();
     AggregateDeclaration *isMember2();
@@ -504,10 +550,12 @@ struct FuncDeclaration : Declaration
     int isMain();
     int isWinMain();
     int isDllMain();
+    enum BUILTIN isBuiltin();
     int isExport();
     int isImportedSymbol();
     int isAbstract();
     int isCodeseg();
+    int isOverloadable();
     virtual int isNested();
     int needThis();
     virtual int isVirtual();
@@ -521,6 +569,7 @@ struct FuncDeclaration : Declaration
     char *kind();
     void toDocBuffer(OutBuffer *buf);
     FuncDeclaration *isUnique();
+    int needsClosure();
 
     static FuncDeclaration *genCfunc(Type *treturn, char *name,
 	Type *t1 = 0, Type *t2 = 0, Type *t3 = 0);
@@ -529,8 +578,9 @@ struct FuncDeclaration : Declaration
 
     Symbol *toSymbol();
     Symbol *toThunkSymbol(target_ptrdiff_t offset);	// thunk version
-    void toObjFile();			// compile to .obj file
+    void toObjFile(int multiobj);			// compile to .obj file
     int cvMember(unsigned char *p);
+    void buildClosure(IRState *irs);
 
     FuncDeclaration *isFuncDeclaration() { return this; }
 };
@@ -578,9 +628,26 @@ struct CtorDeclaration : FuncDeclaration
     CtorDeclaration *isCtorDeclaration() { return this; }
 };
 
+struct PostBlitDeclaration : FuncDeclaration
+{
+    PostBlitDeclaration(Loc loc, Loc endloc);
+    PostBlitDeclaration(Loc loc, Loc endloc, Identifier *id);
+    Dsymbol *syntaxCopy(Dsymbol *);
+    void semantic(Scope *sc);
+    void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
+    int isVirtual();
+    int addPreInvariant();
+    int addPostInvariant();
+    int overloadInsert(Dsymbol *s);
+    void emitComment(Scope *sc);
+
+    PostBlitDeclaration *isPostBlitDeclaration() { return this; }
+};
+
 struct DtorDeclaration : FuncDeclaration
 {
     DtorDeclaration(Loc loc, Loc endloc);
+    DtorDeclaration(Loc loc, Loc endloc, Identifier *id);
     Dsymbol *syntaxCopy(Dsymbol *);
     void semantic(Scope *sc);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
