@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -28,6 +28,7 @@
 #include "init.h"
 #include "import.h"
 #include "template.h"
+#include "attrib.h"
 
 /****************************** Dsymbol ******************************/
 
@@ -382,6 +383,11 @@ int Dsymbol::isDeprecated()
     return FALSE;
 }
 
+int Dsymbol::isOverloadable()
+{
+    return 0;
+}
+
 LabelDsymbol *Dsymbol::isLabel()		// is this a LabelDsymbol()?
 {
     return NULL;
@@ -389,7 +395,9 @@ LabelDsymbol *Dsymbol::isLabel()		// is this a LabelDsymbol()?
 
 AggregateDeclaration *Dsymbol::isMember()	// is this a member of an AggregateDeclaration?
 {
+    //printf("Dsymbol::isMember() %s\n", toChars());
     Dsymbol *parent = toParent();
+    //printf("parent is %s %s\n", parent->kind(), parent->toChars());
     return parent ? parent->isAggregateDeclaration() : NULL;
 }
 
@@ -582,6 +590,23 @@ void Dsymbol::addComment(unsigned char *comment)
 #endif
 }
 
+/********************************* OverloadSet ****************************/
+
+OverloadSet::OverloadSet()
+    : Dsymbol()
+{
+}
+
+void OverloadSet::push(Dsymbol *s)
+{
+    a.push(s);
+}
+
+char *OverloadSet::kind()
+{
+    return "overloadset";
+}
+
 
 /********************************* ScopeDsymbol ****************************/
 
@@ -617,20 +642,21 @@ Dsymbol *ScopeDsymbol::syntaxCopy(Dsymbol *s)
 }
 
 Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
-{   Dsymbol *s;
-    int i;
-
+{
     //printf("%s->ScopeDsymbol::search(ident='%s', flags=x%x)\n", toChars(), ident->toChars(), flags);
+
     // Look in symbols declared in this module
-    s = symtab ? symtab->lookup(ident) : NULL;
+    Dsymbol *s = symtab ? symtab->lookup(ident) : NULL;
     if (s)
     {
 	//printf("\ts = '%s.%s'\n",toChars(),s->toChars());
     }
     else if (imports)
     {
+	OverloadSet *a = NULL;
+
 	// Look in imported modules
-	for (i = 0; i < imports->dim; i++)
+	for (int i = 0; i < imports->dim; i++)
 	{   ScopeDsymbol *ss = (ScopeDsymbol *)imports->data[i];
 	    Dsymbol *s2;
 
@@ -639,6 +665,8 @@ Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
 		continue;
 
 	    //printf("\tscanning import '%s', prots = %d, isModule = %p, isImport = %p\n", ss->toChars(), prots[i], ss->isModule(), ss->isImport());
+	    /* Don't find private members if ss is a module
+	     */
 	    s2 = ss->search(loc, ident, ss->isModule() ? 1 : 0);
 	    if (!s)
 		s = s2;
@@ -646,6 +674,10 @@ Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
 	    {
 		if (s->toAlias() == s2->toAlias())
 		{
+		    /* After following aliases, we found the same symbol,
+		     * so it's not an ambiguity.
+		     * But if one alias is deprecated, prefer the other.
+		     */
 		    if (s->isDeprecated())
 			s = s2;
 		}
@@ -664,7 +696,28 @@ Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
 			 )
 		       )
 		    {
-			if (flags & 4)
+			/* If both s2 and s are overloadable (though we only
+			 * need to check s once)
+			 */
+			if (s2->isOverloadable() && (a || s->isOverloadable()))
+			{   if (!a)
+				a = new OverloadSet();
+			    /* Don't add to a[] if s2 is alias of previous sym
+			     */
+			    for (int j = 0; j < a->a.dim; j++)
+			    {	Dsymbol *s3 = (Dsymbol *)a->a.data[j];
+				if (s2->toAlias() == s3->toAlias())
+				{
+				    if (s3->isDeprecated())
+					a->a.data[j] = (void *)s2;
+				    goto Lcontinue;
+				}
+			    }
+			    a->push(s2);
+			Lcontinue:
+			    continue;
+			}
+			if (flags & 4)		// if return NULL on ambiguity
 			    return NULL;
 			if (!(flags & 2))
 			    ss->multiplyDefined(loc, s, s2);
@@ -673,10 +726,20 @@ Dsymbol *ScopeDsymbol::search(Loc loc, Identifier *ident, int flags)
 		}
 	    }
 	}
+
+	/* Build special symbol if we had multiple finds
+	 */
+	if (a)
+	{   assert(s);
+	    a->push(s);
+	    s = a;
+	}
+
 	if (s)
 	{
 	    Declaration *d = s->isDeclaration();
-	    if (d && d->protection == PROTprivate && !d->parent->isTemplateMixin() &&
+	    if (d && d->protection == PROTprivate &&
+		!d->parent->isTemplateMixin() &&
 		!(flags & 2))
 		error("%s is private", d->toPrettyChars());
 	}
@@ -748,6 +811,7 @@ void ScopeDsymbol::multiplyDefined(Loc loc, Dsymbol *s1, Dsymbol *s2)
 	    s2->toPrettyChars(),
 	    s2->locToChars());
     }
+halt();
 }
 
 Dsymbol *ScopeDsymbol::nameCollision(Dsymbol *s)
@@ -778,6 +842,65 @@ char *ScopeDsymbol::kind()
 }
 
 
+/***************************************
+ * Determine number of Dsymbols, folding in AttribDeclaration members.
+ */
+
+size_t ScopeDsymbol::dim(Array *members)
+{
+    size_t n = 0;
+    if (members)
+    {
+	for (size_t i = 0; i < members->dim; i++)
+	{   Dsymbol *s = (Dsymbol *)members->data[i];
+	    AttribDeclaration *a = s->isAttribDeclaration();
+
+	    if (a)
+	    {
+		n += dim(a->decl);
+	    }
+	    else
+		n++;
+	}
+    }
+    return n;
+}
+
+/***************************************
+ * Get nth Dsymbol, folding in AttribDeclaration members.
+ * Returns:
+ *	Dsymbol*	nth Dsymbol
+ *	NULL		not found, *pn gets incremented by the number
+ *			of Dsymbols
+ */
+
+Dsymbol *ScopeDsymbol::getNth(Array *members, size_t nth, size_t *pn)
+{
+    if (!members)
+	return NULL;
+
+    size_t n = 0;
+    for (size_t i = 0; i < members->dim; i++)
+    {   Dsymbol *s = (Dsymbol *)members->data[i];
+	AttribDeclaration *a = s->isAttribDeclaration();
+
+	if (a)
+	{
+	    s = getNth(a->decl, nth - n, &n);
+	    if (s)
+		return s;
+	}
+	else if (n == nth)
+	    return s;
+	else
+	    n++;
+    }
+
+    if (pn)
+	*pn += n;
+    return NULL;
+}
+
 /*******************************************
  * Look for member of the form:
  *	const(MemberInfo)[] getMembers(string);
@@ -831,29 +954,32 @@ Dsymbol *WithScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 
 /****************************** ArrayScopeSymbol ******************************/
 
-ArrayScopeSymbol::ArrayScopeSymbol(Expression *e)
+ArrayScopeSymbol::ArrayScopeSymbol(Scope *sc, Expression *e)
     : ScopeDsymbol()
 {
     assert(e->op == TOKindex || e->op == TOKslice);
     exp = e;
     type = NULL;
     td = NULL;
+    this->sc = sc;
 }
 
-ArrayScopeSymbol::ArrayScopeSymbol(TypeTuple *t)
+ArrayScopeSymbol::ArrayScopeSymbol(Scope *sc, TypeTuple *t)
     : ScopeDsymbol()
 {
     exp = NULL;
     type = t;
     td = NULL;
+    this->sc = sc;
 }
 
-ArrayScopeSymbol::ArrayScopeSymbol(TupleDeclaration *s)
+ArrayScopeSymbol::ArrayScopeSymbol(Scope *sc, TupleDeclaration *s)
     : ScopeDsymbol()
 {
     exp = NULL;
     type = NULL;
     td = s;
+    this->sc = sc;
 }
 
 Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
@@ -870,7 +996,8 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 	    VarDeclaration *v = new VarDeclaration(loc, Type::tsize_t, Id::dollar, NULL);
 	    Expression *e = new IntegerExp(0, td->objects->dim, Type::tsize_t);
 	    v->init = new ExpInitializer(0, e);
-	    v->storage_class |= STCconst;
+	    v->storage_class |= STCstatic | STCconst;
+	    v->semantic(sc);
 	    return v;
 	}
 
@@ -879,7 +1006,8 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 	    VarDeclaration *v = new VarDeclaration(loc, Type::tsize_t, Id::dollar, NULL);
 	    Expression *e = new IntegerExp(0, type->arguments->dim, Type::tsize_t);
 	    v->init = new ExpInitializer(0, e);
-	    v->storage_class |= STCconst;
+	    v->storage_class |= STCstatic | STCconst;
+	    v->semantic(sc);
 	    return v;
 	}
 
@@ -924,7 +1052,7 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 		 */
 		Expression *e = new IntegerExp(0, ((StringExp *)ce)->len, Type::tsize_t);
 		v->init = new ExpInitializer(0, e);
-		v->storage_class |= STCconst;
+		v->storage_class |= STCstatic | STCconst;
 	    }
 	    else if (ce->op == TOKarrayliteral)
 	    {	/* It is for an array literal, so the
@@ -932,7 +1060,7 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 		 */
 		Expression *e = new IntegerExp(0, ((ArrayLiteralExp *)ce)->elements->dim, Type::tsize_t);
 		v->init = new ExpInitializer(0, e);
-		v->storage_class |= STCconst;
+		v->storage_class |= STCstatic | STCconst;
 	    }
 	    else if (ce->op == TOKtuple)
 	    {	/* It is for an expression tuple, so the
@@ -940,10 +1068,11 @@ Dsymbol *ArrayScopeSymbol::search(Loc loc, Identifier *ident, int flags)
 		 */
 		Expression *e = new IntegerExp(0, ((TupleExp *)ce)->exps->dim, Type::tsize_t);
 		v->init = new ExpInitializer(0, e);
-		v->storage_class |= STCconst;
+		v->storage_class |= STCstatic | STCconst;
 	    }
 	    *pvar = v;
 	}
+	(*pvar)->semantic(sc);
 	return (*pvar);
     }
     return NULL;
