@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -49,6 +49,7 @@ struct HdrGenState;
 struct BinExp;
 struct InterState;
 struct Symbol;		// back end symbol
+struct OverloadSet;
 
 enum TOK;
 
@@ -66,6 +67,7 @@ void initPrecedence();
 
 Expression *resolveProperties(Scope *sc, Expression *e);
 void accessCheck(Loc loc, Scope *sc, Expression *e, Declaration *d);
+Expression *build_overload(Loc loc, Scope *sc, Expression *ethis, Expression *earg, Identifier *id);
 Dsymbol *search_function(ScopeDsymbol *ad, Identifier *funcid);
 void inferApplyArgTypes(enum TOK op, Arguments *arguments, Expression *aggr);
 void argExpTypesToCBuffer(OutBuffer *buf, Expressions *arguments, HdrGenState *hgs);
@@ -73,6 +75,7 @@ void argsToCBuffer(OutBuffer *buf, Expressions *arguments, HdrGenState *hgs);
 void expandTuples(Expressions *exps);
 FuncDeclaration *hasThis(Scope *sc);
 Expression *fromConstInitializer(int result, Expression *e);
+int arrayExpressionCanThrow(Expressions *exps);
 
 struct Expression : Object
 {
@@ -113,7 +116,7 @@ struct Expression : Object
     void checkScalar();
     void checkNoBool();
     Expression *checkIntegral();
-    void checkArithmetic();
+    Expression *checkArithmetic();
     void checkDeprecated(Scope *sc, Dsymbol *s);
     virtual Expression *checkToBoolean();
     Expression *checkToPointer();
@@ -135,6 +138,7 @@ struct Expression : Object
     virtual int isBool(int result);
     virtual int isBit();
     virtual int checkSideEffect(int flag);
+    virtual int canThrow();
 
     virtual int inlineCost(InlineCostState *ics);
     virtual Expression *doInline(InlineDoState *ids);
@@ -246,8 +250,9 @@ struct DollarExp : IdentifierExp
 struct DsymbolExp : Expression
 {
     Dsymbol *s;
+    int hasOverloads;
 
-    DsymbolExp(Loc loc, Dsymbol *s);
+    DsymbolExp(Loc loc, Dsymbol *s, int hasOverloads = 0);
     Expression *semantic(Scope *sc);
     char *toChars();
     void dump(int indent);
@@ -322,6 +327,7 @@ struct StringExp : Expression
     Expression *castTo(Scope *sc, Type *t);
     int compare(Object *obj);
     int isBool(int result);
+    unsigned charAt(size_t i);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     void toMangleBuffer(OutBuffer *buf);
     elem *toElem(IRState *irs);
@@ -346,6 +352,8 @@ struct TupleExp : Expression
     Expression *optimize(int result);
     Expression *interpret(InterState *istate);
     Expression *castTo(Scope *sc, Type *t);
+    elem *toElem(IRState *irs);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -372,6 +380,7 @@ struct ArrayLiteralExp : Expression
     MATCH implicitConvTo(Type *t);
     Expression *castTo(Scope *sc, Type *t);
     dt_t **toDt(dt_t **pdt);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -397,6 +406,7 @@ struct AssocArrayLiteralExp : Expression
     Expression *interpret(InterState *istate);
     MATCH implicitConvTo(Type *t);
     Expression *castTo(Scope *sc, Type *t);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -428,6 +438,8 @@ struct StructLiteralExp : Expression
     Expression *interpret(InterState *istate);
     dt_t **toDt(dt_t **pdt);
     Expression *toLvalue(Scope *sc, Expression *e);
+    int canThrow();
+    MATCH implicitConvTo(Type *t);
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -495,6 +507,7 @@ struct NewExp : Expression
     int checkSideEffect(int flag);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     void scanForNestedRef(Scope *sc);
+    int canThrow();
 
     //int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -516,15 +529,24 @@ struct NewAnonClassExp : Expression
     Expression *semantic(Scope *sc);
     int checkSideEffect(int flag);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
+    int canThrow();
+};
+
+struct SymbolExp : Expression
+{
+    Declaration *var;
+    int hasOverloads;
+
+    SymbolExp(Loc loc, enum TOK op, int size, Declaration *var, int hasOverloads);
+
+    elem *toElem(IRState *irs);
 };
 
 // Offset from symbol
 
-struct SymOffExp : Expression
+struct SymOffExp : SymbolExp
 {
-    Declaration *var;
     target_size_t offset;
-    int hasOverloads;
 
     SymOffExp(Loc loc, Declaration *var, target_size_t offset, int hasOverloads = 0);
     Expression *semantic(Scope *sc);
@@ -537,17 +559,13 @@ struct SymOffExp : Expression
     Expression *castTo(Scope *sc, Type *t);
     void scanForNestedRef(Scope *sc);
 
-    elem *toElem(IRState *irs);
     dt_t **toDt(dt_t **pdt);
 };
 
 // Variable
 
-struct VarExp : Expression
+struct VarExp : SymbolExp
 {
-    Declaration *var;
-    int hasOverloads;
-
     VarExp(Loc loc, Declaration *var, int hasOverloads = 0);
     int equals(Object *o);
     Expression *semantic(Scope *sc);
@@ -559,7 +577,6 @@ struct VarExp : Expression
     void checkEscape();
     Expression *toLvalue(Scope *sc, Expression *e);
     Expression *modifiableLvalue(Scope *sc, Expression *e);
-    elem *toElem(IRState *irs);
     dt_t **toDt(dt_t **pdt);
     void scanForNestedRef(Scope *sc);
 
@@ -568,6 +585,17 @@ struct VarExp : Expression
     //Expression *inlineScan(InlineScanState *iss);
 };
 
+// Overload Set
+
+struct OverExp : Expression
+{
+    OverloadSet *vars;
+
+    OverExp(OverloadSet *s);
+    Expression *toLvalue(Scope *sc, Expression *e);
+};
+
+
 // Function/Delegate literal
 
 struct FuncExp : Expression
@@ -601,6 +629,7 @@ struct DeclarationExp : Expression
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     elem *toElem(IRState *irs);
     void scanForNestedRef(Scope *sc);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -671,6 +700,7 @@ struct UnaExp : Expression
     void dump(int indent);
     void scanForNestedRef(Scope *sc);
     Expression *interpretCommon(InterState *istate, Expression *(*fp)(Type *, Expression *));
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -702,6 +732,7 @@ struct BinExp : Expression
     Expression *interpretCommon(InterState *istate, Expression *(*fp)(Type *, Expression *, Expression *));
     Expression *interpretCommon2(InterState *istate, Expression *(*fp)(TOK, Type *, Expression *, Expression *));
     Expression *interpretAssignCommon(InterState *istate, Expression *(*fp)(Type *, Expression *, Expression *), int post = 0);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -743,6 +774,7 @@ struct AssertExp : UnaExp
     Expression *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
     int checkSideEffect(int flag);
+    int canThrow();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
     int inlineCost(InlineCostState *ics);
@@ -840,6 +872,7 @@ struct CallExp : UnaExp
     elem *toElem(IRState *irs);
     void scanForNestedRef(Scope *sc);
     Expression *toLvalue(Scope *sc, Expression *e);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -867,6 +900,9 @@ struct PtrExp : UnaExp
     elem *toElem(IRState *irs);
     Expression *optimize(int result);
     Expression *interpret(InterState *istate);
+
+    // For operator overloading
+    Identifier *opId();
 };
 
 struct NegExp : UnaExp
@@ -1381,6 +1417,7 @@ struct CondExp : BinExp
     MATCH implicitConvTo(Type *t);
     Expression *castTo(Scope *sc, Type *t);
     void scanForNestedRef(Scope *sc);
+    int canThrow();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -1390,6 +1427,31 @@ struct CondExp : BinExp
 };
 
 
+/****************************************************************/
+
+struct DefaultInitExp : Expression
+{
+    enum TOK subop;		// which of the derived classes this is
+
+    DefaultInitExp(Loc loc, enum TOK subop, int size);
+    virtual Expression *resolve(Loc loc, Scope *sc) = 0;
+    void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
+};
+
+struct FileInitExp : DefaultInitExp
+{
+    FileInitExp(Loc loc);
+    Expression *semantic(Scope *sc);
+    Expression *resolve(Loc loc, Scope *sc);
+};
+
+struct LineInitExp : DefaultInitExp
+{
+    LineInitExp(Loc loc);
+    Expression *semantic(Scope *sc);
+    Expression *resolve(Loc loc, Scope *sc);
+};
+
 /****************************************************************/
 
 /* Special values used by the interpreter
