@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -41,6 +41,7 @@
 FuncDeclaration::FuncDeclaration(Loc loc, Loc endloc, Identifier *id, enum STC storage_class, Type *type)
     : Declaration(id)
 {
+    //printf("FuncDeclaration(id = '%s', type = %p)\n", id->toChars(), type);
     this->storage_class = storage_class;
     this->type = type;
     this->loc = loc;
@@ -70,7 +71,6 @@ FuncDeclaration::FuncDeclaration(Loc loc, Loc endloc, Identifier *id, enum STC s
     inlineAsm = 0;
     cantInterpret = 0;
     semanticRun = 0;
-    nestedFrameRef = 0;
     fes = NULL;
     introducing = 0;
     tintro = NULL;
@@ -80,6 +80,8 @@ FuncDeclaration::FuncDeclaration(Loc loc, Loc endloc, Identifier *id, enum STC s
     nrvo_can = 1;
     nrvo_var = NULL;
     shidden = NULL;
+    builtin = BUILTINunknown;
+    tookAddressOf = 0;
 }
 
 Dsymbol *FuncDeclaration::syntaxCopy(Dsymbol *s)
@@ -113,11 +115,51 @@ void FuncDeclaration::semantic(Scope *sc)
     if (isFuncLiteralDeclaration())
 	printf("\tFuncLiteralDeclaration()\n");
     printf("sc->parent = %s\n", sc->parent->toChars());
-    printf("type: %s\n", type->toChars());
+    printf("type: %p, %s\n", type, type->toChars());
 #endif
 
-    if (type->nextOf())
+    storage_class |= sc->stc;
+    //printf("function storage_class = x%x\n", storage_class);
+
+    if (!originalType)
+	originalType = type;
+    if (!type->deco && type->nextOf())
+    {
+#if 1
+	/* Apply const and invariant storage class
+	 * to the function type
+	 */
 	type = type->semantic(loc, sc);
+	if (storage_class & STCinvariant)
+	{   // Don't use toInvariant(), as that will do a merge()
+	    type = type->makeInvariant();
+	    type->deco = type->merge()->deco;
+	}
+	else if (storage_class & STCconst)
+	{
+	    if (!type->isInvariant())
+	    {	// Don't use toConst(), as that will do a merge()
+		type = type->makeConst();
+		type->deco = type->merge()->deco;
+	    }
+	}
+#else
+	if (storage_class & (STCconst | STCinvariant))
+	{
+	    /* Apply const and invariant storage class
+	     * to the function's return type
+	     */
+	    Type *tn = type->nextOf();
+	    if (storage_class & STCconst)
+		tn = tn->makeConst();
+	    if (storage_class & STCinvariant)
+		tn = tn->makeInvariant();
+	    ((TypeNext *)type)->next = tn;
+	}
+
+	type = type->semantic(loc, sc);
+#endif
+    }
     //type->print();
     if (type->ty != Tfunction)
     {
@@ -125,6 +167,7 @@ void FuncDeclaration::semantic(Scope *sc)
 	return;
     }
     f = (TypeFunction *)(type);
+
     size_t nparams = Argument::dim(f->parameters);
 
     linkage = sc->linkage;
@@ -134,12 +177,10 @@ void FuncDeclaration::semantic(Scope *sc)
 	parent = sc->parent;
     }
     protection = sc->protection;
-    storage_class |= sc->stc;
     if (attributes)
 	attributes->append(sc->attributes);
     else
 	attributes = sc->attributes;
-    //printf("function storage_class = x%x\n", storage_class);
     Dsymbol *parent = toParent();
 
     if (isAuto() || isScope())
@@ -147,6 +188,12 @@ void FuncDeclaration::semantic(Scope *sc)
 
     if (isAbstract() && !isVirtual())
 	error("non-virtual functions cannot be abstract");
+
+    if ((f->isConst() || f->isInvariant()) && !isThis())
+	error("without 'this' cannot be const/invariant");
+
+    if (isAbstract() && isFinal())
+	error("cannot be both final and abstract");
 #if 0
     if (isAbstract() && fbody)
 	error("abstract functions cannot have bodies");
@@ -174,8 +221,8 @@ void FuncDeclaration::semantic(Scope *sc)
     if (sd)
     {
 	// Verify no constructors, destructors, etc.
-	if (isCtorDeclaration() ||
-	    isDtorDeclaration()
+	if (isCtorDeclaration()
+	    //||isDtorDeclaration()
 	    //|| isInvariantDeclaration()
 	    //|| isUnitTestDeclaration()
 	   )
@@ -205,6 +252,7 @@ void FuncDeclaration::semantic(Scope *sc)
 	storage_class |= STCabstract;
 
 	if (isCtorDeclaration() ||
+	    isPostBlitDeclaration() ||
 	    isDtorDeclaration() ||
 	    isInvariantDeclaration() ||
 	    isUnitTestDeclaration() || isNewDeclaration() || isDelete())
@@ -268,115 +316,99 @@ void FuncDeclaration::semantic(Scope *sc)
 	}
 
 	// Find index of existing function in vtbl[] to override
-	if (cd->baseClass)
+	vi = findVtblIndex(&cd->vtbl, cd->baseClass ? cd->baseClass->vtbl.dim : 0);
+	switch (vi)
 	{
-	    for (vi = 0; vi < cd->baseClass->vtbl.dim; vi++)
-	    {
-		FuncDeclaration *fdv = ((Dsymbol *)cd->vtbl.data[vi])->isFuncDeclaration();
+	    case -1:
+		/* Didn't find one, so
+		 * This is an 'introducing' function which gets a new
+		 * slot in the vtbl[].
+		 */
 
-		// BUG: should give error if argument types match,
-		// but return type does not?
-
-		//printf("\tvtbl[%d] = '%s'\n", vi, fdv ? fdv->ident->toChars() : "");
-		if (fdv && fdv->ident == ident)
-		{
-		    int cov = type->covariant(fdv->type);
-		    //printf("\tbaseclass cov = %d\n", cov);
-		    if (cov == 2)
+		// Verify this doesn't override previous final function
+		if (cd->baseClass)
+		{   Dsymbol *s = cd->baseClass->search(loc, ident, 0);
+		    if (s)
 		    {
-			//type->print();
-			//fdv->type->print();
-			//printf("%s %s\n", type->deco, fdv->type->deco);
-			error("of type %s overrides but is not covariant with %s of type %s",
-			    type->toChars(), fdv->toPrettyChars(), fdv->type->toChars());
-		    }
-		    if (cov == 1)
-		    {
-			if (fdv->isFinal())
-			    error("cannot override final function %s", fdv->toPrettyChars());
-
-			if (!isOverride() && global.params.warnings)
-			    error("overrides base class function %s, but is not marked with 'override'", fdv->toPrettyChars());
-
-			if (fdv->toParent() == parent)
-			{
-			    // If both are mixins, then error.
-			    // If either is not, the one that is not overrides
-			    // the other.
-			    if (fdv->parent->isClassDeclaration())
-				goto L1;
-			    if (!this->parent->isClassDeclaration()
-#if !BREAKABI
-				&& !isDtorDeclaration()
-#endif
-				)
-				error("multiple overrides of same function");
-			}
-			cd->vtbl.data[vi] = (void *)this;
-			vtblIndex = vi;
-
-			/* This works by whenever this function is called,
-			 * it actually returns tintro, which gets dynamically
-			 * cast to type. But we know that tintro is a base
-			 * of type, so we could optimize it by not doing a
-			 * dynamic cast, but just subtracting the isBaseOf()
-			 * offset if the value is != null.
-			 */
-
-			if (fdv->tintro)
-			    tintro = fdv->tintro;
-			else if (!type->equals(fdv->type))
-			{
-			    /* Only need to have a tintro if the vptr
-			     * offsets differ
-			     */
-			    target_ptrdiff_t offset;
-			    if (fdv->type->nextOf()->isBaseOf(type->nextOf(), &offset))
-			    {
-				tintro = fdv->type;
-			    }
-			}
-			goto L1;
-		    }
-		    if (cov == 3)
-		    {
-			cd->sizeok = 2;	// can't finish due to forward reference
-			return;
+			FuncDeclaration *f = s->isFuncDeclaration();
+			f = f->overloadExactMatch(type);
+			if (f && f->isFinal() && f->prot() != PROTprivate)
+			    error("cannot override final function %s", f->toPrettyChars());
 		    }
 		}
+
+		if (isFinal())
+		{
+		    cd->vtblFinal.push(this);
+		}
+		else
+		{
+		    // Append to end of vtbl[]
+		    //printf("\tintroducing function\n");
+		    introducing = 1;
+		    vi = cd->vtbl.dim;
+		    cd->vtbl.push(this);
+		    vtblIndex = vi;
+		}
+
+		break;
+
+	    case -2:	// can't determine because of fwd refs
+		cd->sizeok = 2;	// can't finish due to forward reference
+		return;
+
+	    default:
+	    {   FuncDeclaration *fdv = (FuncDeclaration *)cd->vtbl.data[vi];
+		// This function is covariant with fdv
+		if (fdv->isFinal())
+		    error("cannot override final function %s", fdv->toPrettyChars());
+
+		if (!isOverride() && global.params.warnings)
+		    error("overrides base class function %s, but is not marked with 'override'", fdv->toPrettyChars());
+
+		if (fdv->toParent() == parent)
+		{
+		    // If both are mixins, then error.
+		    // If either is not, the one that is not overrides
+		    // the other.
+		    if (fdv->parent->isClassDeclaration())
+			break;
+		    if (!this->parent->isClassDeclaration()
+#if !BREAKABI
+			&& !isDtorDeclaration()
+#endif
+			&& !isPostBlitDeclaration()
+			)
+			error("multiple overrides of same function");
+		}
+		cd->vtbl.data[vi] = (void *)this;
+		vtblIndex = vi;
+
+		/* This works by whenever this function is called,
+		 * it actually returns tintro, which gets dynamically
+		 * cast to type. But we know that tintro is a base
+		 * of type, so we could optimize it by not doing a
+		 * dynamic cast, but just subtracting the isBaseOf()
+		 * offset if the value is != null.
+		 */
+
+		if (fdv->tintro)
+		    tintro = fdv->tintro;
+		else if (!type->equals(fdv->type))
+		{
+		    /* Only need to have a tintro if the vptr
+		     * offsets differ
+		     */
+		    target_ptrdiff_t offset;
+		    if (fdv->type->nextOf()->isBaseOf(type->nextOf(), &offset))
+		    {
+			tintro = fdv->type;
+		    }
+		}
+		break;
 	    }
 	}
 
-	// This is an 'introducing' function.
-
-	// Verify this doesn't override previous final function
-	if (cd->baseClass)
-	{   Dsymbol *s = cd->baseClass->search(loc, ident, 0);
-	    if (s)
-	    {
-		FuncDeclaration *f = s->isFuncDeclaration();
-		f = f->overloadExactMatch(type);
-		if (f && f->isFinal() && f->prot() != PROTprivate)
-		    error("cannot override final function %s", f->toPrettyChars());
-	    }
-	}
-
-	if (isFinal())
-	{
-	    cd->vtblFinal.push(this);
-	}
-	else
-	{
-	    // Append to end of vtbl[]
-	    //printf("\tintroducing function\n");
-	    introducing = 1;
-	    vi = cd->vtbl.dim;
-	    cd->vtbl.push(this);
-	    vtblIndex = vi;
-	}
-
-    L1: ;
-
 	/* Go through all the interface bases.
 	 * If this function is covariant with any members of those interface
 	 * functions, set the tintro.
@@ -384,63 +416,51 @@ void FuncDeclaration::semantic(Scope *sc)
 	for (int i = 0; i < cd->interfaces_dim; i++)
 	{
 	    BaseClass *b = cd->interfaces[i];
-	    for (vi = 0; vi < b->base->vtbl.dim; vi++)
+	    vi = findVtblIndex(&b->base->vtbl, b->base->vtbl.dim);
+	    switch (vi)
 	    {
-		Dsymbol *s = (Dsymbol *)b->base->vtbl.data[vi];
-		//printf("interface %d vtbl[%d] %p %s\n", i, vi, s, s->toChars());
-		FuncDeclaration *fdv = s->isFuncDeclaration();
-		if (fdv && fdv->ident == ident)
-		{
-		    int cov = type->covariant(fdv->type);
-		    //printf("\tcov = %d\n", cov);
-		    if (cov == 2)
-		    {
-			//type->print();
-			//fdv->type->print();
-			//printf("%s %s\n", type->deco, fdv->type->deco);
-			error("of type %s overrides but is not covariant with %s of type %s",
-			    type->toChars(), fdv->toPrettyChars(), fdv->type->toChars());
-		    }
-		    if (cov == 1)
-		    {	Type *ti = NULL;
+		case -1:
+		    break;
 
-			if (fdv->tintro)
-			    ti = fdv->tintro;
-			else if (!type->equals(fdv->type))
-			{
-			    /* Only need to have a tintro if the vptr
-			     * offsets differ
-			     */
-			    target_ptrdiff_t offset;
-			    if (fdv->type->nextOf()->isBaseOf(type->nextOf(), &offset))
-			    {
-				ti = fdv->type;
-#if 0
-				if (offset)
-				    ti = fdv->type;
-				else if (type->nextOf()->ty == Tclass)
-				{   ClassDeclaration *cdn = ((TypeClass *)type->nextOf())->sym;
-				    if (cdn && cdn->sizeok != 1)
-					ti = fdv->type;
-				}
-#endif
-			    }
-			}
-			if (ti)
-			{
-			    if (tintro && !tintro->equals(ti))
-			    {
-				error("incompatible covariant types %s and %s", tintro->toChars(), ti->toChars());
-			    }
-			    tintro = ti;
-			}
-			goto L2;
-		    }
-		    if (cov == 3)
+		case -2:
+		    cd->sizeok = 2;	// can't finish due to forward reference
+		    return;
+
+		default:
+		{   FuncDeclaration *fdv = (FuncDeclaration *)b->base->vtbl.data[vi];
+		    Type *ti = NULL;
+
+		    if (fdv->tintro)
+			ti = fdv->tintro;
+		    else if (!type->equals(fdv->type))
 		    {
-			cd->sizeok = 2;	// can't finish due to forward reference
-			return;
+			/* Only need to have a tintro if the vptr
+			 * offsets differ
+			 */
+			target_ptrdiff_t offset;
+			if (fdv->type->nextOf()->isBaseOf(type->nextOf(), &offset))
+			{
+			    ti = fdv->type;
+#if 0
+			    if (offset)
+				ti = fdv->type;
+			    else if (type->nextOf()->ty == Tclass)
+			    {   ClassDeclaration *cdn = ((TypeClass *)type->nextOf())->sym;
+				if (cdn && cdn->sizeok != 1)
+				    ti = fdv->type;
+			    }
+#endif
+			}
 		    }
+		    if (ti)
+		    {
+			if (tintro && !tintro->equals(ti))
+			{
+			    error("incompatible covariant types %s and %s", tintro->toChars(), ti->toChars());
+			}
+			tintro = ti;
+		    }
+		    goto L2;
 		}
 	    }
 	}
@@ -548,6 +568,11 @@ Ldone:
     return;
 
 Lassignerr:
+    if (sd)
+    {
+	sd->hasIdentityAssign = 1;	// don't need to generate it
+	goto Ldone;
+    }
     error("identity assignment operator overload is illegal");
 }
 
@@ -565,10 +590,12 @@ void FuncDeclaration::semantic3(Scope *sc)
 
     if (!parent)
     {
-	printf("FuncDeclaration::semantic3(%s '%s', sc = %p)\n", kind(), toChars(), sc);
+	if (global.errors)
+	    return;
+	//printf("FuncDeclaration::semantic3(%s '%s', sc = %p)\n", kind(), toChars(), sc);
 	assert(0);
     }
-    //printf("FuncDeclaration::semantic3('%s.%s', sc = %p)\n", parent->toChars(), toChars(), sc);
+    //printf("FuncDeclaration::semantic3('%s.%s', sc = %p, loc = %s)\n", parent->toChars(), toChars(), sc, loc.toChars());
     //fflush(stdout);
     //{ static int x; if (++x == 2) *(char*)0=0; }
     //printf("\tlinkage = %d\n", sc->linkage);
@@ -581,7 +608,6 @@ void FuncDeclaration::semantic3(Scope *sc)
     if (!type || type->ty != Tfunction)
 	return;
     f = (TypeFunction *)(type);
-    size_t nparams = Argument::dim(f->parameters);
 
     // Check the 'throws' clause
     if (fthrows)
@@ -599,15 +625,15 @@ void FuncDeclaration::semantic3(Scope *sc)
 
     if (fbody || frequire)
     {
-	// Establish function scope
-	ScopeDsymbol *ss;
-	Scope *sc2;
-
+	/* Symbol table into which we place parameters and nested functions,
+	 * solely to diagnose name collisions.
+	 */
 	localsymtab = new DsymbolTable();
 
-	ss = new ScopeDsymbol();
+	// Establish function scope
+	ScopeDsymbol *ss = new ScopeDsymbol();
 	ss->parent = sc->scopesym;
-	sc2 = sc->push(ss);
+	Scope *sc2 = sc->push(ss);
 	sc2->func = this;
 	sc2->parent = this;
 	sc2->callSuper = 0;
@@ -616,7 +642,7 @@ void FuncDeclaration::semantic3(Scope *sc)
 	sc2->sw = NULL;
 	sc2->fes = fes;
 	sc2->linkage = LINKd;
-	sc2->stc &= ~(STCauto | STCscope | STCstatic | STCabstract | STCdeprecated | STCconst | STCfinal | STCinvariant);
+	sc2->stc &= ~(STCauto | STCscope | STCstatic | STCabstract | STCdeprecated | STCconst | STCfinal | STCinvariant | STCtls);
 	sc2->attributes = NULL;
 	sc2->protection = PROTpublic;
 	sc2->explicitProtection = 0;
@@ -640,7 +666,7 @@ void FuncDeclaration::semantic3(Scope *sc)
 		assert(!isNested());	// can't be both member and nested
 		assert(ad->handle);
 		Type *thandle = ad->handle;
-		if (storage_class & STCconst)
+		if (storage_class & STCconst || type->isConst())
 		{
 		    if (thandle->ty == Tclass)
 			thandle = thandle->constOf();
@@ -649,7 +675,7 @@ void FuncDeclaration::semantic3(Scope *sc)
 			thandle = thandle->nextOf()->constOf()->pointerTo();
 		    }
 		}
-		else if (storage_class & STCinvariant)
+		else if (storage_class & STCinvariant || type->isInvariant())
 		{
 		    if (thandle->ty == Tclass)
 			thandle = thandle->invariantOf();
@@ -669,9 +695,11 @@ void FuncDeclaration::semantic3(Scope *sc)
 	}
 	else if (isNested())
 	{
-	    VarDeclaration *v;
-
-	    v = new ThisDeclaration(Type::tvoid->pointerTo());
+	    /* The 'this' for a nested function is the link to the
+	     * enclosing function's stack frame.
+	     * Note that nested functions and member functions are disjoint.
+	     */
+	    VarDeclaration *v = new ThisDeclaration(Type::tvoid->pointerTo());
 	    v->storage_class |= STCparameter;
 	    v->semantic(sc2);
 	    if (!sc2->insert(v))
@@ -688,11 +716,12 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    {	// Declare _arguments[]
 #if BREAKABI
 		v_arguments = new VarDeclaration(0, Type::typeinfotypelist->type, Id::_arguments_typeinfo, NULL);
-		v_arguments->storage_class = STCparameter | STCin;
+		v_arguments->storage_class = STCparameter;
 		v_arguments->semantic(sc2);
 		sc2->insert(v_arguments);
 		v_arguments->parent = this;
 
+		//t = Type::typeinfo->type->constOf()->arrayOf();
 		t = Type::typeinfo->type->arrayOf();
 		_arguments = new VarDeclaration(0, t, Id::_arguments, NULL);
 		_arguments->semantic(sc2);
@@ -721,7 +750,7 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    }
 	}
 
-	// Propagate storage class from tuple arguments to their sub-arguments.
+	// Propagate storage class from tuple parameters to their element-parameters.
 	if (f->parameters)
 	{
 	    for (size_t i = 0; i < f->parameters->dim; i++)
@@ -738,10 +767,14 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    }
 	}
 
-	// Declare all the function parameters as variables
+	/* Declare all the function parameters as variables
+	 * and install them in parameters[]
+	 */
+	size_t nparams = Argument::dim(f->parameters);
 	if (nparams)
-	{   // parameters[] has all the tuples removed, as the back end
-	    // doesn't know about tuples
+	{   /* parameters[] has all the tuples removed, as the back end
+	     * doesn't know about tuples
+	     */
 	    parameters = new Dsymbols();
 	    parameters->reserve(nparams);
 	    for (size_t i = 0; i < nparams; i++)
@@ -750,21 +783,21 @@ void FuncDeclaration::semantic3(Scope *sc)
 		Identifier *id = arg->ident;
 		if (!id)
 		{
-		    //error("no identifier for parameter %d of %s", i + 1, toChars());
+		    /* Generate identifier for un-named parameter,
+		     * because we need it later on.
+		     */
 		    OutBuffer buf;
 		    buf.printf("_param_%"PRIuSIZE, i);
 		    char *name = (char *)buf.extractData();
 		    id = new Identifier(name, TOKidentifier);
 		    arg->ident = id;
 		}
-		VarDeclaration *v = new VarDeclaration(0, arg->type, id, NULL);
+		VarDeclaration *v = new VarDeclaration(loc, arg->type, id, NULL);
 		//printf("declaring parameter %s of type %s\n", v->toChars(), v->type->toChars());
 		v->storage_class |= STCparameter;
 		if (f->varargs == 2 && i + 1 == nparams)
 		    v->storage_class |= STCvariadic;
-		v->storage_class |= arg->storageClass & (STCin | STCout | STCref | STClazy | STCfinal | STCconst | STCinvariant);
-		if (v->storage_class & STClazy)
-		    v->storage_class |= STCin;
+		v->storage_class |= arg->storageClass & (STCin | STCout | STCref | STClazy | STCfinal | STCconst | STCinvariant | STCnodtor);
 		v->semantic(sc2);
 		if (!sc2->insert(v))
 		    error("parameter %s.%s is already defined", toChars(), v->toChars());
@@ -794,11 +827,11 @@ void FuncDeclaration::semantic3(Scope *sc)
 			assert(narg->ident);
 			VarDeclaration *v = sc2->search(0, narg->ident, NULL)->isVarDeclaration();
 			assert(v);
-			Expression *e = new VarExp(0, v);
+			Expression *e = new VarExp(v->loc, v);
 			exps->data[j] = (void *)e;
 		    }
 		    assert(arg->ident);
-		    TupleDeclaration *v = new TupleDeclaration(0, arg->ident, exps);
+		    TupleDeclaration *v = new TupleDeclaration(loc, arg->ident, exps);
 		    //printf("declaring tuple %s\n", v->toChars());
 		    v->isexp = 1;
 		    if (!sc2->insert(v))
@@ -809,10 +842,14 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    }
 	}
 
+	/* Do the semantic analysis on the [in] preconditions and
+	 * [out] postconditions.
+	 */
 	sc2->incontract++;
 
 	if (frequire)
-	{
+	{   /* frequire is composed of the [in] contracts
+	     */
 	    // BUG: need to error if accessing out parameters
 	    // BUG: need to treat parameters as const
 	    // BUG: need to disallow returns and throws
@@ -822,10 +859,9 @@ void FuncDeclaration::semantic3(Scope *sc)
 	}
 
 	if (fensure || addPostInvariant())
-	{
-	    ScopeDsymbol *sym;
-
-	    sym = new ScopeDsymbol();
+	{   /* fensure is composed of the [out] contracts
+	     */
+	    ScopeDsymbol *sym = new ScopeDsymbol();
 	    sym->parent = sc2->scopesym;
 	    sc2 = sc2->push(sym);
 
@@ -929,6 +965,8 @@ void FuncDeclaration::semantic3(Scope *sc)
 	if (fbody)
 	{   ClassDeclaration *cd = isClassMember();
 
+	    /* If this is a class constructor
+	     */
 	    if (isCtorDeclaration() && cd)
 	    {
 		for (int i = 0; i < cd->fields.dim; i++)
@@ -953,19 +991,24 @@ void FuncDeclaration::semantic3(Scope *sc)
 		f = (TypeFunction *)type;
 	    }
 
-	    int offend = fbody ? fbody->fallOffEnd() : TRUE;
-
 	    if (isStaticCtorDeclaration())
 	    {	/* It's a static constructor. Ensure that all
 		 * ctor consts were initialized.
 		 */
 
-		ScopeDsymbol *ad = toParent()->isScopeDsymbol();
-		assert(ad);
-		for (int i = 0; i < ad->members->dim; i++)
-		{   Dsymbol *s = (Dsymbol *)ad->members->data[i];
+		Dsymbol *p = toParent();
+		ScopeDsymbol *ad = p->isScopeDsymbol();
+		if (!ad)
+		{
+		    error("static constructor can only be member of struct/class/module, not %s %s", p->kind(), p->toChars());
+		}
+		else
+		{
+		    for (int i = 0; i < ad->members->dim; i++)
+		    {   Dsymbol *s = (Dsymbol *)ad->members->data[i];
 
-		    s->checkCtorConstInit();
+			s->checkCtorConstInit();
+		    }
 		}
 	    }
 
@@ -1015,6 +1058,9 @@ void FuncDeclaration::semantic3(Scope *sc)
 		error("expected to return a value of type %s", type->nextOf()->toChars());
 	    else if (!inlineAsm)
 	    {
+		int offend = fbody ? fbody->blockExit() & BEfallthru : TRUE;
+		//int offend = fbody ? fbody->fallOffEnd() : TRUE;
+
 		if (type->nextOf()->ty == Tvoid)
 		{
 		    if (offend && isMain())
@@ -1061,9 +1107,8 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    // Merge in initialization of 'out' parameters
 	    if (parameters)
 	    {	for (size_t i = 0; i < parameters->dim; i++)
-		{   VarDeclaration *v;
-
-		    v = (VarDeclaration *)parameters->data[i];
+		{
+		    VarDeclaration *v = (VarDeclaration *)parameters->data[i];
 		    if (v->storage_class & STCout)
 		    {
 			assert(v->init);
@@ -1114,6 +1159,7 @@ void FuncDeclaration::semantic3(Scope *sc)
 		e = new DotIdExp(0, e, Id::elements);
 		Expression *e1 = new VarExp(0, _arguments);
 		e = new AssignExp(0, e1, e);
+		e->op = TOKconstruct;
 		e = e->semantic(sc);
 		a->push(new ExpStatement(0, e));
 	    }
@@ -1192,6 +1238,35 @@ void FuncDeclaration::semantic3(Scope *sc)
 	    }
 
 	    fbody = new CompoundStatement(0, a);
+
+	    /* Append destructor calls for parameters as finally blocks.
+	     */
+	    if (parameters)
+	    {	for (size_t i = 0; i < parameters->dim; i++)
+		{
+		    VarDeclaration *v = (VarDeclaration *)parameters->data[i];
+
+		    if (v->storage_class & (STCref | STCout))
+			continue;
+
+		    /* Don't do this for static arrays, since static
+		     * arrays are called by reference. Remove this
+		     * when we change them to call by value.
+		     */
+		    if (v->type->toBasetype()->ty == Tsarray)
+			continue;
+
+		    Expression *e = v->callAutoDtor(sc);
+		    if (e)
+		    {	Statement *s = new ExpStatement(0, e);
+			s = s->semantic(sc);
+			if (fbody->blockExit() == BEfallthru)
+			    fbody = new CompoundStatement(0, fbody, s);
+			else
+			    fbody = new TryFinallyStatement(0, fbody, s);
+		    }
+		}
+	    }
 	}
 
 	sc2->callSuper = 0;
@@ -1274,6 +1349,50 @@ int FuncDeclaration::overrides(FuncDeclaration *fd)
     return result;
 }
 
+/*************************************************
+ * Find index of function in vtbl[0..dim] that
+ * this function overrides.
+ * Returns:
+ *	-1	didn't find one
+ *	-2	can't determine because of forward references
+ */
+
+int FuncDeclaration::findVtblIndex(Array *vtbl, int dim)
+{
+    for (int vi = 0; vi < dim; vi++)
+    {
+	FuncDeclaration *fdv = ((Dsymbol *)vtbl->data[vi])->isFuncDeclaration();
+	if (fdv && fdv->ident == ident)
+	{
+	    int cov = type->covariant(fdv->type);
+	    //printf("\tbaseclass cov = %d\n", cov);
+	    switch (cov)
+	    {
+		case 0:		// types are distinct
+		    break;
+
+		case 1:
+		    return vi;
+
+		case 2:
+		    //type->print();
+		    //fdv->type->print();
+		    //printf("%s %s\n", type->deco, fdv->type->deco);
+		    error("of type %s overrides but is not covariant with %s of type %s",
+			type->toChars(), fdv->toPrettyChars(), fdv->type->toChars());
+		    break;
+
+		case 3:
+		    return -2;	// forward references
+
+		default:
+		    assert(0);
+	    }
+	}
+    }
+    return -1;
+}
+
 /****************************************************
  * Overload this FuncDeclaration with the new one f.
  * Return !=0 if successful; i.e. no conflict.
@@ -1303,13 +1422,25 @@ int FuncDeclaration::overloadInsert(Dsymbol *s)
     if (!f)
 	return FALSE;
 
+#if 0
+    /* Disable this check because:
+     *	const void foo();
+     * semantic() isn't run yet on foo(), so the const hasn't been
+     * applied yet.
+     */
+    if (type)
+    {   printf("type = %s\n", type->toChars());
+	printf("f->type = %s\n", f->type->toChars());
+    }
     if (type && f->type &&	// can be NULL for overloaded constructors
 	f->type->covariant(type) &&
+	f->type->mod == type->mod &&
 	!isFuncAliasDeclaration())
     {
 	//printf("\tfalse: conflict %s\n", kind());
 	return FALSE;
     }
+#endif
 
     if (overnext)
 	return overnext->overloadInsert(f);
@@ -1452,6 +1583,7 @@ FuncDeclaration *FuncDeclaration::overloadExactMatch(Type *t)
 struct Param2
 {
     Match *m;
+    Expression *ethis;
     Expressions *arguments;
 };
 
@@ -1463,11 +1595,9 @@ int fp2(void *param, FuncDeclaration *f)
 
     if (f != m->lastf)		// skip duplicates
     {
-	TypeFunction *tf;
-
 	m->anyf = f;
-	tf = (TypeFunction *)f->type;
-	match = (MATCH) tf->callMatch(arguments);
+	TypeFunction *tf = (TypeFunction *)f->type;
+	match = (MATCH) tf->callMatch(f->needThis() ? p->ethis : NULL, arguments);
 	//printf("match = %d\n", match);
 	if (match != MATCHnomatch)
 	{
@@ -1484,6 +1614,20 @@ int fp2(void *param, FuncDeclaration *f)
 	    else if (f->overrides(m->lastf))
 		goto LfIsBetter;
 
+	    /* Try to disambiguate using template-style partial ordering rules.
+	     * In essence, if f() and g() are ambiguous, if f() can call g(),
+	     * but g() cannot call f(), then pick f().
+	     * This is because f() is "more specialized."
+	     */
+	    {
+	    MATCH c1 = f->leastAsSpecialized(m->lastf);
+	    MATCH c2 = m->lastf->leastAsSpecialized(f);
+	    if (c1 > c2)
+		goto LfIsBetter;
+	    if (c1 < c2)
+		goto LlastIsBetter;
+	    }
+
 	Lambiguous:
 	    m->nextf = f;
 	    m->count++;
@@ -1502,16 +1646,18 @@ int fp2(void *param, FuncDeclaration *f)
     return 0;
 }
 
-void overloadResolveX(Match *m, FuncDeclaration *fstart, Expressions *arguments)
+void overloadResolveX(Match *m, FuncDeclaration *fstart,
+	Expression *ethis, Expressions *arguments)
 {
     Param2 p;
     p.m = m;
+    p.ethis = ethis;
     p.arguments = arguments;
     overloadApply(fstart, &fp2, &p);
 }
 
 
-FuncDeclaration *FuncDeclaration::overloadResolve(Loc loc, Expressions *arguments)
+FuncDeclaration *FuncDeclaration::overloadResolve(Loc loc, Expression *ethis, Expressions *arguments, int flags)
 {
     TypeFunction *tf;
     Match m;
@@ -1534,7 +1680,7 @@ if (arguments)
 
     memset(&m, 0, sizeof(m));
     m.last = MATCHnomatch;
-    overloadResolveX(&m, this, arguments);
+    overloadResolveX(&m, this, ethis, arguments);
 
     if (m.count == 1)		// exactly one match
     {
@@ -1553,6 +1699,9 @@ if (arguments)
 
 	if (m.last == MATCHnomatch)
 	{
+	    if (flags & 1)		// if do not print error messages
+		return NULL;		// no match
+
 	    tf = (TypeFunction *)type;
 
 	    //printf("tf = %s, args = %s\n", tf->deco, ((Expression *)arguments->data[0])->type->deco);
@@ -1582,6 +1731,79 @@ if (arguments)
     }
 }
 
+/*************************************
+ * Determine partial specialization order of 'this' vs g.
+ * This is very similar to TemplateDeclaration::leastAsSpecialized().
+ * Returns:
+ *	match	'this' is at least as specialized as g
+ *	0	g is more specialized than 'this'
+ */
+
+MATCH FuncDeclaration::leastAsSpecialized(FuncDeclaration *g)
+{
+#define LOG_LEASTAS     0
+
+#if LOG_LEASTAS
+    printf("%s.leastAsSpecialized(%s)\n", toChars(), g->toChars());
+#endif
+
+    /* This works by calling g() with f()'s parameters, and
+     * if that is possible, then f() is at least as specialized
+     * as g() is.
+     */
+
+    TypeFunction *tf = (TypeFunction *)type;
+    TypeFunction *tg = (TypeFunction *)g->type;
+    size_t nfparams = Argument::dim(tf->parameters);
+    size_t ngparams = Argument::dim(tg->parameters);
+    MATCH match = MATCHexact;
+
+    /* If both functions have a 'this' pointer, and the mods are not
+     * the same and g's is not const, then this is less specialized.
+     */
+    if (needThis() && g->needThis())
+    {
+	if (tf->mod != tg->mod)
+	{
+	    if (tg->mod == MODconst)
+		match = MATCHconst;
+	    else
+		return MATCHnomatch;
+	}
+    }
+
+    /* Create a dummy array of arguments out of the parameters to f()
+     */
+    Expressions args;
+    args.setDim(nfparams);
+    for (int u = 0; u < nfparams; u++)
+    {
+	Argument *p = Argument::getNth(tf->parameters, u);
+	Expression *e = p->type->defaultInit();
+	args.data[u] = e;
+    }
+
+    MATCH m = (MATCH) tg->callMatch(NULL, &args);
+    if (m)
+    {
+        /* A variadic template is less specialized than a
+         * non-variadic one.
+         */
+        if (tf->varargs && !tg->varargs)
+            goto L1;	// less specialized
+
+#if LOG_LEASTAS
+        printf("  matches %d, so is least as specialized\n", m);
+#endif
+        return m;
+    }
+  L1:
+#if LOG_LEASTAS
+    printf("  doesn't match, so is not as specialized\n");
+#endif
+    return MATCHnomatch;
+}
+
 /********************************
  * Labels are in a separate scope, one per function.
  */
@@ -1601,6 +1823,12 @@ LabelDsymbol *FuncDeclaration::searchLabel(Identifier *ident)
     return (LabelDsymbol *)s;
 }
 
+/****************************************
+ * If non-static member function that has a 'this' pointer,
+ * return the aggregate it is a member of.
+ * Otherwise, return NULL.
+ */
+
 AggregateDeclaration *FuncDeclaration::isThis()
 {   AggregateDeclaration *ad;
 
@@ -1716,8 +1944,16 @@ int FuncDeclaration::isMain()
 
 int FuncDeclaration::isWinMain()
 {
+    //printf("FuncDeclaration::isWinMain() %s\n", toChars());
+#if 0
+    int x = ident == Id::WinMain &&
+	linkage != LINKc && !isMember();
+    printf("%s\n", x ? "yes" : "no");
+    return x;
+#else
     return ident == Id::WinMain &&
 	linkage != LINKc && !isMember();
+#endif
 }
 
 int FuncDeclaration::isDllMain()
@@ -1781,6 +2017,11 @@ int FuncDeclaration::isCodeseg()
     return TRUE;		// functions are always in the code segment
 }
 
+int FuncDeclaration::isOverloadable()
+{
+    return 1;			// functions can be overloaded
+}
+
 // Determine if function needs
 // a static frame pointer to its lexically enclosing function
 
@@ -1788,7 +2029,7 @@ int FuncDeclaration::isNested()
 {
     //if (!toParent())
 	//printf("FuncDeclaration::isNested('%s') parent=%p\n", toChars(), parent);
-    //printf("\ttoParent() = '%s'\n", toParent()->toChars());
+    //printf("\ttoParent2() = '%s'\n", toParent2()->toChars());
     return ((storage_class & STCstatic) == 0) &&
 	   (toParent2()->isFuncDeclaration() != NULL);
 }
@@ -1810,7 +2051,8 @@ int FuncDeclaration::addPreInvariant()
 	    //ad->isClassDeclaration() &&
 	    global.params.useInvariants &&
 	    (protection == PROTpublic || protection == PROTexport) &&
-	    !naked);
+	    !naked &&
+	    ident != Id::cpctor);
 }
 
 int FuncDeclaration::addPostInvariant()
@@ -1821,7 +2063,8 @@ int FuncDeclaration::addPostInvariant()
 	    //ad->isClassDeclaration() &&
 	    global.params.useInvariants &&
 	    (protection == PROTpublic || protection == PROTexport) &&
-	    !naked);
+	    !naked &&
+	    ident != Id::cpctor);
 }
 
 /**********************************
@@ -1884,6 +2127,60 @@ char *FuncDeclaration::kind()
     return "function";
 }
 
+/*******************************
+ * Look at all the variables in this function that are referenced
+ * by nested functions, and determine if a closure needs to be
+ * created for them.
+ */
+
+#if V2
+int FuncDeclaration::needsClosure()
+{
+    /* Need a closure for all the closureVars[] if any of the
+     * closureVars[] are accessed by a
+     * function that escapes the scope of this function.
+     * We take the conservative approach and decide that any function that:
+     * 1) is a virtual function
+     * 2) has its address taken
+     * 3) has a parent that escapes
+     *
+     * Note that since a non-virtual function can be called by
+     * a virtual one, if that non-virtual function accesses a closure
+     * var, the closure still has to be taken. Hence, we check for isThis()
+     * instead of isVirtual(). (thanks to David Friedman)
+     */
+
+    //printf("FuncDeclaration::needsClosure() %s\n", toChars());
+    for (int i = 0; i < closureVars.dim; i++)
+    {	VarDeclaration *v = (VarDeclaration *)closureVars.data[i];
+	assert(v->isVarDeclaration());
+	//printf("\tv = %s\n", v->toChars());
+
+	for (int j = 0; j < v->nestedrefs.dim; j++)
+	{   FuncDeclaration *f = (FuncDeclaration *)v->nestedrefs.data[j];
+	    assert(f != this);
+
+	    //printf("\t\tf = %s, %d, %d\n", f->toChars(), f->isVirtual(), f->tookAddressOf);
+	    if (f->isThis() || f->tookAddressOf)
+		goto Lyes;	// assume f escapes this function's scope
+
+	    // Look to see if any parents of f that are below this escape
+	    for (Dsymbol *s = f->parent; s != this; s = s->parent)
+	    {
+		f = s->isFuncDeclaration();
+		if (f && (f->isThis() || f->tookAddressOf))
+		    goto Lyes;
+	    }
+	}
+    }
+    return 0;
+
+Lyes:
+    //printf("\tneeds closure\n");
+    return 1;
+}
+#endif
+
 /****************************** FuncAliasDeclaration ************************/
 
 // Used as a way to import a set of functions from another scope into this one.
@@ -1995,6 +2292,8 @@ void CtorDeclaration::semantic(Scope *sc)
     Type *tret;
 
     //printf("CtorDeclaration::semantic()\n");
+    if (type)
+	return;
 
     sc = sc->push();
     sc->stc &= ~STCstatic;		// not a static constructor
@@ -2004,7 +2303,7 @@ void CtorDeclaration::semantic(Scope *sc)
     cd = parent->isClassDeclaration();
     if (!cd)
     {
-	error("constructors only are for class definitions");
+	error("constructors are only for class definitions");
 	tret = Type::tvoid;
     }
     else
@@ -2069,6 +2368,78 @@ void CtorDeclaration::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
     bodyToCBuffer(buf, hgs);
 }
 
+/********************************* PostBlitDeclaration ****************************/
+
+PostBlitDeclaration::PostBlitDeclaration(Loc loc, Loc endloc)
+    : FuncDeclaration(loc, endloc, Id::_postblit, STCundefined, NULL)
+{
+}
+
+PostBlitDeclaration::PostBlitDeclaration(Loc loc, Loc endloc, Identifier *id)
+    : FuncDeclaration(loc, endloc, id, STCundefined, NULL)
+{
+}
+
+Dsymbol *PostBlitDeclaration::syntaxCopy(Dsymbol *s)
+{
+    assert(!s);
+    PostBlitDeclaration *dd = new PostBlitDeclaration(loc, endloc, ident);
+    return FuncDeclaration::syntaxCopy(dd);
+}
+
+
+void PostBlitDeclaration::semantic(Scope *sc)
+{
+    //printf("PostBlitDeclaration::semantic() %s\n", toChars());
+    //printf("ident: %s, %s, %p, %p\n", ident->toChars(), Id::dtor->toChars(), ident, Id::dtor);
+    parent = sc->parent;
+    Dsymbol *parent = toParent();
+    StructDeclaration *ad = parent->isStructDeclaration();
+    if (!ad)
+    {
+	error("post blits are only for struct/union definitions, not %s %s", parent->kind(), parent->toChars());
+    }
+    else if (ident == Id::_postblit)
+	ad->postblits.push(this);
+    type = new TypeFunction(NULL, Type::tvoid, FALSE, LINKd);
+
+    sc = sc->push();
+    sc->stc &= ~STCstatic;		// not static
+    sc->linkage = LINKd;
+
+    FuncDeclaration::semantic(sc);
+
+    sc->pop();
+}
+
+int PostBlitDeclaration::overloadInsert(Dsymbol *s)
+{
+    return FALSE;	// cannot overload postblits
+}
+
+int PostBlitDeclaration::addPreInvariant()
+{
+    return FALSE;
+}
+
+int PostBlitDeclaration::addPostInvariant()
+{
+    return (vthis && global.params.useInvariants);
+}
+
+int PostBlitDeclaration::isVirtual()
+{
+    return FALSE;
+}
+
+void PostBlitDeclaration::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
+{
+    if (hgs->hdrgen)
+	return;
+    buf->writestring("=this()");
+    bodyToCBuffer(buf, hgs);
+}
+
 /********************************* DtorDeclaration ****************************/
 
 DtorDeclaration::DtorDeclaration(Loc loc, Loc endloc)
@@ -2076,29 +2447,32 @@ DtorDeclaration::DtorDeclaration(Loc loc, Loc endloc)
 {
 }
 
+DtorDeclaration::DtorDeclaration(Loc loc, Loc endloc, Identifier *id)
+    : FuncDeclaration(loc, endloc, id, STCundefined, NULL)
+{
+}
+
 Dsymbol *DtorDeclaration::syntaxCopy(Dsymbol *s)
 {
-    DtorDeclaration *dd;
-
     assert(!s);
-    dd = new DtorDeclaration(loc, endloc);
+    DtorDeclaration *dd = new DtorDeclaration(loc, endloc, ident);
     return FuncDeclaration::syntaxCopy(dd);
 }
 
 
 void DtorDeclaration::semantic(Scope *sc)
 {
-    ClassDeclaration *cd;
-
+    //printf("DtorDeclaration::semantic() %s\n", toChars());
+    //printf("ident: %s, %s, %p, %p\n", ident->toChars(), Id::dtor->toChars(), ident, Id::dtor);
     parent = sc->parent;
     Dsymbol *parent = toParent();
-    cd = parent->isClassDeclaration();
-    if (!cd)
+    AggregateDeclaration *ad = parent->isAggregateDeclaration();
+    if (!ad)
     {
-	error("destructors only are for class definitions");
+	error("destructors are only for class/struct/union definitions, not %s %s", parent->kind(), parent->toChars());
     }
-    else
-	cd->dtors.push(this);
+    else if (ident == Id::dtor)
+	ad->dtors.push(this);
     type = new TypeFunction(NULL, Type::tvoid, FALSE, LINKd);
 
     sc = sc->push();
@@ -2211,7 +2585,7 @@ int StaticCtorDeclaration::addPostInvariant()
 void StaticCtorDeclaration::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 {
     if (hgs->hdrgen)
-    {	buf->writestring("static this(){}\n");
+    {	buf->writestring("static this();\n");
 	return;
     }
     buf->writestring("static this()");
@@ -2375,12 +2749,7 @@ void InvariantDeclaration::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 
 static Identifier *unitTestId()
 {
-    static int n;
-    char buffer[10 + sizeof(n)*3 + 1];
-
-    sprintf(buffer,"__unittest%d", n);
-    n++;
-    return Lexer::idPool(buffer);
+    return Lexer::uniqueId("__unittest");
 }
 
 UnitTestDeclaration::UnitTestDeclaration(Loc loc, Loc endloc)
