@@ -40,6 +40,7 @@ struct TypedefDeclaration;
 struct TypeInfoDeclaration;
 struct Dsymbol;
 struct TemplateInstance;
+struct CppMangleState;
 enum LINK;
 
 struct TypeBasic;
@@ -104,6 +105,7 @@ enum ENUMTY
     Ttypeof,
     Ttuple,
     Tslice,
+    Treturn,
     TMAX
 };
 typedef unsigned char TY;	// ENUMTY
@@ -219,6 +221,9 @@ struct Type : Object
     virtual void toCBuffer(OutBuffer *buf, Identifier *ident, HdrGenState *hgs);
     virtual void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     void toCBuffer3(OutBuffer *buf, HdrGenState *hgs, int mod);
+#if IN_GCC || TARGET_LINUX
+    virtual void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
     virtual int isintegral();
     virtual int isfloating();	// real, imaginary, or complex
     virtual int isreal();
@@ -228,6 +233,7 @@ struct Type : Object
     virtual int isunsigned();
     virtual int isauto();
     virtual int isString();
+    virtual int isAssignable();
     virtual int checkBoolean();	// if can be converted to boolean value
     virtual void checkDeprecated(Loc loc, Scope *sc);
     int isConst()	{ return mod == MODconst; }
@@ -243,7 +249,7 @@ struct Type : Object
     virtual Type *makeInvariant();
     virtual Dsymbol *toDsymbol(Scope *sc);
     virtual Type *toBasetype();
-    virtual Type *toCanonConst();
+    virtual Type *toHeadMutable();
     virtual int isBaseOf(Type *t, target_ptrdiff_t *poffset);
     virtual MATCH constConv(Type *to);
     virtual MATCH implicitConvTo(Type *to);
@@ -251,7 +257,7 @@ struct Type : Object
     virtual Expression *getProperty(Loc loc, Identifier *ident);
     virtual Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
     virtual unsigned memalign(unsigned salign);
-    virtual Expression *defaultInit();
+    virtual Expression *defaultInit(Loc loc = 0);
     virtual int isZeroInit();		// if initializer is 0
     virtual dt_t **toDt(dt_t **pdt);
     Identifier *getTypeInfoIdent(int internal);
@@ -305,6 +311,9 @@ struct TypeBasic : Type
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
     char *toChars();
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
     int isintegral();
     int isbit();
     int isfloating();
@@ -314,7 +323,7 @@ struct TypeBasic : Type
     int isscalar();
     int isunsigned();
     MATCH implicitConvTo(Type *to);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     int builtinTypeInfo();
 
@@ -347,13 +356,16 @@ struct TypeSArray : TypeArray
     unsigned memalign(unsigned salign);
     MATCH constConv(Type *to);
     MATCH implicitConvTo(Type *to);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     dt_t **toDt(dt_t **pdt);
     dt_t **toDtElem(dt_t **pdt, Expression *e);
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     TypeInfoDeclaration *getTypeInfoDeclaration();
     Expression *toExpression();
     int hasPointers();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
     type *toCParamtype();
@@ -374,10 +386,13 @@ struct TypeDArray : TypeArray
     int isZeroInit();
     int checkBoolean();
     MATCH implicitConvTo(Type *to);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int builtinTypeInfo();
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 };
@@ -390,14 +405,20 @@ struct TypeAArray : TypeArray
     Type *syntaxCopy();
     d_uns64 size(Loc loc);
     Type *semantic(Loc loc, Scope *sc);
+    void resolve(Loc loc, Scope *sc, Expression **pe, Type **pt, Dsymbol **ps);
     void toDecoBuffer(OutBuffer *buf, int flag);
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     int checkBoolean();
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
+    MATCH implicitConvTo(Type *to);
+    MATCH constConv(Type *to);
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     // Back end
     Symbol *aaGetSymbol(char *func, int flags);
@@ -414,10 +435,13 @@ struct TypePointer : TypeNext
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     MATCH implicitConvTo(Type *to);
     int isscalar();
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 };
@@ -430,8 +454,11 @@ struct TypeReference : TypeNext
     d_uns64 size(Loc loc);
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 };
 
 enum RET
@@ -447,6 +474,8 @@ struct TypeFunction : TypeNext
     Arguments *parameters;	// function parameters
     int varargs;	// 1: T t, ...) style for variable number of arguments
 			// 2: T t ...) style for variable number of arguments
+    bool isnothrow;	// true: nothrow
+    bool ispure;	// true: pure
     enum LINK linkage;	// calling convention
 
     int inuse;
@@ -460,8 +489,11 @@ struct TypeFunction : TypeNext
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     TypeInfoDeclaration *getTypeInfoDeclaration();
     Type *reliesOnTident();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
-    int callMatch(Expressions *toargs);
+    int callMatch(Expression *ethis, Expressions *toargs);
     type *toCtype();
     enum RET retStyle();
 
@@ -477,12 +509,15 @@ struct TypeDelegate : TypeNext
     Type *semantic(Loc loc, Scope *sc);
     d_uns64 size(Loc loc);
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     int checkBoolean();
     TypeInfoDeclaration *getTypeInfoDeclaration();
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
     int hasPointers();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 };
@@ -546,6 +581,15 @@ struct TypeTypeof : TypeQualified
     d_uns64 size(Loc loc);
 };
 
+struct TypeReturn : TypeQualified
+{
+    TypeReturn(Loc loc);
+    Type *syntaxCopy();
+    Dsymbol *toDsymbol(Scope *sc);
+    Type *semantic(Loc loc, Scope *sc);
+    void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
+};
+
 struct TypeStruct : Type
 {
     StructDeclaration *sym;
@@ -561,8 +605,9 @@ struct TypeStruct : Type
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
     unsigned memalign(unsigned salign);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
+    int isAssignable();
     int checkBoolean();
     dt_t **toDt(dt_t **pdt);
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
@@ -570,7 +615,10 @@ struct TypeStruct : Type
     int hasPointers();
     MATCH implicitConvTo(Type *to);
     MATCH constConv(Type *to);
-    Type *toCanonConst();
+    Type *toHeadMutable();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 };
@@ -596,11 +644,14 @@ struct TypeEnum : Type
     MATCH implicitConvTo(Type *to);
     MATCH constConv(Type *to);
     Type *toBasetype();
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 };
@@ -619,6 +670,7 @@ struct TypeTypedef : Type
     void toDecoBuffer(OutBuffer *buf, int flag);
     void toCBuffer2(OutBuffer *buf, HdrGenState *hgs, int mod);
     Expression *dotExp(Scope *sc, Expression *e, Identifier *ident);
+    Expression *getProperty(Loc loc, Identifier *ident);
     int isbit();
     int isintegral();
     int isfloating();
@@ -628,16 +680,20 @@ struct TypeTypedef : Type
     int isscalar();
     int isunsigned();
     int checkBoolean();
+    int isAssignable();
     Type *toBasetype();
     MATCH implicitConvTo(Type *to);
     MATCH constConv(Type *to);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     dt_t **toDt(dt_t **pdt);
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
-    Type *toCanonConst();
+    Type *toHeadMutable();
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
     type *toCParamtype();
@@ -659,15 +715,18 @@ struct TypeClass : Type
     ClassDeclaration *isClassHandle();
     int isBaseOf(Type *t, target_ptrdiff_t *poffset);
     MATCH implicitConvTo(Type *to);
-    Expression *defaultInit();
+    Expression *defaultInit(Loc loc);
     int isZeroInit();
     MATCH deduceType(Scope *sc, Type *tparam, TemplateParameters *parameters, Objects *dedtypes);
     int isauto();
     int checkBoolean();
     TypeInfoDeclaration *getTypeInfoDeclaration();
     int hasPointers();
-    Type *toCanonConst();
+    Type *toHeadMutable();
     MATCH constConv(Type *to);
+#if IN_GCC || TARGET_LINUX
+    void toCppMangle(OutBuffer *buf, CppMangleState *cms);
+#endif
 
     type *toCtype();
 
@@ -720,6 +779,7 @@ struct Argument : Object
     void toDecoBuffer(OutBuffer *buf);
     static Arguments *arraySyntaxCopy(Arguments *args);
     static char *argsTypesToChars(Arguments *args, int varargs);
+    static void argsCppMangle(OutBuffer *buf, CppMangleState *cms, Arguments *arguments, int varargs);
     static void argsToCBuffer(OutBuffer *buf, HdrGenState *hgs, Arguments *arguments, int varargs);
     static void argsToDecoBuffer(OutBuffer *buf, Arguments *arguments);
     static size_t dim(Arguments *arguments);
