@@ -45,10 +45,9 @@
 #include "aggregate.h"
 #include "template.h"
 
-static Expression *build_overload(Loc loc, Scope *sc, Expression *ethis, Expression *earg, Identifier *id);
 static void inferApplyArgTypesX(FuncDeclaration *fstart, Arguments *arguments);
 static int inferApplyArgTypesY(TypeFunction *tf, Arguments *arguments);
-static void templateResolve(Match *m, TemplateDeclaration *td, Scope *sc, Loc loc, Objects *targsi, Expressions *arguments);
+static void templateResolve(Match *m, TemplateDeclaration *td, Scope *sc, Loc loc, Objects *targsi, Expression *ethis, Expressions *arguments);
 
 /******************************** Expression **************************/
 
@@ -162,7 +161,7 @@ int CmpExp::isCommutative()  { return TRUE; }
 Identifier *CmpExp::opId()   { return Id::cmp; }
 
 Identifier *ArrayExp::opId()	{ return Id::index; }
-
+Identifier *PtrExp::opId()	{ return Id::opStar; }
 
 /************************************
  * Operator overload.
@@ -279,11 +278,11 @@ Expression *BinExp::op_overload(Scope *sc)
 	    fd = s->isFuncDeclaration();
 	    if (fd)
 	    {
-		overloadResolveX(&m, fd, &args2);
+		overloadResolveX(&m, fd, NULL, &args2);
 	    }
 	    else
 	    {   td = s->isTemplateDeclaration();
-		templateResolve(&m, td, sc, loc, NULL, &args2);
+		templateResolve(&m, td, sc, loc, NULL, NULL, &args2);
 	    }
 	}
 	
@@ -294,11 +293,11 @@ Expression *BinExp::op_overload(Scope *sc)
 	    fd = s_r->isFuncDeclaration();
 	    if (fd)
 	    {
-		overloadResolveX(&m, fd, &args1);
+		overloadResolveX(&m, fd, NULL, &args1);
 	    }
 	    else
 	    {   td = s_r->isTemplateDeclaration();
-		templateResolve(&m, td, sc, loc, NULL, &args1);
+		templateResolve(&m, td, sc, loc, NULL, NULL, &args1);
 	    }
 	}
 
@@ -368,11 +367,11 @@ Expression *BinExp::op_overload(Scope *sc)
 		fd = s_r->isFuncDeclaration();
 		if (fd)
 		{
-		    overloadResolveX(&m, fd, &args2);
+		    overloadResolveX(&m, fd, NULL, &args2);
 		}
 		else
 		{   td = s_r->isTemplateDeclaration();
-		    templateResolve(&m, td, sc, loc, NULL, &args2);
+		    templateResolve(&m, td, sc, loc, NULL, NULL, &args2);
 		}
 	    }
 	    lastf = m.lastf;
@@ -382,11 +381,11 @@ Expression *BinExp::op_overload(Scope *sc)
 		fd = s->isFuncDeclaration();
 		if (fd)
 		{
-		    overloadResolveX(&m, fd, &args1);
+		    overloadResolveX(&m, fd, NULL, &args1);
 		}
 		else
 		{   td = s->isTemplateDeclaration();
-		    templateResolve(&m, td, sc, loc, NULL, &args1);
+		    templateResolve(&m, td, sc, loc, NULL, NULL, &args1);
 		}
 	    }
 
@@ -445,7 +444,7 @@ Expression *BinExp::op_overload(Scope *sc)
  * Utility to build a function call out of this reference and argument.
  */
 
-static Expression *build_overload(Loc loc, Scope *sc, Expression *ethis, Expression *earg, Identifier *id)
+Expression *build_overload(Loc loc, Scope *sc, Expression *ethis, Expression *earg, Identifier *id)
 {
     Expression *e;
 
@@ -734,12 +733,12 @@ static int inferApplyArgTypesY(TypeFunction *tf, Arguments *arguments)
 /**************************************
  */
 
-static void templateResolve(Match *m, TemplateDeclaration *td, Scope *sc, Loc loc, Objects *targsi, Expressions *arguments)
+static void templateResolve(Match *m, TemplateDeclaration *td, Scope *sc, Loc loc, Objects *targsi, Expression *ethis, Expressions *arguments)
 {
     FuncDeclaration *fd;
 
     assert(td);
-    fd = td->deduce(sc, loc, targsi, arguments);
+    fd = td->deduceFunctionTemplate(sc, loc, targsi, ethis, arguments);
     if (!fd)
 	return;
     m->anyf = fd;
