@@ -42,53 +42,96 @@ static real_t zero;	// work around DMC bug for now
 
 
 /*************************************
- * If expression is a variable with a const initializer,
+ * If variable has a const initializer,
  * return that initializer.
  */
 
+Expression *expandVar(int result, VarDeclaration *v)
+{
+    //printf("expandVar(result = %d, v = %s)\n", result, v ? v->toChars() : "null");
+    Expression *e = NULL;
+    if (v && (v->isConst() || v->isInvariant() || v->storage_class & STCmanifest))
+    {
+	Type *tb = v->type->toBasetype();
+	if (result & WANTinterpret ||
+	    v->storage_class & STCmanifest ||
+	    (tb->ty != Tsarray && tb->ty != Tstruct)
+	   )
+	{
+	    if (v->init)
+	    {
+		if (v->inuse)
+		    goto L1;
+		Expression *ei = v->init->toExpression();
+		if (!ei)
+		    goto L1;
+		if (ei->op == TOKconstruct || ei->op == TOKblit)
+		{   AssignExp *ae = (AssignExp *)ei;
+		    ei = ae->e2;
+		    if (ei->isConst() != 1 && ei->op != TOKstring)
+			goto L1;
+		    if (ei->type != v->type)
+			goto L1;
+		}
+		if (v->scope)
+		{
+		    v->inuse++;
+		    e = ei->syntaxCopy();
+		    e = e->semantic(v->scope);
+		    e = e->implicitCastTo(v->scope, v->type);
+		    v->scope = NULL;
+		    v->inuse--;
+		}
+		else if (!ei->type)
+		{
+		    goto L1;
+		}
+		else
+		    // Should remove the copy() operation by
+		    // making all mods to expressions copy-on-write
+		    e = ei->copy();
+	    }
+	    else
+	    {
+#if 1
+		goto L1;
+#else
+		// BUG: what if const is initialized in constructor?
+		e = v->type->defaultInit();
+		e->loc = e1->loc;
+#endif
+	    }
+	    if (e->type != v->type)
+	    {
+		e = e->castTo(NULL, v->type);
+	    }
+	    e = e->optimize(result);
+	}
+    }
+L1:
+    //if (e) printf("\te = %s, e->type = %s\n", e->toChars(), e->type->toChars());
+    return e;
+}
+
+
 Expression *fromConstInitializer(int result, Expression *e1)
 {
     //printf("fromConstInitializer(result = %x, %s)\n", result, e1->toChars());
+    //static int xx; if (xx++ == 10) assert(0);
     Expression *e = e1;
     if (e1->op == TOKvar)
     {	VarExp *ve = (VarExp *)e1;
 	VarDeclaration *v = ve->var->isVarDeclaration();
-	if (v && (v->isConst() || v->isInvariant()))
-	{
-	    Type *tb = v->type->toBasetype();
-	    if (result & WANTinterpret ||
-		(tb->ty != Tsarray && tb->ty != Tstruct)
-	       )
-	    {
-		if (v->init)
-		{
-		    if (!v->inuse)
-		    {
-			Expression *ei = v->init->toExpression();
-			if (ei && ei->type)
-			    // Should remove the copy() operation by
-			    // making all mods to expressions copy-on-write
-			    e = ei->copy();
-		    }
-		}
-		else
-		{
-		    e = v->type->defaultInit();
-		    e->loc = e1->loc;
-		}
-		if (e->type != v->type)
-		{
-		    e = e->castTo(NULL, v->type);
-		}
-		if (e->type != e1->type)
-		{   // Type 'paint' operation
-		    e = e->copy();
-		    e->type = e1->type;
-		}
-		e = e->optimize(result);
-		//printf("e = %s, e1->type = %s, v->type = %s, e->type = %s\n", e->toChars(), e1->type->toChars(), v->type->toChars(), e->type->toChars());
+	e = expandVar(result, v);
+	if (e)
+	{   if (e->type != e1->type)
+	    {   // Type 'paint' operation
+		e = e->copy();
+		e->type = e1->type;
 	    }
 	}
+	else
+	    e = e1;
     }
     return e;
 }
@@ -228,7 +271,26 @@ Expression *AddrExp::optimize(int result)
 {   Expression *e;
 
     //printf("AddrExp::optimize(result = %d) %s\n", result, toChars());
-    e1 = e1->optimize(result);
+
+    /* Rewrite &(a,b) as (a,&b)
+     */
+    if (e1->op == TOKcomma)
+    {	CommaExp *ce = (CommaExp *)e1;
+	AddrExp *ae = new AddrExp(loc, ce->e2);
+	ae->type = type;
+	e = new CommaExp(ce->loc, ce->e1, ae);
+	e->type = type;
+	return e->optimize(result);
+    }
+
+    if (e1->op == TOKvar)
+    {	VarExp *ve = (VarExp *)e1;
+	if (ve->var->storage_class & STCmanifest)
+	    e1 = e1->optimize(result);
+    }
+    else
+	e1 = e1->optimize(result);
+
     // Convert &*ex to ex
     if (e1->op == TOKstar)
     {	Expression *ex;
@@ -315,23 +377,46 @@ Expression *PtrExp::optimize(int result)
 	    return e;
     }
 
+    if (e1->op == TOKsymoff)
+    {	SymOffExp *se = (SymOffExp *)e1;
+	VarDeclaration *v = se->var->isVarDeclaration();
+	Expression *e = expandVar(result, v);
+	if (e && e->op == TOKstructliteral)
+	{   StructLiteralExp *sle = (StructLiteralExp *)e;
+	    e = sle->getField(type, se->offset);
+	    if (e != EXP_CANT_INTERPRET)
+		return e;
+	}
+    }
     return this;
 }
 
 Expression *CallExp::optimize(int result)
-{   Expression *e = this;
+{
+    //printf("CallExp::optimize(result = %d) %s\n", result, toChars());
+    Expression *e = this;
 
     e1 = e1->optimize(result);
-    if (e1->op == TOKvar && result & WANTinterpret)
+    if (e1->op == TOKvar)
     {
 	FuncDeclaration *fd = ((VarExp *)e1)->var->isFuncDeclaration();
 	if (fd)
 	{
-	    Expression *eresult = fd->interpret(NULL, arguments);
-	    if (eresult)
-		e = eresult;
+	    enum BUILTIN b = fd->isBuiltin();
+	    if (b)
+	    {
+		e = eval_builtin(b, arguments);
+		if (!e)			// failed
+		    e = this;		// evaluate at runtime
+	    }
 	    else if (result & WANTinterpret)
-		error("cannot evaluate %s at compile time", toChars());
+	    {
+		Expression *eresult = fd->interpret(NULL, arguments);
+		if (eresult && eresult != EXP_VOID_INTERPRET)
+		    e = eresult;
+		else
+		    error("cannot evaluate %s at compile time", toChars());
+	    }
 	}
     }
     return e;
@@ -343,9 +428,11 @@ Expression *CastExp::optimize(int result)
     //printf("CastExp::optimize(result = %d) %s\n", result, toChars());
     //printf("from %s to %s\n", type->toChars(), to->toChars());
     //printf("from %s\n", type->toChars());
+    //printf("e1->type %s\n", e1->type->toChars());
     //printf("type = %p\n", type);
     assert(type);
     enum TOK op1 = e1->op;
+#define X 0
 
     e1 = e1->optimize(result);
     e1 = fromConstInitializer(result, e1);
@@ -356,7 +443,7 @@ Expression *CastExp::optimize(int result)
        )
     {
 	e1 = e1->castTo(NULL, type);
-	//printf(" returning1 %s\n", e1->toChars());
+	if (X) printf(" returning1 %s\n", e1->toChars());
 	return e1;
     }
 
@@ -364,7 +451,7 @@ Expression *CastExp::optimize(int result)
 	e1->type->implicitConvTo(type) >= MATCHconst)
     {
 	e1->type = type;
-	//printf(" returning2 %s\n", e1->toChars());
+	if (X) printf(" returning2 %s\n", e1->toChars());
 	return e1;
     }
 
@@ -376,6 +463,7 @@ Expression *CastExp::optimize(int result)
 	(type->ty == Tpointer || type->ty == Tclass || type->ty == Tarray))
     {
 	e1->type = type;
+	if (X) printf(" returning3 %s\n", e1->toChars());
 	return e1;
     }
 
@@ -391,14 +479,17 @@ Expression *CastExp::optimize(int result)
 	if (cdto->isBaseOf(cdfrom, &offset) && offset == 0)
 	{
 	    e1->type = type;
+	    if (X) printf(" returning4 %s\n", e1->toChars());
 	    return e1;
 	}
     }
 
     // We can convert 'head const' to mutable
-    if (to->constConv(e1->type) >= MATCHconst)
+    if (to->constOf()->equals(e1->type->constOf()))
+//    if (to->constConv(e1->type) >= MATCHconst)
     {
 	e1->type = type;
+	if (X) printf(" returning5 %s\n", e1->toChars());
 	return e1;
     }
 
@@ -423,15 +514,29 @@ Expression *CastExp::optimize(int result)
     }
     else
 	e = this;
-    //printf(" returning3 %s\n", e->toChars());
+    if (X) printf(" returning6 %s\n", e->toChars());
     return e;
+#undef X
 }
 
 Expression *BinExp::optimize(int result)
 {
     //printf("BinExp::optimize(result = %d) %s\n", result, toChars());
-    e1 = e1->optimize(result);
+    if (op != TOKconstruct && op != TOKblit)	// don't replace const variable with its initializer
+	e1 = e1->optimize(result);
     e2 = e2->optimize(result);
+    if (op == TOKshlass || op == TOKshrass || op == TOKushrass)
+    {
+	if (e2->isConst() == 1)
+	{
+	    integer_t i2 = e2->toInteger();
+	    d_uns64 sz = e1->type->size() * 8;
+	    if (i2 < 0 || i2 > sz)
+	    {   error("shift assign by %jd is outside the range 0..%zu", i2, sz);
+		e2 = new IntegerExp(0);
+	    }
+	}
+    }
     return this;
 }
 
@@ -512,55 +617,41 @@ Expression *ModExp::optimize(int result)
     return e;
 }
 
-Expression *ShlExp::optimize(int result)
-{   Expression *e;
+Expression *shift_optimize(int result, BinExp *e, Expression *(*shift)(Type *, Expression *, Expression *))
+{   Expression *ex = e;
 
-    //printf("ShlExp::optimize(result = %d) %s\n", result, toChars());
-    e1 = e1->optimize(result);
-    e2 = e2->optimize(result);
-    e = this;
-    if (e2->isConst() == 1)
+    e->e1 = e->e1->optimize(result);
+    e->e2 = e->e2->optimize(result);
+    if (e->e2->isConst() == 1)
     {
-	integer_t i2 = e2->toInteger();
-	if (i2 < 0 || i2 > e1->type->size() * 8)
-	{   error("shift left by %"PRIdMAX" exceeds %"PRIuSIZE, i2, e2->type->size() * 8);
-	    e2 = new IntegerExp(0);
+	integer_t i2 = e->e2->toInteger();
+	target_size_t sz = e->e1->type->size() * 8;
+	if (i2 < 0 || i2 > sz)
+	{   error("shift left by %"PRIdMAX" exceeds %"PRIuTSIZE, i2, sz);
+	    e->e2 = new IntegerExp(0);
 	}
-	if (e1->isConst() == 1)
-	    e = new IntegerExp(loc, e1->toInteger() << e2->toInteger(), type);
+	if (e->e1->isConst() == 1)
+	    ex = (*shift)(e->type, e->e1, e->e2);
     }
-    return e;
+    return ex;
+}
+
+Expression *ShlExp::optimize(int result)
+{
+    //printf("ShlExp::optimize(result = %d) %s\n", result, toChars());
+    return shift_optimize(result, this, Shl);
 }
 
 Expression *ShrExp::optimize(int result)
-{   Expression *e;
-
-    e1 = e1->optimize(result);
-    e2 = e2->optimize(result);
-
-    if (e1->isConst() == 1 && e2->isConst() == 1)
-    {
-	e = Shr(type, e1, e2);
-    }
-    else
-	e = this;
-    return e;
+{
+    //printf("ShrExp::optimize(result = %d) %s\n", result, toChars());
+    return shift_optimize(result, this, Shr);
 }
 
 Expression *UshrExp::optimize(int result)
-{   Expression *e;
-
-    //printf("UshrExp::optimize() %s\n", toChars());
-    e1 = e1->optimize(result);
-    e2 = e2->optimize(result);
-
-    if (e1->isConst() == 1 && e2->isConst() == 1)
-    {
-	e = Ushr(type, e1, e2);
-    }
-    else
-	e = this;
-    return e;
+{
+    //printf("UshrExp::optimize(result = %d) %s\n", result, toChars());
+    return shift_optimize(result, this, Ushr);
 }
 
 Expression *AndExp::optimize(int result)
@@ -766,13 +857,14 @@ Expression *CmpExp::optimize(int result)
 {   Expression *e;
 
     //printf("CmpExp::optimize() %s\n", toChars());
-    e1 = e1->optimize(result);
-    e2 = e2->optimize(result);
-    if (e1->isConst() == 1 && e2->isConst() == 1)
-    {
-	e = Cmp(op, type, this->e1, this->e2);
-    }
-    else
+    e1 = e1->optimize(WANTvalue | (result & WANTinterpret));
+    e2 = e2->optimize(WANTvalue | (result & WANTinterpret));
+
+    Expression *e1 = fromConstInitializer(result, this->e1);
+    Expression *e2 = fromConstInitializer(result, this->e2);
+
+    e = Cmp(op, type, e1, e2);
+    if (e == EXP_CANT_INTERPRET)
 	e = this;
     return e;
 }
