@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -32,6 +32,7 @@
 #include "id.h"
 #include "hdrgen.h"
 #include "parse.h"
+#include "template.h"
 
 /******************************** Statement ***************************/
 
@@ -119,6 +120,16 @@ int Statement::usesEH()
     return FALSE;
 }
 
+/* Only valid after semantic analysis
+ */
+int Statement::blockExit()
+{
+    printf("Statement::blockExit(%p)\n", this);
+    printf("%s\n", toChars());
+    assert(0);
+    return BEany;
+}
+
 // TRUE if statement may fall off the end without a throw or return
 
 int Statement::fallOffEnd()
@@ -144,7 +155,7 @@ int Statement::comeFrom()
  *	*sfinally	code executed in finally block
  */
 
-void Statement::scopeCode(Statement **sentry, Statement **sexception, Statement **sfinally)
+void Statement::scopeCode(Scope *sc, Statement **sentry, Statement **sexception, Statement **sfinally)
 {
     //printf("Statement::scopeCode()\n");
     //print();
@@ -203,6 +214,25 @@ Statement *ExpStatement::semantic(Scope *sc)
     return this;
 }
 
+int ExpStatement::blockExit()
+{   int result = BEfallthru;
+
+    if (exp)
+    {
+	if (exp->op == TOKhalt)
+	    return BEhalt;
+	if (exp->op == TOKassert)
+	{   AssertExp *a = (AssertExp *)exp;
+
+	    if (a->e1->isBool(FALSE))	// if it's an assert(0)
+		return BEhalt;
+	}
+	if (exp->canThrow())
+	    result |= BEthrow;
+    }
+    return result;
+}
+
 int ExpStatement::fallOffEnd()
 {
     if (exp)
@@ -289,7 +319,7 @@ Statement *DeclarationStatement::syntaxCopy()
     return ds;
 }
 
-void DeclarationStatement::scopeCode(Statement **sentry, Statement **sexception, Statement **sfinally)
+void DeclarationStatement::scopeCode(Scope *sc, Statement **sentry, Statement **sexception, Statement **sfinally)
 {
     //printf("DeclarationStatement::scopeCode()\n");
     //print();
@@ -307,7 +337,7 @@ void DeclarationStatement::scopeCode(Statement **sentry, Statement **sexception,
 	    if (v)
 	    {	Expression *e;
 
-		e = v->callAutoDtor();
+		e = v->callAutoDtor(sc);
 		if (e)
 		{
 		    //printf("dtor is: "); e->print();
@@ -381,7 +411,7 @@ Statement *CompoundStatement::semantic(Scope *sc)
 		Statement *sexception;
 		Statement *sfinally;
 
-		s->scopeCode(&sentry, &sexception, &sfinally);
+		s->scopeCode(sc, &sentry, &sexception, &sfinally);
 		if (sentry)
 		{
 		    sentry = sentry->semantic(sc);
@@ -420,10 +450,7 @@ Statement *CompoundStatement::semantic(Scope *sc)
 			body = new CompoundStatement(0, a);
 			body = new ScopeStatement(0, body);
 
-			static int num;
-			char name[3 + sizeof(num) * 3 + 1];
-			sprintf(name, "__o%d", ++num);
-			Identifier *id = Lexer::idPool(name);
+			Identifier *id = Lexer::uniqueId("__o");
 
 			Statement *handler = new ThrowStatement(0, new IdentifierExp(0, id));
 			handler = new CompoundStatement(0, sexception, handler);
@@ -525,6 +552,31 @@ int CompoundStatement::usesEH()
     return FALSE;
 }
 
+int CompoundStatement::blockExit()
+{
+    //printf("CompoundStatement::blockExit(%p) %d\n", this, statements->dim);
+    int result = BEfallthru;
+    for (size_t i = 0; i < statements->dim; i++)
+    {	Statement *s = (Statement *) statements->data[i];
+	if (s)
+	{
+//printf("result = x%x\n", result);
+//printf("%s\n", s->toChars());
+	    if (!(result & BEfallthru) && !s->comeFrom())
+	    {
+		if (global.params.warnings)
+		{   fprintf(stdmsg, "warning - ");
+		    s->error("statement is not reachable");
+		}
+	    }
+
+	    result &= ~BEfallthru;
+	    result |= s->blockExit();
+	}
+    }
+    return result;
+}
+
 int CompoundStatement::fallOffEnd()
 {   int falloff = TRUE;
 
@@ -535,11 +587,13 @@ int CompoundStatement::fallOffEnd()
 	if (!s)
 	    continue;
 
+#if 0
 	if (!falloff && global.params.warnings && !s->comeFrom())
 	{
 	    fprintf(stdmsg, "warning - ");
 	    s->error("statement is not reachable");
 	}
+#endif
 	falloff = s->fallOffEnd();
     }
     return falloff;
@@ -638,33 +692,37 @@ int UnrolledLoopStatement::hasContinue()
 int UnrolledLoopStatement::usesEH()
 {
     for (size_t i = 0; i < statements->dim; i++)
-    {	Statement *s;
-
-	s = (Statement *) statements->data[i];
+    {	Statement *s = (Statement *) statements->data[i];
 	if (s && s->usesEH())
 	    return TRUE;
     }
     return FALSE;
 }
 
-int UnrolledLoopStatement::fallOffEnd()
-{   int falloff = TRUE;
+int UnrolledLoopStatement::blockExit()
+{
+    int result = BEfallthru;
+    for (size_t i = 0; i < statements->dim; i++)
+    {	Statement *s = (Statement *) statements->data[i];
+	if (s)
+	{
+	    int r = s->blockExit();
+	    result |= r & ~(BEbreak | BEcontinue);
+	}
+    }
+    return result;
+}
 
+int UnrolledLoopStatement::fallOffEnd()
+{
     //printf("UnrolledLoopStatement::fallOffEnd()\n");
     for (size_t i = 0; i < statements->dim; i++)
     {	Statement *s = (Statement *)statements->data[i];
 
-	if (!s)
-	    continue;
-
-	if (!falloff && global.params.warnings && !s->comeFrom())
-	{
-	    fprintf(stdmsg, "warning - ");
-	    s->error("statement is not reachable");
-	}
-	falloff = s->fallOffEnd();
+	if (s)
+	    s->fallOffEnd();
     }
-    return falloff;
+    return TRUE;
 }
 
 int UnrolledLoopStatement::comeFrom()
@@ -725,7 +783,7 @@ Statement *ScopeStatement::semantic(Scope *sc)
 	    Statement *sexception;
 	    Statement *sfinally;
 
-	    statement->scopeCode(&sentry, &sexception, &sfinally);
+	    statement->scopeCode(sc, &sentry, &sexception, &sfinally);
 	    if (sfinally)
 	    {
 		//printf("adding sfinally\n");
@@ -754,6 +812,12 @@ int ScopeStatement::usesEH()
     return statement ? statement->usesEH() : FALSE;
 }
 
+int ScopeStatement::blockExit()
+{
+    //printf("ScopeStatement::blockExit(%p)\n", statement);
+    return statement ? statement->blockExit() : BEfallthru;
+}
+
 int ScopeStatement::fallOffEnd()
 {
     return statement ? statement->fallOffEnd() : TRUE;
@@ -852,6 +916,24 @@ int WhileStatement::usesEH()
     return body ? body->usesEH() : 0;
 }
 
+int WhileStatement::blockExit()
+{
+    //printf("WhileStatement::blockExit(%p)\n", this);
+
+    int result = BEnone;
+    if (condition->canThrow())
+	result |= BEthrow;
+    if (body)
+    {	result |= body->blockExit();
+	if (result & BEbreak)
+	    result |= BEfallthru;
+	result &= ~(BEbreak | BEcontinue);
+    }
+    else
+	result |= BEfallthru;
+    return result;
+}
+
 int WhileStatement::fallOffEnd()
 {
     if (body)
@@ -922,6 +1004,28 @@ int DoStatement::usesEH()
     return body ? body->usesEH() : 0;
 }
 
+int DoStatement::blockExit()
+{   int result;
+
+    if (body)
+    {	result = body->blockExit();
+	if (result & BEbreak)
+	{
+	    if (result == BEbreak)
+		return BEfallthru;
+	    result |= BEfallthru;
+	}
+	if (result & BEcontinue)
+	    result |= BEfallthru;
+	result &= ~(BEbreak | BEcontinue);
+    }
+    else
+	result = BEfallthru;
+    if (result & BEfallthru && condition->canThrow())
+	result |= BEthrow;
+    return result;
+}
+
 int DoStatement::fallOffEnd()
 {
     if (body)
@@ -980,14 +1084,14 @@ Statement *ForStatement::semantic(Scope *sc)
     sc = sc->push(sym);
     if (init)
 	init = init->semantic(sc);
-    if (!condition)
-	// Use a default value
-	condition = new IntegerExp(loc, 1, Type::tboolean);
     sc->noctor++;
-    condition = condition->semantic(sc);
-    condition = resolveProperties(sc, condition);
-    condition = condition->optimize(WANTvalue);
-    condition = condition->checkToBoolean();
+    if (condition)
+    {
+	condition = condition->semantic(sc);
+	condition = resolveProperties(sc, condition);
+	condition = condition->optimize(WANTvalue);
+	condition = condition->checkToBoolean();
+    }
     if (increment)
 	increment = increment->semantic(sc);
 
@@ -1000,14 +1104,14 @@ Statement *ForStatement::semantic(Scope *sc)
     return this;
 }
 
-void ForStatement::scopeCode(Statement **sentry, Statement **sexception, Statement **sfinally)
+void ForStatement::scopeCode(Scope *sc, Statement **sentry, Statement **sexception, Statement **sfinally)
 {
     //printf("ForStatement::scopeCode()\n");
     //print();
     if (init)
-	init->scopeCode(sentry, sexception, sfinally);
+	init->scopeCode(sc, sentry, sexception, sfinally);
     else
-	Statement::scopeCode(sentry, sexception, sfinally);
+	Statement::scopeCode(sc, sentry, sexception, sfinally);
 }
 
 int ForStatement::hasBreak()
@@ -1026,6 +1130,31 @@ int ForStatement::usesEH()
     return (init && init->usesEH()) || body->usesEH();
 }
 
+int ForStatement::blockExit()
+{   int result = BEfallthru;
+
+    if (init)
+    {	result = init->blockExit();
+	if (!(result & BEfallthru))
+	    return result;
+    }
+    if (condition)
+    {	if (condition->canThrow())
+	    result |= BEthrow;
+    }
+    else
+	result &= ~BEfallthru;	// the body must do the exiting
+    if (body)
+    {	int r = body->blockExit();
+	if (r & BEbreak)
+	    result |= BEfallthru;
+	result |= r & ~(BEbreak | BEcontinue);
+    }
+    if (increment && increment->canThrow())
+	result |= BEthrow;
+    return result;
+}
+
 int ForStatement::fallOffEnd()
 {
     if (body)
@@ -1121,6 +1250,7 @@ Statement *ForeachStatement::semantic(Scope *sc)
 
     aggr = aggr->semantic(sc);
     aggr = resolveProperties(sc, aggr);
+    aggr = aggr->optimize(WANTvalue);
     if (!aggr->type)
     {
 	error("invalid foreach aggregate %s", aggr->toChars());
@@ -1188,7 +1318,7 @@ Statement *ForeachStatement::semantic(Scope *sc)
 		}
 		Initializer *ie = new ExpInitializer(0, new IntegerExp(k));
 		VarDeclaration *var = new VarDeclaration(loc, arg->type, arg->ident, ie);
-		var->storage_class |= STCconst;
+		var->storage_class |= STCmanifest;
 		DeclarationExp *de = new DeclarationExp(loc, var);
 		st->push(new ExpStatement(loc, de));
 		arg = (Argument *)arguments->data[1];	// value
@@ -1198,9 +1328,8 @@ Statement *ForeachStatement::semantic(Scope *sc)
 		error("no storage class for value %s", arg->ident->toChars());
 	    Dsymbol *var;
 	    if (te)
-	    {
-		if (e->type->toBasetype()->ty == Tfunction &&
-		    e->op == TOKvar)
+	    {	Type *tb = e->type->toBasetype();
+		if ((tb->ty == Tfunction || tb->ty == Tsarray) && e->op == TOKvar)
 		{   VarExp *ve = (VarExp *)e;
 		    var = new AliasDeclaration(loc, arg->ident, ve->var);
 		}
@@ -1211,8 +1340,6 @@ Statement *ForeachStatement::semantic(Scope *sc)
 		    VarDeclaration *v = new VarDeclaration(loc, arg->type, arg->ident, ie);
 		    if (e->isConst())
 			v->storage_class |= STCconst;
-		    else
-			v->storage_class |= STCfinal;
 		    var = v;
 		}
 	    }
@@ -1291,14 +1418,14 @@ Statement *ForeachStatement::semantic(Scope *sc)
 
 		var = new VarDeclaration(loc, arg->type, arg->ident, NULL);
 		var->storage_class |= STCforeach;
-		var->storage_class |= arg->storageClass & (STCin | STCout | STCref | STCfinal | STCconst);
+		var->storage_class |= arg->storageClass & (STCin | STCout | STCref | STCconst | STCinvariant);
 		if (dim == 2 && i == 0)
 		{   key = var;
-		    var->storage_class |= STCfinal;
+		    //var->storage_class |= STCfinal;
 		}
 		else
-		{   if (!(arg->storageClass & STCref))
-			var->storage_class |= STCfinal;
+		{   //if (!(arg->storageClass & STCref))
+			//var->storage_class |= STCfinal;
 		    value = var;
 		}
 #if 1
@@ -1320,7 +1447,8 @@ Statement *ForeachStatement::semantic(Scope *sc)
 		if (aggr->op == TOKstring)
 		    aggr = aggr->implicitCastTo(sc, value->type->arrayOf());
 		else
-		    error("foreach: %s is not an array of %s", tab->toChars(), value->type->toChars());
+		    error("foreach: %s is not an array of %s",
+			tab->toChars(), value->type->toChars());
 	    }
 
 	    if (key &&
@@ -1395,10 +1523,8 @@ Statement *ForeachStatement::semantic(Scope *sc)
 		    // a reference.
 		    VarDeclaration *v;
 		    Initializer *ie;
-		    char applyArg[10 + sizeof(i)*3 + 1];
 
-		    sprintf(applyArg, "__applyArg%d", i);
-		    id = Lexer::idPool(applyArg);
+		    id = Lexer::uniqueId("__applyArg", i);
 
 		    ie = new ExpInitializer(0, new IdentifierExp(0, id));
 		    v = new VarDeclaration(0, arg->type, arg->ident, ie);
@@ -1413,6 +1539,7 @@ Statement *ForeachStatement::semantic(Scope *sc)
 	    fld->fbody = body;
 	    flde = new FuncExp(loc, fld);
 	    flde = flde->semantic(sc);
+	    fld->tookAddressOf = 0;
 
 	    // Resolve any forward referenced goto's
 	    for (int i = 0; i < gotos.dim; i++)
@@ -1583,6 +1710,19 @@ int ForeachStatement::usesEH()
     return body->usesEH();
 }
 
+int ForeachStatement::blockExit()
+{   int result = BEfallthru;
+
+    if (aggr->canThrow())
+	result |= BEthrow;
+
+    if (body)
+    {
+	result |= body->blockExit() & ~(BEbreak | BEcontinue);
+    }
+    return result;
+}
+
 int ForeachStatement::fallOffEnd()
 {
     if (body)
@@ -1730,6 +1870,21 @@ int ForeachRangeStatement::usesEH()
     return body->usesEH();
 }
 
+int ForeachRangeStatement::blockExit()
+{   int result = BEfallthru;
+
+    if (lwr && lwr->canThrow())
+	result |= BEthrow;
+    else if (upr && upr->canThrow())
+	result |= BEthrow;
+
+    if (body)
+    {
+	result |= body->blockExit() & ~(BEbreak | BEcontinue);
+    }
+    return result;
+}
+
 int ForeachRangeStatement::fallOffEnd()
 {
     if (body)
@@ -1853,6 +2008,25 @@ int IfStatement::usesEH()
     return (ifbody && ifbody->usesEH()) || (elsebody && elsebody->usesEH());
 }
 
+int IfStatement::blockExit()
+{
+    //printf("IfStatement::blockExit(%p)\n", this);
+
+    int result = BEnone;
+    if (condition->canThrow())
+	result |= BEthrow;
+    if (ifbody)
+	result |= ifbody->blockExit();
+    else
+	result |= BEfallthru;
+    if (elsebody)
+	result |= elsebody->blockExit();
+    else
+	result |= BEfallthru;
+    //printf("IfStatement::blockExit(%p) = x%x\n", this, result);
+    return result;
+}
+
 int IfStatement::fallOffEnd()
 {
     if (!ifbody || ifbody->fallOffEnd() ||
@@ -2028,6 +2202,26 @@ Statement *PragmaStatement::semantic(Scope *sc)
 	    }
 	}
     }
+    else if (ident == Id::startaddress)
+    {
+	if (!args || args->dim != 1)
+	    error("function name expected for start address");
+	else
+	{
+	    Expression *e = (Expression *)args->data[0];
+	    e = e->semantic(sc);
+	    e = e->optimize(WANTvalue | WANTinterpret);
+	    args->data[0] = (void *)e;
+	    Dsymbol *sa = getDsymbol(e);
+	    if (!sa || !sa->isFuncDeclaration())
+		error("function name expected for start address, not '%s'", e->toChars());
+	    if (body)
+	    {
+		body = body->semantic(sc);
+	    }
+	    return this;
+	}
+    }
     else
         error("unrecognized pragma(%s)", ident->toChars());
 
@@ -2043,6 +2237,18 @@ int PragmaStatement::usesEH()
     return body && body->usesEH();
 }
 
+int PragmaStatement::blockExit()
+{
+    int result = BEfallthru;
+#if 0 // currently, no code is generated for Pragma's, so it's just fallthru
+    if (arrayExpressionCanThrow(args))
+	result |= BEthrow;
+    if (body)
+	result |= body->blockExit();
+#endif
+    return result;
+}
+
 int PragmaStatement::fallOffEnd()
 {
     if (body)
@@ -2113,8 +2319,10 @@ SwitchStatement::SwitchStatement(Loc loc, Expression *c, Statement *b)
     condition = c;
     body = b;
     sdefault = NULL;
+    tf = NULL;
     cases = NULL;
     hasNoDefault = 0;
+    hasVars = 0;
 }
 
 Statement *SwitchStatement::syntaxCopy()
@@ -2127,6 +2335,7 @@ Statement *SwitchStatement::syntaxCopy()
 Statement *SwitchStatement::semantic(Scope *sc)
 {
     //printf("SwitchStatement::semantic(%p)\n", this);
+    tf = sc->tf;
     assert(!cases);		// ensure semantic() is only run once
     condition = condition->semantic(sc);
     condition = resolveProperties(sc, condition);
@@ -2143,6 +2352,7 @@ Statement *SwitchStatement::semantic(Scope *sc)
     {	condition = condition->integralPromotions(sc);
 	condition->checkIntegral();
     }
+    condition = condition->optimize(WANTvalue);
 
     sc = sc->push();
     sc->sbreak = this;
@@ -2228,6 +2438,24 @@ int SwitchStatement::usesEH()
     return body ? body->usesEH() : 0;
 }
 
+int SwitchStatement::blockExit()
+{   int result = BEnone;
+    if (condition->canThrow())
+	result |= BEthrow;
+
+    if (body)
+    {	result |= body->blockExit();
+	if (result & BEbreak)
+	{   result |= BEfallthru;
+	    result &= ~BEbreak;
+	}
+    }
+    else
+	result |= BEfallthru;
+
+    return result;
+}
+
 int SwitchStatement::fallOffEnd()
 {
     if (body)
@@ -2264,6 +2492,7 @@ CaseStatement::CaseStatement(Loc loc, Expression *exp, Statement *s)
 {
     this->exp = exp;
     this->statement = s;
+    index = 0;
     cblock = NULL;
 }
 
@@ -2279,17 +2508,33 @@ Statement *CaseStatement::semantic(Scope *sc)
     //printf("CaseStatement::semantic() %s\n", toChars());
     exp = exp->semantic(sc);
     if (sw)
-    {	int i;
-
+    {
 	exp = exp->implicitCastTo(sc, sw->condition->type);
 	exp = exp->optimize(WANTvalue | WANTinterpret);
+
+	/* This is where variables are allowed as case expressions.
+	 */
+	if (exp->op == TOKvar)
+	{   VarExp *ve = (VarExp *)exp;
+	    VarDeclaration *v = ve->var->isVarDeclaration();
+	    Type *t = exp->type->toBasetype();
+	    if (v && (t->isintegral() || t->ty == Tclass))
+	    {	/* Flag that we need to do special code generation
+		 * for this, i.e. generate a sequence of if-then-else
+		 */
+		sw->hasVars = 1;
+		goto L1;
+	    }
+	}
+
 	if (exp->op != TOKstring && exp->op != TOKint64)
 	{
 	    error("case must be a string or an integral constant, not %s", exp->toChars());
 	    exp = new IntegerExp(0);
 	}
 
-	for (i = 0; i < sw->cases->dim; i++)
+    L1:
+	for (int i = 0; i < sw->cases->dim; i++)
 	{
 	    CaseStatement *cs = (CaseStatement *)sw->cases->data[i];
 
@@ -2303,7 +2548,7 @@ Statement *CaseStatement::semantic(Scope *sc)
 	sw->cases->push(this);
 
 	// Resolve any goto case's with no exp to this case statement
-	for (i = 0; i < sw->gotoCases.dim; i++)
+	for (int i = 0; i < sw->gotoCases.dim; i++)
 	{
 	    GotoCaseStatement *gcs = (GotoCaseStatement *)sw->gotoCases.data[i];
 
@@ -2313,6 +2558,9 @@ Statement *CaseStatement::semantic(Scope *sc)
 		sw->gotoCases.remove(i);	// remove from array
 	    }
 	}
+
+	if (sc->sw->tf != sc->tf)
+	    error("switch and case are in different finally blocks");
     }
     else
 	error("case not in switch statement");
@@ -2333,6 +2581,12 @@ int CaseStatement::usesEH()
     return statement->usesEH();
 }
 
+int CaseStatement::blockExit()
+{
+    // Assume the worst
+    return BEany;
+}
+
 int CaseStatement::fallOffEnd()
 {
     return statement->fallOffEnd();
@@ -2371,6 +2625,7 @@ Statement *DefaultStatement::syntaxCopy()
 
 Statement *DefaultStatement::semantic(Scope *sc)
 {
+    //printf("DefaultStatement::semantic()\n");
     if (sc->sw)
     {
 	if (sc->sw->sdefault)
@@ -2378,6 +2633,9 @@ Statement *DefaultStatement::semantic(Scope *sc)
 	    error("switch statement already has a default");
 	}
 	sc->sw->sdefault = this;
+
+	if (sc->sw->tf != sc->tf)
+	    error("switch and default are in different finally blocks");
     }
     else
 	error("default not in switch statement");
@@ -2390,6 +2648,12 @@ int DefaultStatement::usesEH()
     return statement->usesEH();
 }
 
+int DefaultStatement::blockExit()
+{
+    // Assume the worst
+    return BEany;
+}
+
 int DefaultStatement::fallOffEnd()
 {
     return statement->fallOffEnd();
@@ -2428,6 +2692,11 @@ Statement *GotoDefaultStatement::semantic(Scope *sc)
     return this;
 }
 
+int GotoDefaultStatement::blockExit()
+{
+    return BEgoto;
+}
+
 int GotoDefaultStatement::fallOffEnd()
 {
     return FALSE;
@@ -2473,6 +2742,11 @@ Statement *GotoCaseStatement::semantic(Scope *sc)
     return this;
 }
 
+int GotoCaseStatement::blockExit()
+{
+    return BEgoto;
+}
+
 int GotoCaseStatement::fallOffEnd()
 {
     return FALSE;
@@ -2496,6 +2770,11 @@ SwitchErrorStatement::SwitchErrorStatement(Loc loc)
 {
 }
 
+int SwitchErrorStatement::blockExit()
+{
+    return BEthrow;
+}
+
 int SwitchErrorStatement::fallOffEnd()
 {
     return FALSE;
@@ -2590,9 +2869,14 @@ Statement *ReturnStatement::semantic(Scope *sc)
 
 	    if (!v || v->isOut() || v->isRef())
 		fd->nrvo_can = 0;
+	    else if (tbret->ty == Tstruct && ((TypeStruct *)tbret)->sym->dtor)
+		// Struct being returned has destructors
+		fd->nrvo_can = 0;
 	    else if (fd->nrvo_var == NULL)
 	    {	if (!v->isDataseg() && !v->isParameter() && v->toParent2() == fd)
+		{   //printf("Setting nrvo to %s\n", v->toChars());
 		    fd->nrvo_var = v;
+		}
 		else
 		    fd->nrvo_can = 0;
 	    }
@@ -2767,6 +3051,14 @@ Statement *ReturnStatement::semantic(Scope *sc)
     return this;
 }
 
+int ReturnStatement::blockExit()
+{   int result = BEreturn;
+
+    if (exp && exp->canThrow())
+	result |= BEthrow;
+    return result;
+}
+
 int ReturnStatement::fallOffEnd()
 {
     return FALSE;
@@ -2797,6 +3089,7 @@ Statement *BreakStatement::syntaxCopy()
 
 Statement *BreakStatement::semantic(Scope *sc)
 {
+    //printf("BreakStatement::semantic()\n");
     // If:
     //	break Identifier;
     if (ident)
@@ -2855,6 +3148,12 @@ Statement *BreakStatement::semantic(Scope *sc)
     return this;
 }
 
+int BreakStatement::blockExit()
+{
+    //printf("BreakStatement::blockExit(%p) = x%x\n", this, ident ? BEgoto : BEbreak);
+    return ident ? BEgoto : BEbreak;
+}
+
 int BreakStatement::fallOffEnd()
 {
     return FALSE;
@@ -2954,6 +3253,11 @@ Statement *ContinueStatement::semantic(Scope *sc)
     return this;
 }
 
+int ContinueStatement::blockExit()
+{
+    return ident ? BEgoto : BEcontinue;
+}
+
 int ContinueStatement::fallOffEnd()
 {
     return FALSE;
@@ -3033,6 +3337,11 @@ int SynchronizedStatement::usesEH()
     return TRUE;
 }
 
+int SynchronizedStatement::blockExit()
+{
+    return body ? body->blockExit() : BEfallthru;
+}
+
 int SynchronizedStatement::fallOffEnd()
 {
     return body ? body->fallOffEnd() : TRUE;
@@ -3143,6 +3452,18 @@ int WithStatement::usesEH()
     return body ? body->usesEH() : 0;
 }
 
+int WithStatement::blockExit()
+{
+    int result = BEnone;
+    if (exp->canThrow())
+	result = BEthrow;
+    if (body)
+	result |= body->blockExit();
+    else
+	result |= BEfallthru;
+    return result;
+}
+
 int WithStatement::fallOffEnd()
 {
     return body ? body->fallOffEnd() : TRUE;
@@ -3176,14 +3497,14 @@ Statement *TryCatchStatement::semantic(Scope *sc)
 {
     body = body->semanticScope(sc, NULL /*this*/, NULL);
 
-    for (int i = 0; i < catches->dim; i++)
-    {   Catch *c;
-
-	c = (Catch *)catches->data[i];
+    /* Even if body is NULL, still do semantic analysis on catches
+     */
+    for (size_t i = 0; i < catches->dim; i++)
+    {   Catch *c = (Catch *)catches->data[i];
 	c->semantic(sc);
 
 	// Determine if current catch 'hides' any previous catches
-	for (int j = 0; j < i; j++)
+	for (size_t j = 0; j < i; j++)
 	{   Catch *cj = (Catch *)catches->data[j];
 	    char *si = c->loc.toChars();
 	    char *sj = cj->loc.toChars();
@@ -3192,6 +3513,10 @@ Statement *TryCatchStatement::semantic(Scope *sc)
 		error("catch at %s hides catch at %s", sj, si);
 	}
     }
+
+    if (!body)
+	return NULL;
+
     return this;
 }
 
@@ -3205,6 +3530,20 @@ int TryCatchStatement::usesEH()
     return TRUE;
 }
 
+int TryCatchStatement::blockExit()
+{   int result;
+
+    assert(body);
+    result = body->blockExit();
+
+    for (size_t i = 0; i < catches->dim; i++)
+    {
+        Catch *c = (Catch *)catches->data[i];
+        result |= c->blockExit();
+    }
+    return result;
+}
+
 int TryCatchStatement::fallOffEnd()
 {
     int result = FALSE;
@@ -3227,8 +3566,7 @@ void TryCatchStatement::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
     buf->writenl();
     if (body)
         body->toCBuffer(buf, hgs);
-    int i;
-    for (i = 0; i < catches->dim; i++)
+    for (size_t i = 0; i < catches->dim; i++)
     {
         Catch *c = (Catch *)catches->data[i];
         c->toCBuffer(buf, hgs);
@@ -3294,6 +3632,11 @@ void Catch::semantic(Scope *sc)
     sc->pop();
 }
 
+int Catch::blockExit()
+{
+    return handler ? handler->blockExit() : BEfallthru;
+}
+
 void Catch::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 {
     buf->writestring("catch");
@@ -3305,7 +3648,8 @@ void Catch::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
     buf->writenl();
     buf->writebyte('{');
     buf->writenl();
-    handler->toCBuffer(buf, hgs);
+    if (handler)
+	handler->toCBuffer(buf, hgs);
     buf->writebyte('}');
     buf->writenl();
 }
@@ -3336,6 +3680,14 @@ Statement *TryFinallyStatement::semantic(Scope *sc)
     sc->scontinue = NULL;	// no break or continue out of finally block
     finalbody = finalbody->semantic(sc);
     sc->pop();
+    if (!body)
+	return finalbody;
+    if (!finalbody)
+	return body;
+    if (body->blockExit() == BEfallthru)
+    {	Statement *s = new CompoundStatement(loc, body, finalbody);
+	return s;
+    }
     return this;
 }
 
@@ -3364,6 +3716,12 @@ int TryFinallyStatement::usesEH()
     return TRUE;
 }
 
+int TryFinallyStatement::blockExit()
+{
+    int result = body->blockExit();
+    return result;
+}
+
 int TryFinallyStatement::fallOffEnd()
 {   int result;
 
@@ -3395,6 +3753,11 @@ Statement *OnScopeStatement::semantic(Scope *sc)
     return this;
 }
 
+int OnScopeStatement::blockExit()
+{   // At this point, this statement is just an empty placeholder
+    return BEfallthru;
+}
+
 void OnScopeStatement::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 {
     buf->writestring(Token::toChars(tok));
@@ -3404,10 +3767,10 @@ void OnScopeStatement::toCBuffer(OutBuffer *buf, HdrGenState *hgs)
 
 int OnScopeStatement::usesEH()
 {
-    return (tok != TOKon_scope_success);
+    return 1;
 }
 
-void OnScopeStatement::scopeCode(Statement **sentry, Statement **sexception, Statement **sfinally)
+void OnScopeStatement::scopeCode(Scope *sc, Statement **sentry, Statement **sexception, Statement **sfinally)
 {
     //printf("OnScopeStatement::scopeCode()\n");
     //print();
@@ -3431,10 +3794,7 @@ void OnScopeStatement::scopeCode(Statement **sentry, Statement **sexception, Sta
 	     *	sexception:    x = 1;
 	     *	sfinally: if (!x) statement;
 	     */
-	    static int num;
-	    char name[5 + sizeof(num) * 3 + 1];
-	    sprintf(name, "__osf%d", ++num);
-	    Identifier *id = Lexer::idPool(name);
+	    Identifier *id = Lexer::uniqueId("__os");
 
 	    ExpInitializer *ie = new ExpInitializer(loc, new IntegerExp(0));
 	    VarDeclaration *v = new VarDeclaration(loc, Type::tint32, id, ie);
@@ -3486,6 +3846,11 @@ Statement *ThrowStatement::semantic(Scope *sc)
     return this;
 }
 
+int ThrowStatement::blockExit()
+{
+    return BEthrow;  // obviously
+}
+
 int ThrowStatement::fallOffEnd()
 {
     return FALSE;
@@ -3538,6 +3903,11 @@ Statements *VolatileStatement::flatten(Scope *sc)
     return a;
 }
 
+int VolatileStatement::blockExit()
+{
+    return statement ? statement->blockExit() : BEfallthru;
+}
+
 int VolatileStatement::fallOffEnd()
 {
     return statement ? statement->fallOffEnd() : TRUE;
@@ -3599,6 +3969,12 @@ Statement *GotoStatement::semantic(Scope *sc)
     return this;
 }
 
+int GotoStatement::blockExit()
+{
+    //printf("GotoStatement::blockExit(%p)\n", this);
+    return BEgoto;
+}
+
 int GotoStatement::fallOffEnd()
 {
     return FALSE;
@@ -3680,6 +4056,12 @@ int LabelStatement::usesEH()
     return statement ? statement->usesEH() : FALSE;
 }
 
+int LabelStatement::blockExit()
+{
+    //printf("LabelStatement::blockExit(%p)\n", this);
+    return statement ? statement->blockExit() : BEfallthru;
+}
+
 int LabelStatement::fallOffEnd()
 {
     return statement ? statement->fallOffEnd() : TRUE;
