@@ -1,6 +1,6 @@
 
 // Compiler implementation of the D programming language
-// Copyright (c) 1999-2007 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -43,6 +43,7 @@ struct AsmStatement;
 struct GotoStatement;
 struct ScopeStatement;
 struct TryCatchStatement;
+struct TryFinallyStatement;
 struct HdrGenState;
 struct InterState;
 
@@ -60,6 +61,21 @@ struct elem;
 #endif
 struct code;
 
+/* How a statement exits
+ */
+enum BE
+{
+    BEnone =	 0,
+    BEfallthru = 1,
+    BEthrow =    2,
+    BEreturn =   4,
+    BEgoto =     8,
+    BEhalt =	 0x10,
+    BEbreak =	 0x20,
+    BEcontinue = 0x40,
+    BEany = (BEfallthru | BEthrow | BEreturn | BEgoto | BEhalt),
+};
+
 struct Statement : Object
 {
     Loc loc;
@@ -85,8 +101,9 @@ struct Statement : Object
     virtual int hasContinue();
     virtual int usesEH();
     virtual int fallOffEnd();
+    virtual int blockExit();
     virtual int comeFrom();
-    virtual void scopeCode(Statement **sentry, Statement **sexit, Statement **sfinally);
+    virtual void scopeCode(Scope *sc, Statement **sentry, Statement **sexit, Statement **sfinally);
     virtual Statements *flatten(Scope *sc);
     virtual Expression *interpret(InterState *istate);
 
@@ -114,6 +131,7 @@ struct ExpStatement : Statement
     Statement *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
     int fallOffEnd();
+    int blockExit();
 
     int inlineCost(InlineCostState *ics);
     Expression *doInline(InlineDoState *ids);
@@ -141,7 +159,7 @@ struct DeclarationStatement : ExpStatement
     DeclarationStatement(Loc loc, Expression *exp);
     Statement *syntaxCopy();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
-    void scopeCode(Statement **sentry, Statement **sexit, Statement **sfinally);
+    void scopeCode(Scope *sc, Statement **sentry, Statement **sexit, Statement **sfinally);
 
     DeclarationStatement *isDeclarationStatement() { return this; }
 };
@@ -156,6 +174,7 @@ struct CompoundStatement : Statement
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     Statement *semantic(Scope *sc);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Statements *flatten(Scope *sc);
@@ -184,6 +203,7 @@ struct UnrolledLoopStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -208,6 +228,7 @@ struct ScopeStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -228,6 +249,7 @@ struct WhileStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -249,6 +271,7 @@ struct DoStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -269,10 +292,11 @@ struct ForStatement : Statement
     ForStatement(Loc loc, Statement *init, Expression *condition, Expression *increment, Statement *body);
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
-    void scopeCode(Statement **sentry, Statement **sexit, Statement **sfinally);
+    void scopeCode(Scope *sc, Statement **sentry, Statement **sexit, Statement **sfinally);
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -304,6 +328,7 @@ struct ForeachStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -331,6 +356,7 @@ struct ForeachRangeStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -356,6 +382,7 @@ struct IfStatement : Statement
     Expression *interpret(InterState *istate);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     IfStatement *isIfStatement() { return this; }
 
@@ -391,9 +418,12 @@ struct PragmaStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
 
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
+
+    void toIR(IRState *irs);
 };
 
 struct StaticAssertStatement : Statement
@@ -411,17 +441,20 @@ struct SwitchStatement : Statement
 {
     Expression *condition;
     Statement *body;
-    DefaultStatement *sdefault;
 
+    DefaultStatement *sdefault;
+    TryFinallyStatement *tf;
     Array gotoCases;		// array of unresolved GotoCaseStatement's
     Array *cases;		// array of CaseStatement's
     int hasNoDefault;		// !=0 if no default statement
+    int hasVars;		// !=0 if has variable case values
 
     SwitchStatement(Loc loc, Expression *c, Statement *b);
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     int hasBreak();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     Expression *interpret(InterState *istate);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
@@ -443,6 +476,7 @@ struct CaseStatement : Statement
     Statement *semantic(Scope *sc);
     int compare(Object *obj);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -464,6 +498,7 @@ struct DefaultStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -482,6 +517,7 @@ struct GotoDefaultStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -497,6 +533,7 @@ struct GotoCaseStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -506,6 +543,7 @@ struct GotoCaseStatement : Statement
 struct SwitchErrorStatement : Statement
 {
     SwitchErrorStatement(Loc loc);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -520,6 +558,7 @@ struct ReturnStatement : Statement
     Statement *syntaxCopy();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     Statement *semantic(Scope *sc);
+    int blockExit();
     int fallOffEnd();
     Expression *interpret(InterState *istate);
 
@@ -540,6 +579,7 @@ struct BreakStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -554,6 +594,7 @@ struct ContinueStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     Expression *interpret(InterState *istate);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -571,6 +612,7 @@ struct SynchronizedStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -593,6 +635,7 @@ struct WithStatement : Statement
     Statement *semantic(Scope *sc);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
 
     Statement *inlineScan(InlineScanState *iss);
@@ -610,6 +653,7 @@ struct TryCatchStatement : Statement
     Statement *semantic(Scope *sc);
     int hasBreak();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
 
     Statement *inlineScan(InlineScanState *iss);
@@ -630,6 +674,7 @@ struct Catch : Object
     Catch(Loc loc, Type *t, Identifier *id, Statement *handler);
     Catch *syntaxCopy();
     void semantic(Scope *sc);
+    int blockExit();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 };
 
@@ -645,6 +690,7 @@ struct TryFinallyStatement : Statement
     int hasBreak();
     int hasContinue();
     int usesEH();
+    int blockExit();
     int fallOffEnd();
 
     Statement *inlineScan(InlineScanState *iss);
@@ -659,10 +705,11 @@ struct OnScopeStatement : Statement
 
     OnScopeStatement(Loc loc, TOK tok, Statement *statement);
     Statement *syntaxCopy();
+    int blockExit();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
     Statement *semantic(Scope *sc);
     int usesEH();
-    void scopeCode(Statement **sentry, Statement **sexit, Statement **sfinally);
+    void scopeCode(Scope *sc, Statement **sentry, Statement **sexit, Statement **sfinally);
 
     void toIR(IRState *irs);
 };
@@ -675,6 +722,7 @@ struct ThrowStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
+    int blockExit();
     int fallOffEnd();
 
     Statement *inlineScan(InlineScanState *iss);
@@ -690,6 +738,7 @@ struct VolatileStatement : Statement
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
     Statements *flatten(Scope *sc);
+    int blockExit();
     int fallOffEnd();
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
 
@@ -707,6 +756,7 @@ struct GotoStatement : Statement
     GotoStatement(Loc loc, Identifier *ident);
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
+    int blockExit();
     int fallOffEnd();
     Expression *interpret(InterState *istate);
 
@@ -728,6 +778,7 @@ struct LabelStatement : Statement
     Statement *semantic(Scope *sc);
     Statements *flatten(Scope *sc);
     int usesEH();
+    int blockExit();
     int fallOffEnd();
     int comeFrom();
     Expression *interpret(InterState *istate);
@@ -761,6 +812,7 @@ struct AsmStatement : Statement
     AsmStatement(Loc loc, Token *tokens);
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
+    int blockExit();
     int comeFrom();
 
     void toCBuffer(OutBuffer *buf, HdrGenState *hgs);
@@ -785,6 +837,7 @@ struct ExtAsmStatement : Statement
 	Expressions *argConstraints, int nOutputArgs, Expressions *clobbers);
     Statement *syntaxCopy();
     Statement *semantic(Scope *sc);
+    int blockExit();
     void toIR(IRState *irs);
 };
 
