@@ -1,5 +1,5 @@
 
-// Copyright (c) 1999-2004 by Digital Mars
+// Copyright (c) 1999-2008 by Digital Mars
 // All Rights Reserved
 // written by Walter Bright
 // http://www.digitalmars.com
@@ -98,9 +98,11 @@ static void find_module_deps(Module * the_module, Array * out_deps)
 
 void Module::genmoduleinfo()
 {
+    //printf("Module::genmoduleinfo() %s\n", toChars());
+
     Symbol *msym = toSymbol();
     unsigned offset;
-    unsigned sizeof_ModuleInfo = 13 * PTRSIZE;
+    unsigned sizeof_ModuleInfo = 14 * PTRSIZE;
 
     //////////////////////////////////////////////
 
@@ -119,6 +121,7 @@ void Module::genmoduleinfo()
 	    void *dtor;
 	    void *unitTest;
 	    const(MemberInfo[]) function(string) xgetMembers;	// module getMembers() function
+	    void *ictor;
        }
      */
     dt_t *dt = NULL;
@@ -126,7 +129,9 @@ void Module::genmoduleinfo()
     if (moduleinfo)
 	dtxoff(&dt, moduleinfo->toVtblSymbol(), 0, TYnptr); // vtbl for ModuleInfo
     else
+    {	//printf("moduleinfo is null\n");
 	dtdword(&dt, 0);		// BUG: should be an assert()
+    }
     dtdword(&dt, 0);			// monitor
 
     // name[]
@@ -136,14 +141,11 @@ void Module::genmoduleinfo()
     dtabytes(&dt, TYnptr, 0, namelen + 1, name);
 
     ClassDeclarations aclasses;
-    int i;
 
     //printf("members->dim = %d\n", members->dim);
-    for (i = 0; i < members->dim; i++)
-    {
-	Dsymbol *member;
+    for (int i = 0; i < members->dim; i++)
+    {	Dsymbol *member = (Dsymbol *)members->data[i];
 
-	member = (Dsymbol *)members->data[i];
 	//printf("\tmember '%s'\n", member->toChars());
 	member->addLocalClass(&aclasses);
     }
@@ -162,7 +164,7 @@ void Module::genmoduleinfo()
     else
 #endif
 #endif
-    for (i = 0; i < aimports.dim; i++)
+    for (int i = 0; i < aimports.dim; i++)
     {	Module *m = (Module *)aimports.data[i];
 	if (!m->needModuleInfo())
 	    aimports_dim--;
@@ -181,9 +183,9 @@ void Module::genmoduleinfo()
 	dtdword(&dt, 0);
 
     if (needmoduleinfo)
-	dti32(&dt, 0, true);		// flags (4 means MIstandalone)
+	dti32(&dt, 8|0, true);		// flags (4 means MIstandalone)
     else
-	dti32(&dt, 4, true);		// flags (4 means MIstandalone)
+	dti32(&dt, 8|4, true);		// flags (4 means MIstandalone)
 
     if (sctor)
 	dtxoff(&dt, sctor, 0, TYnptr);
@@ -206,6 +208,11 @@ void Module::genmoduleinfo()
     else
 	dtdword(&dt, 0);			// xgetMembers
 
+    if (sictor)
+	dtxoff(&dt, sictor, 0, TYnptr);
+    else
+	dtdword(&dt, 0);
+
     //////////////////////////////////////////////
 
 #ifdef IN_GCC
@@ -224,11 +231,9 @@ void Module::genmoduleinfo()
     else
 #endif
 #endif
-    for (i = 0; i < aimports.dim; i++)
-    {
-	Module *m;
+    for (int i = 0; i < aimports.dim; i++)
+    {	Module *m = (Module *)aimports.data[i];
 
-	m = (Module *)aimports.data[i];
 	if (m->needModuleInfo())
 	{   Symbol *s = m->toSymbol();
 	    s->Sflags |= SFLweak;
@@ -236,11 +241,9 @@ void Module::genmoduleinfo()
 	}
     }
 
-    for (i = 0; i < aclasses.dim; i++)
+    for (int i = 0; i < aclasses.dim; i++)
     {
-	ClassDeclaration *cd;
-
-	cd = (ClassDeclaration *)aclasses.data[i];
+	ClassDeclaration *cd = (ClassDeclaration *)aclasses.data[i];
 	dtxoff(&dt, cd->toSymbol(), 0, TYnptr);
     }
 
@@ -257,7 +260,7 @@ void Module::genmoduleinfo()
 }
 
 /* ================================================================== */
-void Dsymbol::toObjFile()
+void Dsymbol::toObjFile(int multiobj)
 {
     //printf("Dsymbol::toObjFile('%s')\n", toChars());
     // ignore
@@ -277,7 +280,7 @@ void Dsymbol::toObjFile()
 		ds = ((DsymbolExp *) o)->s;
 		d = ds->isDeclaration();
 		if (d)
-		    d->toObjFile();
+		    d->toObjFile(multiobj);
 	    }
 		
 	}		
@@ -286,11 +289,10 @@ void Dsymbol::toObjFile()
 }
 /* ================================================================== */
 
-void ClassDeclaration::toObjFile()
+void ClassDeclaration::toObjFile(int multiobj)
 {   unsigned i;
     unsigned offset;
     Symbol *sinit;
-    Symbol *sdtor;
     enum_SC scclass;
 
     //printf("ClassDeclaration::toObjFile('%s')\n", toChars());
@@ -314,11 +316,12 @@ void ClassDeclaration::toObjFile()
 	Dsymbol *member;
 
 	member = (Dsymbol *)members->data[i];
-	member->toObjFile();
+	member->toObjFile(0);
     }
 
+#if 0
     // Build destructor by aggregating dtors[]
-#ifndef IN_GCC
+    Symbol *sdtor;
     switch (dtors.dim)
     {	case 0:
 	    // No destructors for this class
@@ -375,8 +378,6 @@ void ClassDeclaration::toObjFile()
 	    writefunc(sdtor);
 	}
     }
-#else
-    sdtor = d_gcc_aggregate_dtors(this);
 #endif
 
     // Generate C symbols
@@ -466,8 +467,8 @@ void ClassDeclaration::toObjFile()
 	dtdword(&dt, 0);
 
     // destructor
-    if (sdtor)
-	dtxoff(&dt, sdtor, 0, TYnptr);
+    if (dtor)
+	dtxoff(&dt, dtor->toSymbol(), 0, TYnptr);
     else
 	dtdword(&dt, 0);
 
@@ -478,7 +479,7 @@ void ClassDeclaration::toObjFile()
 	dtdword(&dt, 0);
 
     // flags
-    int flags = 16 | 4 | com;
+    int flags = 16 | 4 | isCOMclass();
     if (ctor)
 	flags |= 8;
     for (ClassDeclaration *cd = this; cd; cd = cd->baseClass)
@@ -717,7 +718,10 @@ void ClassDeclaration::toObjFile()
 	{
 	    if (isFuncHidden(fd))
 	    {
-		//printf("%s %s is hidden in %s\n", fd->toParent()->toChars(), fd->toChars(), toChars());
+		if (global.params.warnings)
+		{   fprintf(stdmsg, "warning - ");
+		    error("%s %s is hidden in %s\n", fd->toParent()->toChars(), fd->toChars(), toChars());
+		}
 		dtxoff(&dt, rtlsym[RTLSYM_DHIDDENFUNC], 0, TYnptr);
 	    }
 	    else
@@ -810,7 +814,7 @@ unsigned ClassDeclaration::baseVtblOffset(BaseClass *bc)
 
 /* ================================================================== */
 
-void InterfaceDeclaration::toObjFile()
+void InterfaceDeclaration::toObjFile(int multiobj)
 {   unsigned i;
     unsigned offset;
     Symbol *sinit;
@@ -836,7 +840,7 @@ void InterfaceDeclaration::toObjFile()
 
 	member = (Dsymbol *)members->data[i];
 	if (!member->isFuncDeclaration())
-	    member->toObjFile();
+	    member->toObjFile(0);
     }
 
     // Generate C symbols
@@ -911,7 +915,7 @@ void InterfaceDeclaration::toObjFile()
     dtdword(&dt, 0);
 
     // flags
-    dti32(&dt, 4 | com, true);
+    dti32(&dt, 4 | isCOMinterface(), true);
 
     // deallocator
     dtdword(&dt, 0);
@@ -958,7 +962,7 @@ void InterfaceDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void StructDeclaration::toObjFile()
+void StructDeclaration::toObjFile(int multiobj)
 {   unsigned i;
 
     //printf("StructDeclaration::toObjFile('%s')\n", toChars());
@@ -1018,23 +1022,30 @@ void StructDeclaration::toObjFile()
 	    Dsymbol *member;
 
 	    member = (Dsymbol *)members->data[i];
-	    member->toObjFile();
+	    member->toObjFile(0);
 	}
     }
 }
 
 /* ================================================================== */
 
-void VarDeclaration::toObjFile()
+void VarDeclaration::toObjFile(int multiobj)
 {
     Symbol *s;
     unsigned sz;
     Dsymbol *parent;
 
     //printf("VarDeclaration::toObjFile(%p '%s' type=%s) protection %d\n", this, toChars(), type->toChars(), protection);
+    //printf("\talign = %d\n", type->alignsize());
 
     if (aliassym)
-    {	toAlias()->toObjFile();
+    {	toAlias()->toObjFile(0);
+	return;
+    }
+
+    // Do not store variables we cannot take the address of
+    if (!canTakeAddressOf())
+    {
 	return;
     }
 
@@ -1044,12 +1055,14 @@ void VarDeclaration::toObjFile()
 	sz = type->size();
 
 	parent = this->toParent();
-#if 1	/* private statics should still get a global symbol, in case
+#if 0	/* private statics should still get a global symbol, in case
 	 * another module inlines a function that references it.
 	 */
 	if (/*protection == PROTprivate ||*/
 	    !parent || parent->ident == NULL || parent->isFuncDeclaration())
+	{
 	    s->Sclass = SCstatic;
+	}
 	else
 #endif
 	{
@@ -1058,28 +1071,29 @@ void VarDeclaration::toObjFile()
 	    else
 		s->Sclass = SCglobal;
 
-	    do
-	    {
-		/* Global template data members need to be in comdat's
-		 * in case multiple .obj files instantiate the same
-		 * template with the same types.
-		 */
-		if (parent->isTemplateInstance() && !parent->isTemplateMixin())
-		{
-		    /* These symbol constants have already been copied,
-		     * so no reason to output them.
-		     * Note that currently there is no way to take
-		     * the address of such a const.
-		     */
-		    if (isConst() && type->toBasetype()->ty != Tsarray &&
-			init && init->isExpInitializer())
-			return;
-
-		    s->Sclass = SCcomdat;
-		    break;
-		}
-		parent = parent->parent;
-	    } while (parent);
+            do
+            {
+                /* Global template data members need to be in comdat's
+                 * in case multiple .obj files instantiate the same
+                 * template with the same types.
+                 */
+                if (parent->isTemplateInstance() && !parent->isTemplateMixin())
+                {
+#if 0
+                    /* These symbol constants have already been copied,
+                     * so no reason to output them.
+                     * Note that currently there is no way to take
+                     * the address of such a const.
+                     */
+                    if (isConst() && type->toBasetype()->ty != Tsarray &&
+                        init && init->isExpInitializer())
+                        return;
+#endif
+                    s->Sclass = SCcomdat;
+                    break;
+                }
+                parent = parent->parent;
+            } while (parent);
 	}
 	s->Sfl = FLdata;
 
@@ -1157,7 +1171,7 @@ void VarDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void TypedefDeclaration::toObjFile()
+void TypedefDeclaration::toObjFile(int multiobj)
 {
     //printf("TypedefDeclaration::toObjFile('%s')\n", toChars());
 
@@ -1196,17 +1210,20 @@ void TypedefDeclaration::toObjFile()
 
 /* ================================================================== */
 
-void EnumDeclaration::toObjFile()
+void EnumDeclaration::toObjFile(int multiobj)
 {
     //printf("EnumDeclaration::toObjFile('%s')\n", toChars());
 
+    if (isAnonymous())
+	return;
+
     if (global.params.symdebug)
 	toDebug();
 
     type->getTypeInfo(NULL);	// generate TypeInfo
 
     TypeEnum *tc = (TypeEnum *)type;
-    if (type->isZeroInit() || !tc->sym->defaultval)
+    if (!tc->sym->defaultval || type->isZeroInit())
 	;
     else
     {
@@ -1227,7 +1244,7 @@ void EnumDeclaration::toObjFile()
 #if ELFOBJ // Burton
 	sinit->Sseg = CDATA;
 #endif /* ELFOBJ */
-	dtnbytes(&sinit->Sdt, tc->size(0), (char *)&tc->sym->defaultval);
+	tc->sym->defaultval->toDt(&sinit->Sdt);
 	//sinit->Sdt = tc->sym->init->toDt();
 	outdata(sinit);
     }
