@@ -71,8 +71,10 @@ Expression *Type::getInternalTypeInfo(Scope *sc)
     switch (t->ty)
     {
 	case Tsarray:
+#if 0
 	    // convert to corresponding dynamic array type
 	    t = t->nextOf()->mutableOf()->arrayOf();
+#endif
 	    break;
 
 	case Tclass:
@@ -142,7 +144,7 @@ Expression *Type::getTypeInfo(Scope *sc)
 	    }
 	    else			// if in obj generation pass
 	    {
-		t->vtinfo->toObjFile();
+		t->vtinfo->toObjFile(global.params.multiobj);
 	    }
 	}
     }
@@ -310,8 +312,12 @@ void TypeInfoEnumDeclaration::toDt(dt_t **pdt)
      *	void[] m_init;
      */
 
-    sd->memtype->getTypeInfo(NULL);
-    dtxoff(pdt, sd->memtype->vtinfo->toSymbol(), 0, TYnptr);	// TypeInfo for enum members
+    if (sd->memtype)
+    {	sd->memtype->getTypeInfo(NULL);
+	dtxoff(pdt, sd->memtype->vtinfo->toSymbol(), 0, TYnptr);	// TypeInfo for enum members
+    }
+    else
+	dtdword(pdt, 0);
 
     char *name = sd->toPrettyChars();
     size_t namelen = strlen(name);
@@ -319,7 +325,7 @@ void TypeInfoEnumDeclaration::toDt(dt_t **pdt)
     dtabytes(pdt, TYnptr, 0, namelen + 1, name);
 
     // void[] init;
-    if (tinfo->isZeroInit() || !sd->defaultval)
+    if (!sd->defaultval || tinfo->isZeroInit())
     {	// 0 initializer, or the same as the base type
 	dtdword(pdt, 0);	// init.length
 	dtdword(pdt, 0);	// init.ptr
@@ -442,7 +448,9 @@ void TypeInfoStructDeclaration::toDt(dt_t **pdt)
      *	int function(in void*, in void*) xopCmp;
      *	string function(const(void)*) xtoString;
      *	uint m_flags;
-     *  xgetMembers
+     *  xgetMembers;
+     *	xdtor;
+     *	xpostblit;
      *
      *	name[]
      */
@@ -475,9 +483,10 @@ void TypeInfoStructDeclaration::toDt(dt_t **pdt)
 	Scope sc;
 
 	tftohash = new TypeFunction(NULL, Type::thash_t, 0, LINKd);
+	tftohash->mod = MODconst;
 	tftohash = (TypeFunction *)tftohash->semantic(0, &sc);
 
-	tftostring = new TypeFunction(NULL, Type::tchar->constOf()->arrayOf(), 0, LINKd);
+	tftostring = new TypeFunction(NULL, Type::tchar->invariantOf()->arrayOf(), 0, LINKd);
 	tftostring = (TypeFunction *)tftostring->semantic(0, &sc);
     }
 
@@ -489,6 +498,7 @@ void TypeInfoStructDeclaration::toDt(dt_t **pdt)
 
 	arguments->push(arg);
 	tfeqptr = new TypeFunction(arguments, Type::tint32, 0, LINKd);
+	tfeqptr->mod = MODconst;
 	tfeqptr = (TypeFunction *)tfeqptr->semantic(0, &sc);
     }
 
@@ -560,6 +570,20 @@ void TypeInfoStructDeclaration::toDt(dt_t **pdt)
     else
 	dtdword(pdt, 0);			// xgetMembers
 
+    // xdtor
+    FuncDeclaration *sdtor = sd->dtor;
+    if (sdtor)
+	dtxoff(pdt, sdtor->toSymbol(), 0, TYnptr);
+    else
+	dtdword(pdt, 0);			// xdtor
+
+    // xpostblit
+    FuncDeclaration *spostblit = sd->postblit;
+    if (spostblit)
+	dtxoff(pdt, spostblit->toSymbol(), 0, TYnptr);
+    else
+	dtdword(pdt, 0);			// xpostblit
+
     // name[]
     dtnbytes(pdt, namelen + 1, name);
 }
@@ -627,7 +651,7 @@ void TypeInfoTupleDeclaration::toDt(dt_t **pdt)
     dtxoff(pdt, s, 0, TYnptr);		    // elements.ptr
 }
 
-void TypeInfoDeclaration::toObjFile()
+void TypeInfoDeclaration::toObjFile(int multiobj)
 {
     Symbol *s;
     unsigned sz;
@@ -635,6 +659,12 @@ void TypeInfoDeclaration::toObjFile()
 
     //printf("TypeInfoDeclaration::toObjFile(%p '%s') protection %d\n", this, toChars(), protection);
 
+    if (multiobj)
+    {
+	obj_append(this);
+	return;
+    }
+
     s = toSymbol();
     sz = type->size();
 
